/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * ils.c
 *
 * Manage the 'kls' command.
 ***************************************************************************/

#include <stdio.h>

#include "interp.h"
#include "kediti.h"

static char *documentation =
"kls [-l] [-oid <node>,<name>]\n\
This command performs an Ls request on the given node. The node is given\n\
either through its identifier, by using the \"-oid <node>\" form, or through\n\
its name. Giving a node by its name is possible only if the \"mkdb\" command\n\
has been run before. Finally if no argument is provided, and a current\n\
directory is defined, kls will list this one.\n\
If the \"-l\" option is present, this function will also querry children\n\
status, to print a full description of children.\n" ;

/***************************************************************************
 * ParseLs: parse the ls command
 *
 * RETURNS: 0
 ***************************************************************************/

static int ParseLs (argc, argv)
    int  argc ;
    char **argv ;

{
    KID 	   oid ;
    KeIdArray	   result ;
    KeStatus       status ;
    int            count, verbose = 0 ;
    KID            *pid ;
    int            children ;

/*
 * Parse first arg(s) to find the target node identifier:
 */

    argc-- ; argv++ ;

    if ( argc && ( ! strcmp (*argv, "-l") ) )
    {
	verbose = 1 ;
	argc-- ; 
	argv++ ;
    }

    if ( ! argc )
    {
	/*
	 * Use current directory, if any:
	 */
	if ( PathCount > 0 )
	{
	    oid = Path[PathCount-1] ;
	}
	else
	{
	    fprintf (stderr, "\tNo current directoy, provide node.\n") ;
	    return (0) ;
	}
    }
    else
    {
	/*
	 * Get the rag using the ParseId function.
	 */
	if ( ParseId (argc, argv, &oid) <= 0 )
	    return (1) ;
    }

/*
 * oid contains the node to be listed, we now perfom the Ls request in
 * synchronous mode.
 */

    if ( verbose )
    {
	int        i, count ;
	KeNodeStat *stats ;

	if ( (count = KecLsl (KecConn, oid, &stats)) > 0 )
	{
	    for ( i = 0 ; i < count ; i++ )
	    {
		fprintf (stdout, "\t%-30.30s\tx%08x (%d) %s %s\n"
			 , stats[i]->name
			 , *pid
			 , *pid
			 , PrintMode (stats[i]->mode)
			 , PrintMask (stats[i]->mask)) ;
	    }
	    KecLslFree (stats) ;
	}
	else
	{
	    printf ("\tKecLsl failed.\n") ;
	}
    }
    else
    {
	if ( status = KecLsSync (KecConn, oid, &result) )
	{
	    KecError error = KecErrorOfConnection (KecConn) ;

	    /*
	     * An error occured. Retreive it from the connection and display 
	     * the associated string message (if any)
	     */

	    printf ("\tkls failed (status = %d).\n", status) ;
	    if ( error->msg )
		printf ("\t%s\n", error->msg) ;
	    
	    return (0) ;
	}


	fprintf (stdout, "Found %d children:\n", result->count) ;
    
	for ( children = result->count, pid = result->nodes
	     ; children 
	     ; children--, pid++ )
	{
	    DBEntry entry = GetNodeEntry (*pid) ;
	    
	    fprintf (stdout, "\t%-30.30s\tx%08x (%d)\n"
		     , entry ? entry->name : ""
		     , *pid
		     , *pid) ;
	}
	
	KeFree (result) ;
    }

    return (0) ;
}

/***************************************************************************
 * LsInitialize: declare the 'kls' command
 *
 * RETURNS: void
 ***************************************************************************/

void LsInitialize ()

{
    DeclareCommand ("kls"
		    , documentation
		    , ParseLs) ;
    return ;
}
