/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * interp.c
 *
 * This file defines a minimal interpretor: it just break input lines into
 * their (argc, argv) form, and pass this, through an indirection, to the
 * argv[0] corresponding function pointer.
 ***************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/times.h>

#include "interp.h"

static Command *Commands     = (Command *) 0 ;
static int     CommandsCount = 0 ;

/***************************************************************************
 * BreakLine: break a line into its argc, argv form
 *
 * This make use of a copy of  the buffer in a static area, so wraning
 * when you call it twice
 *
 * RETURN: 0 on success, error else.
 ***************************************************************************/

static int BreakLine (buffer, size, pargc, pargv)
    char *buffer ;
    int  size ;
    int  *pargc ;
    char **pargv ;

{
    static char static_buffer[MAX_COMMAND_LINE_SIZE] ;
    char        *p ;
    
/*
 * Get a fresh copy of command line
 */

    if ( strlen (buffer) >= MAX_COMMAND_LINE_SIZE )
	return (1) ;
    strcpy (static_buffer, buffer) ;

/*
 * Spli it into argc, argv
 */
    *pargc   = 1 ;
    *pargv++ = static_buffer ;
    for ( p = static_buffer ; *p && ( (*p == ' ') || (*p =='\t') ) ; p++ )
	;
    while ( *p )
    {
	switch (*p)
	{
	  case '\t':
	  case ' ':
	    /* Check for new argv entry */
	    if ( (*pargc) + 1 >= size )
		return (1) ;
	    /* Set an EOS and skipp all trailing blanks */
	    *p++ = '\0' ;
	    while ( *p && ((*p == ' ') || (*p == '\t')) )
		p++ ;
	    /* If still something, this is another field */
	    if ( *p )
	    {
		*pargv++ = p ;
		(*pargc)++ ;
	    }
	    break ;
	  default:
	    /* Skip any 'normal' char */
	    p++ ;
	    break ;
	}
    }

    /* We are done */

    return (0) ;
}

/***************************************************************************
 * DispatchCommand: given an argc, argv array, invoke command handler
 *
 * This is not very efficient, no probs, we are not looking for this.
 *
 * RETURNS: (<0) if command not found, command ret val otherwise
 ***************************************************************************/

int DispatchCommand (argc, argv)
    int  argc ;
    char **argv ;

{
    register Command *pcommand ;
    register int     i ;
    int              error = -1 ;

    for ( pcommand = Commands, i = 0 ; i < CommandsCount ; i++, pcommand++)
    {
	if ( ! strcmp ((*pcommand)->name, *argv) )
	    return ( ((*pcommand)->handler) (argc, argv) ) ;
    }

    return (error) ;

}

/***************************************************************************
 * ParseAndExec: parse the command line, and exec the appropriate command
 *
 * RETURNS: (<0) if command not found, handler retval otherwise
 ***************************************************************************/

int ParseAndExec (buffer)
    char *buffer ;

{
    char *argv[MAX_ARGS] ;
    int  argc ;
    int  error ;
    char *p, *cmd ;

/* 
 * Get rid of empty lines, commented lines and trailing blanks :
 */

    for ( p = buffer ; (*p != '\n') && ((*p == ' ') || (*p == '\t')) ; p++ )
	;
    if ( (*p == '\n') || (*p == '#') )
	return (0) ;
    for ( cmd = p ; (*p) && (*p != '\n') ; p++ )
	;
    *p = '\0' ;

/*
 * Break buffer into argc, argv and dispatch :
 */

    if ( BreakLine (cmd, MAX_ARGS, &argc, argv) )
    {
	printf ("*** Error while breaking line (implementation bug).\n");
	return (-1) ;
    }

    if ( (error = DispatchCommand (argc, argv)) )
    {
	printf ("*** Command failed (bad syntax).\n") ;
    }

    return (error) ;
}

/***************************************************************************
 * DeclareCommand: declare a new command
 *
 * Create a new Commamd entry, and add it to the array of well-knwon 
 * commands.
 * name and doc are not copied, so the caller should not free them
 *
 * RETURNS: none (should return an error in fact)
 ***************************************************************************/

void DeclareCommand (name, doc, handler)
    char *name ;
    char *doc ;
    int (*handler)() ;

{
    Command new ;

/*
 * Reallocate the Command array :
 */
    CommandsCount++ ;
    if ( CommandsCount > 1 )
    {
	if ( ! ( Commands = (Command *) realloc (Commands
						 , CommandsCount * sizeof (Command *))) )
	{
	    perror ("DeclareCommand, malloc ") ;
	    exit (1) ;
	}
    }
    else
    {
	if ( ! (Commands = (Command *) malloc (sizeof (Command *))) )
	{
	    perror ("DeclareCommand, malloc ") ;
	    exit (1) ;
	}
    }
/*
 * Set up the new entry :
 */
    if ( ! (new = (Command) malloc (sizeof (CommandRec))) )
    {
	perror ("DeclareCommand, malloc ") ;
	exit (1) ;
    }
    
    Commands[CommandsCount - 1] = new ;

    new->name    = name ;
    new->doc     = doc ;
    new->handler = handler ;
    return ;
}

/***************************************************************************
 * ParseDoc: parse the doc command
 * ParseHelp: list all available commands
 *
 * This function uses the Commands array, that's why we found it here
 *
 * RETURN: 0
 ***************************************************************************/

static int ParseDoc (argc, argv)
    int argc ;
    char **argv ;

{
    register Command *pcommand ;
    register int     i ;

/*
 * Parse command line :
 */

    if ( argc != 2 )
    {
	printf ("\tdoc <name>\n") ;
	return (1) ;
    }
    argc-- ;
    argv++ ;
	
/*
 * Look for this command :
 */
    for ( i = 0 , pcommand = Commands ; i < CommandsCount ; i++, pcommand++ )
    {
	if ( ! strcmp (*argv, (*pcommand)->name) )
	{
	    printf ("Documentation for [%s]:\n%s\n"
		    , *argv
		    , (*pcommand)->doc) ;
	    return (0) ;
	}
    }

    printf ("No documentation for [%s]\n", *argv) ;
    return (0) ;
}

static int ParseHelp (argc, argv)
    int  argc ;
    char **argv ;

{
    register Command *pcommand ;
    register int i ;

    if ( argc != 1 )
    {
	printf ("\thelp\n") ;
	return (1) ;
    }

    for ( i = 0 , pcommand = Commands ; i < CommandsCount ; i++, pcommand++ )
    {
	char *doc = (*pcommand)->doc ;

	while ( *doc != '\n' )
	    fputc (*doc++, stdout) ;
	fputc ('\n', stdout) ;
    }

    return (0) ;
}

/***************************************************************************
 * ParseLoad: parse a 'load' command
 *
 * Execute line in file
 *
 * RETURN: 0
 ***************************************************************************/

static int ParseLoad (argc, argv)
    int  argc ;
    char **argv ;

{
    char buffer[4096] ;
    FILE *fd ;
    int trace = 0 ;
    char *fn ;

    /* Decode args */
    if ( (argc == 2) && (*argv[1] != '-') ) {
	fn = argv[1] ;
    } else if ( (argc == 3) && ( ! strcmp (argv[1], "-trace")) ) {
	trace = 1 ;
	fn = argv[2] ;
    } else {
	printf ("load [-trace] <filename>\n") ;
	return (1) ;
    }
    /* Do it */
    if ( (fd = fopen (fn, "r")) == NULL )
    {
	printf ("\tcouldn't open file %s\n", fn) ;
	return (0) ;
    }
    else
    {
	while ( fgets (buffer, 4096, fd) )
	{
	    /*
	     * Trace if needed and ParseAndExec :
	     */
	    if ( trace )
		printf (">>> %s", buffer) ;
	    ParseAndExec (buffer) ;
	}
	fclose (fd) ;
	return (0) ;
    }
}
	       
/***************************************************************************
 * ParseShell: parse the "sh" command
 *
 * RETURNS: status of launched command
 ***************************************************************************/

static int ParseShell (argc, argv)
    int argc ;
    char **argv ;

{
    char *p ;

    if ( argc == 1 )
    {
	printf ("sh <command line>.\n") ;
	return (1) ;
    }

/*
 * As *I* now what happend behind, I can do this hack :
 */
    for ( argc -= 2 , argv++ , p = *argv ; argc ; p++ )
    {
	if ( ! *p )
	{
	    argc-- ;
	    *p = ' ' ;
	}
    }

    printf ("\treturned status: %d\n", system (*argv)) ;
    return (0) ;
}

/***************************************************************************
 * ParseExit: parse the exit command
 *
 * RETURNS: !!
 ***************************************************************************/

static int ParseExit (argc, argv) 
    int  argc ;
    char **argv ;

{
    printf ("\t1003.1003: User exited normally.\n") ;
    exit (1) ;
}

/***************************************************************************
 * ParseRepeat: parse the 'repeat' command
 *
 * RETURNS: repeated command error code
 ***************************************************************************/

static int ParseRepeat (argc, argv)
    int  argc ;
    char **argv ;

{
    int count, error ;

    /* Get the repeat count, and reparse end of command */
    if ( argc <= 2 )
    {
	printf ("\trepeat <count> <command>.\n") ;
	return (1) ;
    }
    for ( count = atoi (argv[1]) ; count > 0 ; count-- )
    {
	printf ( "Remaining iterations (;-) %d\n", count) ;
	if ( (error = DispatchCommand (argc - 2, argv + 2)) < 0 )
	    break ;
    }

    return ( error ) ;
}

/***************************************************************************
 * ParseTime: parse the 'times' command
 *
 * RETUNS: subcommanbd error
 ***************************************************************************/

static int ParseTimes (argc, argv)
    int  argc ;
    char **argv ;

{
    struct tms start, stop ;
    int        error ;

    times (&start) ;
    error = DispatchCommand (argc - 1, argv + 1) ;
    times (&stop) ;

    fprintf (stderr, "\tuser delta: %d\n", stop.tms_utime - start.tms_utime) ;
    fprintf (stderr, "\tsys  delta: %d\n", stop.tms_stime - start.tms_stime) ;
    return (error) ;
}



/***************************************************************************
 * InterpInitialize: declare all the predefined functions
 ***************************************************************************/

void InterpInitialize ()

{
    extern int ParseShell(), ParseExit(), ParseLoad(), ParseRepeat() ;

    DeclareCommand ("sh"
		    , "sh <command line>\nInvoke a shell to process <command-line>."
		    , ParseShell) ;

    DeclareCommand ("exit"
		    , "exit\nExit normally from application."
		    , ParseExit) ;

    DeclareCommand ("load"
		    , "load <cmd-file>\nLoad and execute the given file."
		    , ParseLoad) ;
    
    DeclareCommand ("repeat"
		    , "repeat <count> <cmd>\nRepeat command count times."
		    , ParseRepeat) ;

    DeclareCommand ("times"
		    , "times <comand>\ngive the user/sys timing for command\n"
		    , ParseTimes) ;

    DeclareCommand ("doc"
		    , "doc <name>\nretrieve the doc associated with <name>"
		    , ParseDoc) ;

    DeclareCommand ("help"
		    , "help\nList all available commands with their help."
		    , ParseHelp) ;

    return ;
}
