/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * mkdb.c
 *
 * Make a full inspection of application, creating a hashtable from names
 * to oid. Than, prepare to parse command line.
 ***************************************************************************/

#include <stdio.h>

#include <KeClient.h>

#include "kediti.h"

static KeHashtable NodeToInfo = (KeHashtable) 0 ;

/***************************************************************************
 * node_hash_fct
 * node_cmp_fct
 ***************************************************************************/

unsigned long node_hash_fct (entry) 
    DBEntry entry ;

{
    return ((unsigned long) entry->node) ;
}

int node_cmp_fct (entry1, entry2)
    DBEntry entry1 ;
    DBEntry entry2 ;

{
    return ( entry1->node != entry2->node ) ;
}

/***************************************************************************
 * PrintDBEntry: print on fd the given file
 ***************************************************************************/

static void PrintDBEntry (entry, fd)
    DBEntry entry ;
    FILE    *fd ;

{
    fprintf (fd, "%-50.50s x%08x (%d)\n"
	     , entry->name
	     , entry->node
	     , entry->node) ;

    return ;
}

DBEntry GetNodeEntry (node)
    KID node ;

{
    DBEntryRec entry_rec, *entry ;

    if ( ! NodeToInfo )
	return ((DBEntry) 0) ;

    entry_rec.node = node ;

    return ((DBEntry) KeHashtableFind (NodeToInfo, &entry_rec)) ;
}

static void AddNode (node, name) 
    KID  node ;
    char *name ;

{
    DBEntryRec entry_rec ; 
    DBEntry    entry ;

    entry_rec.node = node ;

    if ( entry = (DBEntry) KeHashtableFind (NodeToInfo, &entry_rec) )
    {
	if ( entry->name )
	{
	    if ( strcmp (entry->name, name) )
	    {
		free (entry->name) ;
		entry->name = (char *) strdup (name) ;
	    }
	}
	else
	{
	    entry->name = (char *) strdup (name) ;

	}
	return ;
    }

    if ( entry = (DBEntry) malloc (sizeof (DBEntryRec)) )
    {
	entry->name     = (char *) strdup (name) ;
	entry->node     = node ;
	entry->children = (KeIdArray) 0 ;
	
	KeHashtableAdd (NodeToInfo, entry) ;
    }

    return ;
}

static void SetNodeChildren (node, children)
    KID       node ;
    KeIdArray children ;

{
    DBEntryRec entry_rec, *entry ;

    entry_rec.node = node ;

    if ( entry = (DBEntry) KeHashtableFind (NodeToInfo, &entry_rec) )
    {
	if ( entry->children )
	{
	    KeFree (entry->children) ;
	}
	entry->children = children ;
    }
    else
    {
	fprintf (stderr, "SetNodeChildren: undeclared node x%08x\n", node) ;
    }

    return ;
}

/***************************************************************************
 * KecDBMake: create (or update) the static database.
 *
 * This function is *really* a nightmare. Algo is the following:
 *
 * - Put the root node in ToStat, call PerformStats
 * - PerfomStats: Stat on each node in the ToStat list, with StatHandler
 * - StatHandler looks if stat describes a directory. If it does, than
 *   put the node in the ToLs array.
 * - Once all pending Stat replies have been received, if there are some
 *   nodes in ToLs, ls for all nodes, with LsHandler, otherwise, job is
 *   done.
 * - LsHandler put the nodes in the ToStat array, then calls PerformStats
 *
 * This algorithm ensures that communication with server is always in the
 * same direction: either the client is sending requests (and can manage
 * the reply in the meantime), or the client is listening for replies.
 *
 * Another possible strategy is the following:
 *
 * - Stat on root node with StatHanldler
 * - StatHandler: if stat describes a directory, Ls on it with LsHandler
 * - LsHandler  : stat on each child node with StatHandler
 * Algo finishes when no more requests are pending. This second solution
 * although quite elegant (at least from my point of view) is bugged: 
 * client and server can get into a dead-lock beacuse both of the write on 
 * socket *at* the same time (ie, clients requests are intespeded with 
 * server replies, wich is not the case for the preceding algo).
 *
 * The second versin uses less memory, but rely implicitly on kernel
 * comm buffers, wich is BAD.
 ***************************************************************************/

static KecConnection CurrentConnection ;
static int           NodeCount, MaxLs, MaxStat ;
static KID	     ToStat[32 * 1024], *ToStatPtr ;
static KID	     ToLs[32 * 1024], *ToLsPtr ;

static void StatHandler(), LsHandler() ;

static void LsHanlder (reply, closure)
    KecReply  reply ;
    KePointer closure ;

{
    if ( ! reply->status )
    {
	KeIdArray nodes = KecIdArrayOfReply (reply) ;
	int       i ;
	
	if ( TraceFlag )
	    fprintf (stderr, "Ls handler: got %d nodes.\n", nodes->count) ;

	for ( i = 0 ; i < nodes->count ; i++ )
	    *ToStatPtr++ = nodes->nodes[i] ;

	SetNodeChildren ((KID) closure, nodes) ;
    }
    return ;
}
	
static int PerformLs (closure)
    KePointer closure ;

{
    int count = ToLsPtr - ToLs ;
    int needs_gobble = 0 ;

    if ( ! count )
	return (count) ;

    if ( count > MaxLs )
	MaxLs = count ;

    while ( ToLsPtr > ToLs )
    {
	--ToLsPtr ;
	if ( KecLs (CurrentConnection
		    , *ToLsPtr
		    , LsHanlder
		    , (KePointer) *ToLsPtr) )
	{
	    fprintf (stderr, "Problems while ls'ing.\n") ;
	    exit (1) ;
	}
	if ( needs_gobble++ > 100 )
	{
	    if ( TraceFlag )
		fprintf (stderr, "PerformLs: gobelling data.\n") ;

	    needs_gobble = 0 ;
	    KecGobbleData (CurrentConnection) ;
	}
    }

    return (count) ;
}

static void StatHandler (reply, closure)
    KecReply  reply ;
    KePointer closure ;

{
    char buffer[512] ;

    if ( ! reply->status )
    {
	KeNodeStat stat = KecStatOfReply (reply) ;
	char       *full_path ;
	char	   *path = (char *) closure ;

	/*
	 * Add the node to the data-base (keyed by full-name):
	 */

	AddNode (stat->id, stat->name) ;

	NodeCount++ ;

	/*
	 * If it is a directory, put it in the ToLs stack:
	 */

	if ( stat->mode & KEDIT_DirectoryMode )
	    *ToLsPtr++ = stat->id ;

	KeFree (stat) ;
    }
    return ;
}

static int PerformStat (closure)
    KePointer closure ;

{
    int count        = ToStatPtr - ToStat ;
    int needs_gobble = 0 ;

    if ( ! count )
	return (0) ;

    if ( count > MaxStat )
	MaxStat = count ;

    while ( ToStatPtr != ToStat )
    {
	if ( KecStat (CurrentConnection, *--ToStatPtr, StatHandler, closure ) )
	{
	    fprintf (stderr, "Problems while stat'ing.\n") ;
	    exit (1) ;
	}
	if ( needs_gobble++ > 100 )
	{
	    needs_gobble = 0 ;
	    KecGobbleData (CurrentConnection) ;
	}
    }

    return ( count ) ;
}

KecDBMake (kc)
    KecConnection kc ;

{

    CurrentConnection = kc ;
    NodeCount = 0 ;

/*
 * Initialize KeHashtable if not done:
 */

    if ( ! NodeToInfo )
    {
	NodeToInfo = KeHashtableMake (DBENTRY_HASHTABLE_SIZE
				    , node_hash_fct
				    , node_cmp_fct) ;
    }

/*
 * Start it up :
 */

    ToLsPtr   = ToLs ;
    ToStatPtr = ToStat ;
    NodeCount = 0 ;
    MaxLs     = 0 ;
    MaxStat   = 0 ;

    *ToStatPtr++ = KecRootNodeOfConnection (kc) ;

    while ( 1 )
    {
	if ( ! PerformStat ((KePointer) 0) )
	    goto done ;
	KecGobbleData (kc) ;
	if ( ! PerformLs ((KePointer) 0) )
	    goto done ;
	KecGobbleData (kc) ;
    }

  done:

    fprintf (stderr, "mkdb: %d nodes, %d MaxStat, %d MaxLs\n"
	     , NodeCount
	     , MaxStat
	     , MaxLs) ;

    return (NodeCount) ;
}

/***************************************************************************
 * KecDBReset (kc)
 *
 * Reset the static data base. In fact, we don't need the connection.
 *
 * RETURN: void.
 ***************************************************************************/

void RemoveDBEntry (entry, unused)
    DBEntry entry ;
    void    *unused ;

{
    KeHashtableRemove (NodeToInfo, entry) ;

    if ( entry->name )
	free (entry->name) ;
    if ( entry->children )
	KeFree (entry->children) ;
	    
    free (entry) ;
}

void KecDBReset (kec)
    KecConnection kec ;

{

    KeHashtableMap (NodeToInfo, RemoveDBEntry, (char *) 0) ;

    return ;
}

/***************************************************************************
 * KecDBDump: dump in ascii the current state of db to given fd
 ***************************************************************************/

KecDBDump (kec, fd)
    KecConnection kec ;
    FILE *fd ;

{

    KeHashtableMap (NodeToInfo, PrintDBEntry, fd) ;

    return ;
}

