/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * parse.c
 *
 * Various parsing utility functions.
 ***************************************************************************/

#include <stdio.h>
#include <ctype.h>

#include "interp.h"
#include "kediti.h"

KID  Path[256] ;		/* node for current path */
int  PathCount ;		/* path size */
int  TraceFlag = 1 ;

/***************************************************************************
 * ParseTrace: set/unset the Trace var
 *
 * RETURNS: 0
 ***************************************************************************/

static char *trace_doc = 
"trace <true/false>\n\
This turns on/off various verbosities.\n" ;

static int ParseTrace (argc, argv)
    int  argc ;
    char **argv ;

{
    if ( argc != 2 )
	return (-1) ;
    TraceFlag = toupper ((*(argv+1))[0]) == 'T' ;
    return (0) ;
}

/***************************************************************************
 * ParseId: parse to ge an identifier
 *
 * RETURNS: number of argc used in argv
 ***************************************************************************/

int ParseId (argc, argv, pid)
    int  argc ;
    char **argv ;
    KID *pid ;

{
    char buffer[1024] ;

    if ( (argc >= 2)  && (! strcmp (*argv, "-oid")) )
    {
	*pid = (KID) atoi (*(argv+1)) ;
	return (2) ;
    } 
    else 
    {
	DBEntry entry, dir ;
	int     i ;

	if ( ( ! PathCount )
	    || ( ! (dir = GetNodeEntry (Path[PathCount-1])) )
	    || ( ! dir->children ) )
	{
	    return (-1) ;
	}

	for ( i = 0 ; i < dir->children->count ; i++ )
	{
	    if ( entry = GetNodeEntry (dir->children->nodes[i]) )
	    {
		if ( ! strcmp (entry->name, *argv) )
		{
		    *pid = entry->node ;
		    return (1) ;
		}
	    }
	}
    }

    return (-1) ;
}

/***************************************************************************
 * ParseCd: parse the 'cd' command
 * ParsePwd: parse the 'pwd' command
 *
 * RETURNS: 0
 ***************************************************************************/

static char *cd_doc =
"cd <name>\n\
This command is used in conjunction with \"mkdb\" and related command.\n\
It allows kediti user to see the target server as a file system.\n\
This comand is not fully implemented, here is the precise arg you\n\
can give:\n\
\t-oid <oid>: if there is a current path, and oid identifies a\n\
\tchild of current directory, then go to that directory. If no\n\
\tchild is found the command fails. If directory is the root\n\
\tdirectory of target server the reset path to this.\n\
\t..        : if there is a current path, go up one level.\n\
\t          : reset path (no more current path defined)\n." ;

static int ParseCd (argc, argv)
    int argc ;
    char **argv ;

{
    int     count, i ;
    KID     oid ;
    DBEntry entry ;

    argc-- ; argv++ ;

/*
 * Find back entry corresponding to given object:
 */

    if ( (count = ParseId (argc, argv, &oid)) <= 0 )
    {
	/*
	 * Look for peculiar strings:
	 */
	if ( ! argc ) 
	{
	    PathCount = 0 ;
	    return (0) ;
	} 
	else if ( ! strcmp (*argv, "..") ) 
	{
	    if ( PathCount > 0 ) 
	    {
		PathCount-- ;
		return (0) ;
	    }
	    else
	    {
		fprintf (stderr, "\tAlready at root !\n") ;
		return (0) ;
	    }
	}
	else
	{
	    fprintf (stderr, "\tInvalid arg, use \"help\"\n") ;
	    return (0) ;
	}
    }
    
    if ( ! ( entry = GetNodeEntry (oid)) )
    {
	fprintf (stderr, "\tinvalid path (not registered in cache).\n") ;
	return (0) ;
    }

/*
 * Can this be a relative path ?
 */

    if ( PathCount > 0 )
    {
	DBEntry dir = GetNodeEntry (Path[PathCount-1]) ;

	/*
	 * Look in last path component to find the new component: if
	 * found, than set the path in a relative fashion.
	 */

	if ( dir )
	{
	    for ( i = 0 ; i < dir->children->count ; i++ )
	    {
		if ( dir->children->nodes[i] == oid )
		{
		    Path[PathCount++] = oid ;
		    return (0) ;
		}
	    }
	}
	
	fprintf (stderr, "\tInvalid relative path.\n") ;
	return (0) ;
    }

/*
 * Ok, this is the root :
 */

    if ( oid != KecRootNodeOfConnection (KecConn) )
	fprintf (stderr, "\tInvalid absolute path.\n") ;

    Path[PathCount++] = oid ;

    return (0) ;
}

static char *pwd_doc =
"pwd\n\
returns the current \"working directory\".\n" ;

static int ParsePwd (argc, argv)
    int argc ;
    char **argv ;

{
    DBEntry entry ;
    int     i ;

    if ( argc != 1 )
	return (1) ;

    printf ("/") ;

    for ( i = 0 ; i < PathCount - 1 ; i++ )
    {
	if ( entry = GetNodeEntry (Path[i] ) )
	    printf ("%s/", entry->name) ;
    }

    if ( entry = GetNodeEntry (Path[i]) )
	printf ("%s\n", entry->name) ;
    else
	printf ("\n") ;

    return (0) ;
}
	
/***************************************************************************
 * ParseMkdb:  parse the mkdb command
 * ParseRstdb: parse the resetdb command
 *
 * RETRUNS: 0 
 ***************************************************************************/

static char *mkdb_doc =
"mkdb\n\
This command will fill in a client cache of all the target server exported\n\
nodes, so that following command can design nodes by their symbolic name\n\
instead of their node identifier.\n\
This command should be run at the beginning of the session, so that all other\n\
commands accepts as node designers their symbolic name.\n" ;

static int ParseMkdb (argc, argv)
    int argc ;
    char **argv ;

{
    int nodes ;

    if ( argc != 1 )
	return (1) ;
    if ( TraceFlag )
    {
	printf ("Rebuilding database...") ;
	fflush (stdout) ;
    }
    nodes = KecDBMake (KecConn) ;
    if ( TraceFlag )
	printf ("done (%d nodes).\n", nodes) ;
    return (0) ;
}

static char *resetdb_doc =
"resetdb\n\
This command destroyed the node cache created by the \"mkdb\" command. You\n\
should re-run \"mkdb\" after this to rebuild a clean cache.\n" ;

static int ParseRstdb (argc, argv)
    int argc ;
    char **argv ;

{
    if ( argc != 1 )
	return (1) ;
    KecDBReset (KecConn) ;
    return (0) ;
}

static char *dumpdb_doc =
"dumpdb [-o <file]\n\
Dump current node cache. If no <file> argument is provided, dump stuff will\n\
be directed to stdout, otherwise, into the given file.\n" ;

static int ParseDumpdb (argc, argv)
    int argc ;
    char **argv ;

{
    FILE *fd = stdout ;
    argc-- ;
    argv++ ;

    if ( argc )
    {
	if ( strcmp (*argv, "-o") || ! (argc - 1) )
	{
	    printf ("\tdumpdb [-o file]\n") ;
	    return (1) ;
	}
	else
	{
	    argc-- ;
	    if ( ! (fd = fopen (*++argv, "w+")) )
	    {
		perror ("cannot create output file ") ;
		return (1) ;
	    }
	}
    }

    KecDBDump (KecConn, fd) ;

    if ( fd != stdout )
	fclose (fd) ;
    
    return (0) ;
}
	
static char *rename_doc = 
"rename <full-path> <alias>\n\
This command allows to define an alias to a node. This is very usefull\n\
whe monitoring a node wich is not in the current path. The alias should\n\
always begin with a ` (backquote), in definition as when used.\n\
Example:\n\
# rename /Root/Struct/test-0/a 'a\n\
# cd /Root/Types\n\
# kget 'a\n" ;

static int ParseRename (argc, argv)
    int argc ;
    char **argv ;

{
    KID oid ;

    if ( argc != 3 )
    {
	printf ("rename <full-path> <alias>\n") ;
	return (1) ;
    }
/*
 * Make a new entry in the KecDB (if we can find the source)
 */
    if ( ParseId (1, argv+1, &oid) > 0 )
	fprintf (stderr, "KecDBAddEntry (argv[2], oid): not done\n") ;
    else
	printf ("No object named %s in current path.\n", argv[1]) ;
    return (0) ;
}

/***************************************************************************
 * DBInitialize: declare the database related commands
 *
 * RETURNS: void.
 ***************************************************************************/

void DBInitialize ()

{
    DeclareCommand ("cd"
		    , cd_doc
		    , ParseCd) ;
    DeclareCommand ("pwd"
		    , pwd_doc
		    , ParsePwd) ;
    DeclareCommand ("mkdb"
		    , mkdb_doc
		    , ParseMkdb) ;
    DeclareCommand ("resetdb"
		    , resetdb_doc
		    , ParseRstdb) ;
    DeclareCommand ("dumpdb"
		    , dumpdb_doc
		    , ParseDumpdb) ;
    DeclareCommand ("rename"
		    , rename_doc
		    , ParseRename) ;
    DeclareCommand ("trace"
		    , trace_doc
		    , ParseTrace) ;
		    
    return ;
}
