/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * structsample.c
 *
 * This file demonstrate the usage of the struct interface to k-Edit. 
 * You should perhaps have a look at the documentation before trying to
 * understand what is going on here !
 ***************************************************************************/

#include <stdio.h>

#include <KeServer.h>
#include <KeStruct.h>
#include <KesCodec.h>

/*
 * The datatype we want to export:
 * It is not really safe to define the fields of struct using C
 * data types, we should perhaps use the k-Edit predefined types
 * instead (CARD32, BYTE, STRING...). 
 * Have to think of this...
 */

typedef struct _Test {
    unsigned long b ;		/* a CARD32 in fact */
    char          c ;		/* a BYTE */
    char          *s ;		/* a STRING */
    char          *clone ;	/* a STRING */
} TestRec, *Test ;

/*
 * A few globals:
 */

KesStructClass TestClass ;	/* k-Edit encapsulator for Test structures */
KesServer      Server ;		/* our k-Edit server context */
KesDirectory   StructDirectory;	/* the directory for exported Test objs */

/***************************************************************************
 * TestClone: Exec hook for the 'clone' field
 *
 * Create a clone of given object, and export it. The clone is exported 
 * under the name defined by the value of the 'clone' field in cloned
 * object.
 *
 * RETURNS: a KeStatus.
 ***************************************************************************/

static KeStatus TestClone (object, accessor, emsg, efree)
    KesStruct         object ;
    KesStructAccessor accessor ;
    char	      **emsg ;
    KeErrorFreeFct    *efree ;

{
    Test new    = (Test) malloc (sizeof (TestRec)) ;
    Test cloned = (Test) object->pointer ;

/*
 * Test for a valid name:
 */

    if ( ! cloned->clone ) {
	*emsg  = "Object to clone has no name." ;
	*efree = (KeErrorFreeFct) 0 ;
	return (KEDIT_Success) ;
    }

/*
 * Create the cloned structure:
 */

    fprintf (stderr, "structsample[TestClone]: to object <%s>\n"
	     , cloned->clone) ;

    new->b     = cloned->b ;
    new->c     = cloned->c ;
    new->s     = cloned->s ? (char *) strdup (cloned->s) : (char *) 0 ;
    new->clone = (char *) strdup (cloned->clone) ;

/*
 * Export it :
 */

    object = KesStructMake (Server
			    , TestClass
			    , new->clone
			    , (KePointer) new
			    , KEDIT_CreateMask | KEDIT_DeleteMask) ;

    fprintf (stderr, "structsample: object x%08x node %08x %d\n"
	     , object
	     , object->node
	     , object->node) ;

    KesDirectoryAddChild (StructDirectory, object->node) ;

    return (KEDIT_Success) ;
}

/***************************************************************************
 * main
 *
 * Does all the job:
 * - parse the (simple) command line
 * - open a server context
 * - register the predefined types we are going to use
 * - create the StructClass encapsulator
 * - create a few objects
 * - enter main loop
 *
 * RETURNS: never !
 ***************************************************************************/

main (argc, argv)
    int argc ;
    char **argv ;

{
    int                 i, num_objs = 5 ;
    KesRequest		request ;
    int 		offset ;
    KesHandlerSet	handler_set ;
    KID                 node ;
    unsigned long       full_mask ;

/*
 * Get command line options:
 */

    argc-- ; argv++ ;
    while ( argc ) {
	if ( ! strcmp (*argv, "-n") ) {
	    argv++ ; argc-- ;
	    num_objs = atoi (*argv) ;
	    argv++ ; argc-- ;
	} else {
	    fprintf (stderr, "structsample [-n <# of objects to export>]\n") ;
	    exit (1) ;
	}
    }

/*
 * Create a server context :
 */

    if ( ! ( Server = KesServerInitialize ( "struct-test" ) ) ) 
    {
	fprintf (stderr, "structsample: unable to create a server context\n") ;
	exit (1) ;
    }

/*
 * Pre-register the type we are going to use:
 */

    KesCodecRegister (Server, KeCodecCARD32Get(), &node) ;
    KesCodecRegister (Server, KeCodecSTRINGGet(), &node) ;
    KesCodecRegister (Server, KeCodecBYTEGet(), &node) ;

/*
 * Create a Test encapsulator class :
 *
 * We have to declare an accessor for each of the slot of our struct typedef
 * so that k-Edit can access them. The "type" allows k-Edit to encode/decode
 * properly the fields to/from the net.
 * As we provide no set/get functions, k-Edit will use the 'offset' and
 * access directly to the structure data.
 */

    TestClass = KesStructClassMake ("Test") ;

    full_mask =  ( KEDIT_NotifyMask
		  | KEDIT_ReadMask | KEDIT_ReadWithValueMask
		  | KEDIT_WriteMask | KEDIT_WriteWithValueMask ) ;

    KesStructAccessorDeclare (TestClass
			      , "b"
			      , KEDIT_ReadMode | KEDIT_WriteMode
			      , full_mask
			      , KesOffsetOf (TestRec, b)
			      , KeCodecCARD32Get() 
			      , KesStructNullSlotSet
			      , KesStructNullSlotGet
			      , KesStructNullSlotExec) ;

    KesStructAccessorDeclare (TestClass
			      , "c"
			      , KEDIT_ReadMode | KEDIT_WriteMode
			      , full_mask
			      , KesOffsetOf (TestRec, c)
			      , KeCodecBYTEGet() 
			      , KesStructNullSlotSet
			      , KesStructNullSlotGet
			      , KesStructNullSlotExec) ;

    KesStructAccessorDeclare (TestClass
			      , "s"
			      , KEDIT_ReadMode | KEDIT_WriteMode
			      , full_mask
			      , KesOffsetOf (TestRec, s)
			      , KeCodecSTRINGGet() 
			      , KesStructStringSet
			      , KesStructStringGet
			      , KesStructNullSlotExec) ;

    KesStructAccessorDeclare (TestClass
			      , "clone"
			      , KEDIT_ExecMode|KEDIT_WriteMode|KEDIT_ReadMode
			      , full_mask | KEDIT_ExecMask
			      , KesOffsetOf (TestRec, clone)
			      , KeCodecSTRINGGet()
			      , KesStructStringSet
			      , KesStructStringGet
			      , TestClone) ;

/*
 * Create a Struct directory for our structs, and add it as a child of the
 * server's root directory.
 */

    StructDirectory = KesDirectoryMake (Server, "Struct") ;

    KesDirectoryAddChild (KesRootDirectoryOfServer (Server)
			  , StructDirectory->node) ;

/*
 * Create and export sample instances :
 */

    for (i = 0; i < num_objs ; i++ )
    {
	KesStruct object ;
	Test      test = (Test) malloc (sizeof (TestRec)) ;
	char      name[128] ;

	test->b     = 0 ;
	test->c     = 0 ;
	test->s     = (char *) 0 ;
	test->clone = (char *) strdup ("Hello World") ;

	sprintf (name, "%s-%d", "test", i) ;
	object = KesStructMake (Server
				, TestClass
				, name
				, (KePointer) test
				, KEDIT_CreateMask | KEDIT_DeleteMask) ;

	fprintf (stderr, "structsample: object x%08x node %08x %d\n"
		 , object
		 , object->node
		 , object->node) ;

	KesDirectoryAddChild (StructDirectory, object->node) ;
    }

    fprintf (stderr, "structsample[main]: inited %d objects\n", num_objs) ;

    IceMainLoop () ;
    

}
