/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * kedit-tree.c
 *
 * Maintains the k-edit structures for the tree.
 ***************************************************************************/

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Cardinals.h>	

#include "xkedit.h"

static XTree TheXTree = (XTree) 0 ;

/***************************************************************************
 * MakeNode: add a node to the graph
 *
 * Convenient function.
 *
 * RETURNS: created node
 ***************************************************************************/

XTreeNode MakeNode (kec_conn, parent, stat)
    KecConnection kec_conn ;
    XTreeNode     parent ;
    KeNodeStat    stat ;

{
    NodeInfo info = (NodeInfo) malloc (sizeof (NodeInfoRec)) ;
    void _KeditDeleteHandler() ;
    
    if ( ! info )
    {
	fprintf (stderr, "malloc failed.\n") ;
	exit (1) ;
    }

    info->xnode = XTreeCreateNode (parent
				   , stat->name
				   , ((stat->mode & KEDIT_DirectoryMode) 
				      ? ( XTREE_NODE_DIRECTORY
					 | XTREE_NODE_DIRECTORY_LAZY
					 | XTREE_NODE_FOLDED)
				      : (unsigned long) 0)
				   , (void *) info) ;

    info->flags   = (unsigned long) 0 ;
    info->node    = stat->id ;
    info->ncreate = (KecNotifyId) 0 ;
    info->ndelete = (KecNotifyId) 0 ;
    
    if ( parent ) {
	NodeInfo parinfo = (NodeInfo) XTreeNodeData (parent) ;
	
	info->parent = parinfo->node ;
	if ( parinfo->flags & NI_DELETE_NOTIFY ) {
	    info->ndelete = KecNotifyAdd (kec_conn
					  , info->node
					  , KEDIT_DeleteMask
					  , _KeditDeleteHandler
					  , (void *) info) ;
	    KecConnectionDoSync (kec_conn) ;
	}
    } else {
	info->parent = (KID) -1 ;
    }

    return (info->xnode) ;
}

/***************************************************************************
 * _KeditFiller: filler for the k-edit tree
 * _KeditKiller: killer for a node
 * _KeditAction: action for node (on dble click)
 ***************************************************************************/

static void _KeditKiller (xtree, node, closure)
    XTree     xtree ;
    XTreeNode node ;
    void      *closure ;

{
    Context  ctxt = (Context) closure ;
    NodeInfo info = (NodeInfo) XTreeNodeData (node) ;

    if ( info->ncreate )
	KecNotifyRemove (ctxt->kec_conn, info->ncreate) ;
    if ( info->ndelete )
	KecNotifyRemove (ctxt->kec_conn, info->ndelete) ;

    free (info) ;
    return ;
}

static void _KeditFiller (xtree, node, closure)
    XTree     xtree ;
    XTreeNode node ;
    void      *closure ;

{
    KeNodeStat *stats ;
    Context    ctxt = (Context) closure ;
    int        i, n ;
    NodeInfo   new, info = (NodeInfo) XTreeNodeData (node) ;

    if ( (n = KecLsl (ctxt->kec_conn, info->node, &stats)) < 0)
    {
	SetUserMessage (ctxt, "KecLsl failed on node.") ;
	return ;
    }

    for (i = 0 ; i < n ; i++ )
    {
	(void) MakeNode (ctxt->kec_conn, node, stats[i]) ;
    }

    KecLslFree (stats) ;
    
    return ;
}

static void _KeditAction (xtree, node, ctxt_ptr)
    XTree     xtree ;
    XTreeNode node ;
    void      *ctxt_ptr ;

{
    Context    ctxt = (Context) ctxt_ptr ;
    Inspector inspector ;

    if ( inspector = MakeInspector (ctxt) )
    {
	MakeViewer (ctxt, node) ;
	InspectorResize (inspector) ;
    }

    return ;
}

/***************************************************************************
 * _KeditDeleteHandler: handle the deletion of some node
 * _KeditCreateHandler: handle the creation of some node
 * Automatically add new created children
 ***************************************************************************/

void _KeditDeleteHandler (kec_conn, event, closure)
    KecConnection kec_conn ;
    KecEvent      event ;
    KePointer     closure ;

{
    KecDeleteEvent delete = (KecDeleteEvent) event ;
    NodeInfo       info = (NodeInfo) closure ;

    XTreeDeleteNode (TheXTree, info->xnode) ;
    return ;
}
    
    
void _KeditCreateHandler (kec_conn, event, closure)
    KecConnection kec_conn ;
    KecEvent      event ;
    KePointer     closure ;

{
    KecCreateEvent create = (KecCreateEvent) event ;
    NodeInfo       new, info   = (NodeInfo) closure ;
    KeNodeStat     stat_ret ;
    KeStatus       status ;

/*
 * If father node is folded, unfolding it is sufficient:
 */

    if ( XTREE_FOLDED (info->xnode) )
    {
	XTreeFoldUnfoldNode (TheXTree, info->xnode) ;
	XTreeRefresh (TheXTree, info->xnode) ;
	return ;
    }

/*
 * Else, stat on node, and add the new child:
 */
    
    if ( status = KecStatSync (kec_conn, create->child, &stat_ret) )
    {
	fprintf (stderr, "Stat failed\n") ;
	return ;
    }

    (void) MakeNode (kec_conn, info->xnode, stat_ret) ; 

    XTreeRefresh (TheXTree, info->xnode) ;

    return ;
}

/***************************************************************************
 * ConnectToServer: connect me to some server
 *
 * RETURNS: given context
 ***************************************************************************/

static void _KeditProcessClientMessage (kec_conn_ptr, pfd, input)
    XtPointer kec_conn_ptr ;
    int       *pfd ;
    XtInputId input ;

{
    IceProcessMessages (((KecConnection) kec_conn_ptr)->ice_connection
			, NULL
			, NULL) ;
    return ;
}

Context ConnectToServer (ctxt, name)
    Context ctxt ;
    char    *name ;

{
    XTreeNode  root ;
    KeNodeStat stat_ret ;
    NodeInfo   info ;

/*
 * Free previous stuff:
 */

    if ( ctxt->kec_conn )
    {
	/* SHOULD REMOVE INPUT here */
	KecCloseConnection (ctxt->kec_conn) ;
    }

    if ( ctxt->xtree )
	XTreeDelete (ctxt->xtree) ;

/*
 * Create new stuff:
 */

    if ( ! (ctxt->kec_conn = KecOpenConnection (name)) )
    {
	SetUserMessage (ctxt, "Unable to connect to target server !") ;
	return (ctxt) ;
    }

    ctxt->xt_input = XtAppAddInput (ctxt->app_con
				    , KecFdOfConnection (ctxt->kec_conn)
				    , (XtPointer) XtInputReadMask
				    , _KeditProcessClientMessage
				    , (XtPointer) ctxt->kec_conn) ;

    if ( KecStatSync (ctxt->kec_conn
		      , KecRootNodeOfConnection (ctxt->kec_conn)
		      , &stat_ret) )
    {
	SetUserMessage (ctxt, "Stat on root node failed") ;
	return (ctxt) ;
    }

    if ( ! (info = (NodeInfo) malloc (sizeof (NodeInfoRec))) )
    {
	fprintf (stderr, "malloc failed.\n") ;
	exit (1) ;
    }

    root = MakeNode (ctxt->kec_conn, (XTreeNode) 0, stat_ret) ;

    KeFree (stat_ret) ;

    TheXTree = ctxt->xtree = XTreeCreate (ctxt->wpaned
					  , root
					  , _KeditFiller
					  , (void *) ctxt
					  , _KeditKiller
					  , (void *) ctxt
					  , _KeditAction
					  , (void *) ctxt) ;

    return (ctxt) ;
}
