/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * xaw-inspect.c
 *
 * Create and manage a node editor.
 * Node editors are arranged inside inspectors, they can be multiple
 * inspectors, allowing users to visualize mutliple node in the same
 * "box".
 ***************************************************************************/

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Cardinals.h>	

#include <X11/Shell.h>

#include <X11/Xaw/Box.h>	
#include <X11/Xaw/Command.h>
#include <X11/Xaw/MenuButton.h>	
#include <X11/Xaw/SmeBSB.h>	
#include <X11/Xaw/SimpleMenu.h>	
#include <X11/Xaw/Paned.h>	
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Toggle.h>

#include "xkedit.h"

Inspector TheInspector = (Inspector) 0 ;

/*
 * Prototypes:
 */

static Viewer _CreateViewerTopArea (
#ifdef KEDIT_NeedsFunctionPrototype
      Viewer viewer
    , Widget wtop
#endif
) ;

static void _ViewerSetActionProc (
#ifdef KEDIT_NeedsFunctionPrototype
      Widget widget
    , XEvent *event
    , String *params
    , Cardinal *num_params
#endif
) ;

static Viewer _ViewerByTextWidget (
#ifdef KEDIT_NeedsFunctionPrototype
      Widget wtext 
#endif
) ;

/*
 * Text widget translations:
 */

static char *ViewerTextTranslation = "<Key>Return: viewer-set()" ;

static XtActionsRec ViewerTextActions[] = {
    {
	"viewer-set",
	(XtActionProc) _ViewerSetActionProc
    }
} ;

/***************************************************************************
 * InspectorResize: resize inspector so that all viewers are visble
 *
 * This is really BAD, I'm ashame for it. I cannot find any container
 * that need my needs for inspectors, as I don't want to use scrollbars.
 * The cost for this is thoses HARD-CODED values here (I could recompute
 * the full inspector width, but why bother ?
 *
 * RETURNS: void
 ***************************************************************************/

void InspectorResize (inspector)
    Inspector inspector ;

{
    Viewer    viewer ;
    Dimension iw, w, ih, h ;
    int       i ;
    
    for ( iw = ih = 0, i = 0 ; i < inspector->viewers_count ; i++ )
    {
	XtVaGetValues (inspector->viewers[i]->wtop
		       , XtNwidth, &w
		       , XtNheight, &h
		       , NULL) ;
	
	ih += h ;
	if ( iw < w )
	    iw = w ;
    }
    
    XtVaSetValues (inspector->wtop
		   , XtNwidth, (iw > 0) ? (iw + 10) : 400
		   , XtNheight, ih + 70
		   , NULL) ;

    return ;
}

/***************************************************************************
 * DeleteInspector: delete inspector, free all associated resources
 *
 * RETURN: void
 ***************************************************************************/

void DeleteInspector (inspector)
    Inspector inspector ;

{
    Inspector *prev, me ;

/*
 * Delete all viewers first:
 */

    while ( inspector->viewers_count )
	DeleteViewer (inspector->viewers[0], False, False) ;

    if ( inspector->viewers )
	free (inspector->viewers) ;
    
/*
 * Remove from list, destroy widget hierarchy, and free inspector itself:
 */

    me   = TheInspector ;
    prev = &TheInspector ;

    while ( me )
    {
	if ( me == inspector )
	{
	    *prev = me->next ;
	    break ;
	}
	prev = &(me->next) ;
	me   = me->next ;
    }
	
    XtDestroyWidget (inspector->wtop) ;
    
    free (inspector) ;
    
    return ;
}

/***************************************************************************
 * MakeInspector: make an inspector
 *
 * _CreateInspectorTopArea: create the button bar for inspector
 * _InspectorViewerIndex  : index of viewer in inspector
 ***************************************************************************/

static int _InspectorViewerIndex (viewer)
    Viewer viewer ;

{
    register int idx ;
    Inspector    inspector = viewer->inspector ;

    for ( idx = 0 ; idx < inspector->viewers_count ; idx++ )
    {
	if ( inspector->viewers[idx] == viewer )
	    return (idx) ;
    }

    fprintf (stderr, "_InspectorViewerIndex: implementation bug.\n") ;
    exit (1) ;
}

static void _InspectorDeleteCallback (widget, inspector_ptr, unused)
    Widget    widget ;
    XtPointer inspector_ptr ;
    XtPointer unused ;

{
    DeleteInspector ((Inspector) inspector_ptr) ;
    return ;
}

static void _InspectorHelpCallback (widget, ctxt_ptr, unused)
    Widget    widget ;
    XtPointer ctxt_ptr ;
    XtPointer unused ;

{
    DisplayFile ((Context) ctxt_ptr, INSPECTOR_FILE) ;
    return ;
}

static Inspector _CreateInspectorTopArea (inspector, parent)
    Inspector inspector ;
    Widget    parent ;

{
    Widget box, command ;

    box = XtCreateManagedWidget ("menubar"
				 , boxWidgetClass
				 , parent
				 , NULL, ZERO) ;
    
    command = XtCreateManagedWidget ("help"
				     , commandWidgetClass
				     , box
				     , NULL, ZERO) ;
    
    XtAddCallback (command, XtNcallback
		   , _InspectorHelpCallback
		   , (XtPointer) inspector->context) ;

    command = XtCreateManagedWidget ("delete"
				     , commandWidgetClass
				     , box
				     , NULL, ZERO) ;

    XtAddCallback (command
		   , XtNcallback
		   , _InspectorDeleteCallback
		   , (XtPointer) inspector) ;

    return (inspector) ;
}

Inspector MakeInspector (ctxt)
    Context ctxt ;

{
    Inspector inspector ;
    Widget    paned ;

/*
 * Allocate and set defaults:
 */

    if ( ! (inspector = (Inspector) malloc (sizeof (InspectorRec))) )
	return (inspector) ;

    inspector->next          = (Inspector) 0 ;

    inspector->viewers       = (Viewer *) 0 ;
    inspector->viewers_size  = 0 ;
    inspector->viewers_count = 0 ;

    inspector->context       = ctxt ;

/*
 * Create widget hierarchy:
 */

    inspector->wtop = XtVaCreatePopupShell ("inspector"
					    , topLevelShellWidgetClass
					    , ctxt->wtop
					    , XtNallowShellResize, TRUE
					    , XtNtransientFor, ctxt->wtop
					    , NULL) ;

    paned = XtCreateManagedWidget ("inspector-box"
				   , panedWidgetClass
				   , inspector->wtop
				   , NULL, ZERO) ;

    _CreateInspectorTopArea (inspector, paned) ;

    inspector->wbox = XtCreateManagedWidget ("viewers-box"
					     , boxWidgetClass
					     , paned
					     , NULL, ZERO) ;

    XtRealizeWidget (inspector->wtop) ;

    XtPopup (inspector->wtop, XtGrabNone) ;

/*
 * Update static globals:
 */

    if ( ! TheInspector )
    {
	TheInspector = inspector ;
	XtAppAddActions (ctxt->app_con
			 , ViewerTextActions
			 , XtNumber(ViewerTextActions)) ;
    }
    else
    {
	inspector->next = TheInspector ;
	TheInspector    = inspector ;
    }

    return (inspector) ;
}

/***************************************************************************
 * Various converters from string to value:
 * _CARD16FromString
 * _CARD32FromString
 * _BYTEFromString
 * _STRINGFromString
 ***************************************************************************/

static void *_BYTEFromString (viewer)
    Viewer viewer ;

{
    static BYTE byte ;
    char        *str_value ;

    XtVaGetValues (viewer->wtext, XtNstring, &str_value, NULL) ;
    
    byte = (BYTE) atoi (str_value) ;
    return ((void *) &byte) ;
}

static void _BYTEToString (viewer, pbyte)
    Viewer viewer ;
    BYTE   *pbyte ;

{
    char buffer[32] ;

    sprintf (buffer, "%d", (int) *pbyte) ;
    XtVaSetValues (viewer->wtext, XtNstring, buffer, NULL) ;
    return ;
}

static void *_CARD16FromString (viewer)
    Viewer viewer ;

{
    static CARD16 card16 ;
    char          *str_value ;

    XtVaGetValues (viewer->wtext, XtNstring, &str_value, NULL) ;
    
    card16 = (CARD16) atoi (str_value) ;
    return ((void *) &card16) ;
}

static void _CARD16ToString (viewer, pcard16)
    Viewer viewer ;
    CARD16 *pcard16 ;

{
    char buffer[32] ;

    sprintf (buffer, "%d", (int) *pcard16) ;
    XtVaSetValues (viewer->wtext, XtNstring, buffer, NULL) ;
    return ;
}

static void *_CARD32FromString (viewer)
    Viewer viewer ;

{
    static CARD32 card32 ;
    char          *str_value ;

    XtVaGetValues (viewer->wtext, XtNstring, &str_value, NULL) ;
    
    card32 = (CARD32) atoi (str_value) ;
    return ((void *) &card32) ;
}

static void _CARD32ToString (viewer, pcard32)
    Viewer viewer ;
    CARD32 *pcard32 ;

{
    char buffer[32] ;

    sprintf (buffer, "%d", (int) *pcard32) ;
    XtVaSetValues (viewer->wtext, XtNstring, buffer, NULL) ;
    return ;
}

static void *_STRINGFromString (viewer)
    Viewer viewer ;

{
    static char *string ;

    XtVaGetValues (viewer->wtext, XtNstring, &string, NULL) ;
    
    return ((void *) string) ;
}

static void _STRINGToString (viewer, pstr)
    Viewer viewer ;
    char   *pstr ;

{
    XtVaSetValues (viewer->wtext, XtNstring, pstr, NULL) ;
    return ;
}

/***************************************************************************
 * _ViewerByTextWidget: retreive the viewer wich has this text widget
 *
 * This is needed for those @&#$ ActionProc with no closures
 ***************************************************************************/

static Viewer _ViewerByTextWidget (widget)
    Widget widget ;

{
    Inspector inspector ;
    int       i ;

    for ( inspector = TheInspector ; inspector ; inspector = inspector->next )
    {
	for ( i = 0 ; i < inspector->viewers_count ; i++ )
	{
	    if ( inspector->viewers[i]->wtext == widget )
		return (inspector->viewers[i]) ;
	}
    }

    return ((Viewer) 0) ;
}

/***************************************************************************
 * MakeViewer: make a viewer for given node
 *
 * _CreateViewerTopArea : create the viewer menubar
 * _ViewerDeleteCallback: delete this viewer callback
 * _ViewerSetCallback : apply current value callback
 * _ViewerGetCallback : get current value
 ***************************************************************************/

static void _ViewerHelpCallback (widget, ctxt_ptr, unused)
    Widget    widget ;
    XtPointer ctxt_ptr ;
    XtPointer unused ;
    
{
    DisplayFile ((Context) ctxt_ptr, VIEWER_FILE) ;
    return ;
}

static void _ViewerDeleteCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;
    
{
    DeleteViewer ((Viewer) viewer_ptr, True, True) ;
    return ;
}

static void _ViewerSetCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerSetValue ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerGetCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerGetValue ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerExecCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerExec ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerReadWithValueNotifyCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerReadWithValueMonitor ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerWriteWithValueNotifyCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerWriteWithValueMonitor ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerReadNotifyCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerReadMonitor ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerWriteNotifyCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerWriteMonitor ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerExecNotifyCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerExecMonitor ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerNotifyNotifyCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    ViewerNotifyMonitor ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerEditorCallback (widget, viewer_ptr, unused)
    Widget    widget ;
    XtPointer viewer_ptr ;
    XtPointer unused ;

{
    RunEditor ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerSetActionProc (widget, event, params, num_params)
    Widget   widget ;
    XEvent   *event ;
    String   *params ;
    Cardinal *num_params ;

{
    Viewer viewer = _ViewerByTextWidget (widget) ;

    if ( viewer )
	ViewerSetValue (viewer) ;

    return ;
}


static Viewer _CreateViewerTopArea (viewer, parent)
    Viewer viewer ;
    Widget parent ;

{
    Widget box, button, node_menu, edit_menu, entry;

    box = XtVaCreateManagedWidget ("menubar"
				   , boxWidgetClass
				   , parent
				   , NULL, ZERO) ;

/*
 * The node name:
 */

    XtVaCreateManagedWidget ("title"
			     , labelWidgetClass
			     , box
			     , XtNlabel, viewer->stat->name
			     , NULL, ZERO) ;

/*
 * The edit menu:
 */

    button = XtCreateManagedWidget ("edit"
				    , menuButtonWidgetClass
				    , box
				    , NULL, ZERO) ;

    edit_menu = XtCreatePopupShell ("menu"
				    , simpleMenuWidgetClass
				    , button
				    , NULL, ZERO) ;
    

    entry = XtCreateManagedWidget ("editor"
				   , smeBSBObjectClass
				   , edit_menu
				   , NULL, ZERO) ;
    
    XtAddCallback (entry, XtNcallback
		   , _ViewerEditorCallback, (XtPointer) viewer) ;

    entry = XtCreateManagedWidget ("help"
				   , smeBSBObjectClass
				   , edit_menu
				   , NULL, ZERO) ;

    XtAddCallback (entry, XtNcallback
		   , _ViewerHelpCallback
		   , (XtPointer) viewer->inspector->context) ;

    entry = XtCreateManagedWidget ("delete"
				   , smeBSBObjectClass
				   , edit_menu
				   , NULL, ZERO) ;

    XtAddCallback (entry, XtNcallback
		   , _ViewerDeleteCallback, (XtPointer) viewer) ;


/*
 * The node menu is computed according to the node capacities
 */

    if (viewer->stat->mode & ( KEDIT_ExecMode|KEDIT_WriteMode|KEDIT_ReadMode) )
    {
	button = XtCreateManagedWidget ("node"
					, menuButtonWidgetClass
					, box
					, NULL, ZERO) ;
	
	node_menu = XtCreatePopupShell ("menu"
					, simpleMenuWidgetClass
					, button
					, NULL, ZERO) ;
	
	if ( viewer->stat->mode & KEDIT_ExecMode )
	{
	    entry = XtCreateManagedWidget ("exec"
					   , smeBSBObjectClass
					   , node_menu
					   , NULL, ZERO) ;
	    
	XtAddCallback (entry, XtNcallback
		       , _ViewerExecCallback, (XtPointer) viewer);
	}
	
	if ( viewer->stat->mode & KEDIT_WriteMode )
	{
	    entry = XtCreateManagedWidget ("set"
					   , smeBSBObjectClass
					   , node_menu
					   , NULL, ZERO) ;
	    
	    XtAddCallback (entry, XtNcallback
		       , _ViewerSetCallback, (XtPointer) viewer);
	}
	
	if ( viewer->stat->mode & KEDIT_ReadMode )
	{
	    entry = XtCreateManagedWidget ("get"
					   , smeBSBObjectClass
					   , node_menu
					   , NULL, ZERO) ;
	    
	    XtAddCallback (entry, XtNcallback
			   , _ViewerGetCallback, (XtPointer) viewer);
	}
    }


/*
 * Monitor toggle buttons, according to node status:
 */

    XtCreateManagedWidget ("monitor"
			   , labelWidgetClass
			   , box
			   , NULL, ZERO) ;

    if ( viewer->stat->mask & KEDIT_ReadMask )
    {
	entry = XtCreateManagedWidget ("read-notify"
				       , toggleWidgetClass
				       , box
				       , NULL, ZERO) ;

	XtAddCallback (entry, XtNcallback
		       , _ViewerReadNotifyCallback
		       , (XtPointer) viewer) ;
    }

    if ( viewer->stat->mask & KEDIT_WriteMask )
    {
	entry = XtCreateManagedWidget ("write-notify"
				       , toggleWidgetClass
				       , box
				       , NULL, ZERO) ;
	
	XtAddCallback (entry, XtNcallback
		       , _ViewerWriteNotifyCallback
		       , (XtPointer) viewer) ;
    }

    if ( viewer->stat->mask & KEDIT_ExecMask )
    {
	entry = XtCreateManagedWidget ("exec-notify"
				       , toggleWidgetClass
				       , box
				       , NULL, ZERO) ;

	XtAddCallback (entry, XtNcallback
		       , _ViewerExecNotifyCallback
		       , (XtPointer) viewer) ;
    }

    if ( viewer->stat->mask & KEDIT_NotifyMask )
    {
	entry = XtCreateManagedWidget ("notify-notify"
				       , toggleWidgetClass
				       , box
				       , NULL, ZERO) ;

	XtAddCallback (entry, XtNcallback
		       , _ViewerNotifyNotifyCallback
		       , (XtPointer) viewer) ;
    }

    XtCreateManagedWidget ("update"
			   , labelWidgetClass
			   , box
			   , NULL, ZERO) ;


    if ( viewer->stat->mask & KEDIT_ReadWithValueMask )
    {
	entry = XtCreateManagedWidget ("read-with-value-notify"
				       , toggleWidgetClass
				       , box
				       , NULL, ZERO) ;

	XtAddCallback (entry, XtNcallback
		       , _ViewerReadWithValueNotifyCallback
		       , (XtPointer) viewer) ;
    }

    if ( viewer->stat->mask & KEDIT_WriteWithValueMask )
    {
	entry = XtCreateManagedWidget ("write-with-value-notify"
				       , toggleWidgetClass
				       , box
				       , NULL, ZERO) ;

	XtAddCallback (entry, XtNcallback
		       , _ViewerWriteWithValueNotifyCallback
		       , (XtPointer) viewer) ;
    }

    return (viewer) ;
}


Viewer MakeViewer (ctxt, node)
    Context   ctxt ;
    XTreeNode node ;

{
    Viewer    viewer ;
    Inspector inspector = TheInspector ;
    KeStatus  status ;
    NodeInfo  info = (NodeInfo) XTreeNodeData (node) ;

/*
 * Allocate, set defaults:
 */

    if ( ! ( viewer = (Viewer) malloc (sizeof (ViewerRec))) )
	return (viewer) ;

    viewer->parent_node      = info->parent ;

    viewer->inspector        = inspector ;
    viewer->timeout          = (XtIntervalId) 0 ;
    viewer->flash_count      = 0 ;
    viewer->background       = (Pixel) 0 ;

    viewer->read             = (KecNotifyId) 0 ;
    viewer->read_with_value  = (KecNotifyId) 0 ;
    viewer->write            = (KecNotifyId) 0 ;
    viewer->write_with_value = (KecNotifyId) 0 ;
    viewer->exec             = (KecNotifyId) 0 ;
    viewer->notify           = (KecNotifyId) 0 ;

    if ( status = KecStatSync (ctxt->kec_conn
			       , info->node
			       , &(viewer->stat)) )
    {
	DisplayKeditError (ctxt, "Stat on node:", status) ;
	free (viewer) ;
	return ;
    }

    viewer->codec      = GetCodec (ctxt, viewer->stat->info) ;
    viewer->fromstring = ((void *(*)()) 0) ;

    if ( viewer->codec == KeCodecBYTEGet () ) {
	viewer->tostring   = _BYTEToString ;
	viewer->fromstring = _BYTEFromString ;
    } else if ( viewer->codec == KeCodecCARD16Get () ) {
	viewer->tostring   = _CARD16ToString ;
	viewer->fromstring = _CARD16FromString ;
    } else if ( viewer->codec == KeCodecCARD32Get () ) {
	viewer->tostring   = _CARD32ToString ;
	viewer->fromstring = _CARD32FromString ;
    } else if ( viewer->codec == KeCodecSTRINGGet () ) {
	viewer->tostring   = _STRINGToString ;
	viewer->fromstring = _STRINGFromString ;
    }

/*
 * Create Widget hierarchy:
 */

    viewer->wtop = XtVaCreateWidget ("viewer"
				     , boxWidgetClass
				     , inspector->wbox
				     , NULL) ;

    _CreateViewerTopArea (viewer, viewer->wtop) ;

    viewer->wtext = XtVaCreateManagedWidget ("text"
					     , asciiTextWidgetClass
					     , viewer->wtop
					     , XtNeditType, XawtextEdit
					     , NULL) ;
    XtOverrideTranslations (viewer->wtext
			    , XtParseTranslationTable (ViewerTextTranslation));
	
    if ( ! ( viewer->stat->mode & KEDIT_WriteMode ) )
    {
	XtVaSetValues (viewer->wtext
		       , XtNsensitive, False
		       , NULL) ;
    }

    XtManageChild (viewer->wtop) ;

/*
 * Add it to current inspector:
 */

    if ( ! inspector->viewers_size )
    {
	int size = DEFAULT_VIEWERS_SIZE ;

	inspector->viewers = (Viewer *) malloc (sizeof (Viewer) * size) ;
	inspector->viewers_size = size ;
    }
    else if ( inspector->viewers_count + 1 >= inspector->viewers_size )
    {
	int size = inspector->viewers_size + DEFAULT_VIEWERS_INC ;

	inspector->viewers = (Viewer *) realloc (inspector->viewers 
						  , sizeof (Viewer) * size) ;
	inspector->viewers_size = size ;
    }

    inspector->viewers[inspector->viewers_count++] = viewer ;

/*
 * Get current background, if someone make the viewer flash:
 */

    XtVaGetValues (viewer->wtop
		   , XtNbackground, &(viewer->background)
		   , NULL) ;

/*
 * If node is readable, get current value:
 */

    if ( viewer->stat->mode & KEDIT_ReadMode )
	ViewerGetValue (viewer) ;

    return (viewer) ;
}

/***************************************************************************
 * DeleteViewer: delete given viewer
 *
 * TRy to free all associated resources (it should be okay)
 *
 * RETURN: void
 ***************************************************************************/

void DeleteViewer (viewer, resize, delete)
    Viewer    viewer ;
    int       resize ;
    int       delete ;

{
    Inspector inspector = viewer->inspector ;
    Context   ctxt      = inspector->context ;
    int       idx       = _InspectorViewerIndex (viewer) ;

/*
 * Remove all pending stuff for the viewer:
 */

    if ( viewer->timeout )
	XtRemoveTimeOut (viewer->timeout) ;

    if ( viewer->read )
	KecNotifyRemove (ctxt->kec_conn, viewer->read) ;

    if ( viewer->read_with_value )
	KecNotifyRemove (ctxt->kec_conn, viewer->read_with_value) ;

    if ( viewer->write )
	KecNotifyRemove (ctxt->kec_conn, viewer->write) ;

    if ( viewer->write_with_value )
	KecNotifyRemove (ctxt->kec_conn, viewer->write_with_value) ;

    if ( viewer->exec )
	KecNotifyRemove (ctxt->kec_conn, viewer->exec) ;

    if ( viewer->notify )
	KecNotifyRemove (ctxt->kec_conn, viewer->notify) ;

    KecConnectionDoSync (ctxt->kec_conn) ;

/*
 * Delete and free widget hierarchy:
 */

    XtDestroyWidget (viewer->wtop) ;

/*
 * Set up the inspector:
 */

    if ( idx < inspector->viewers_count - 1 )
    {
	bcopy (inspector->viewers + idx + 1
	       , inspector->viewers + idx
	       , sizeof (Viewer) * (inspector->viewers_count - idx)) ;
    }

    inspector->viewers_count-- ;

    if ( delete && ( ! inspector->viewers_count ) )
	DeleteInspector (inspector) ;

    if ( resize )
	InspectorResize (inspector) ;

/*
 * Free resources:
 */

    KeFree (viewer->stat) ;
    free (viewer) ;

    return ;
}


/***************************************************************************
 * ViewerFlash: make some visual kludge to show viewer
 *
 * _ViewerFlashHandler: is a Kedit notify handler to flah a viewer
 * _ViewerFlasher     : is the timeout proc, to change viewer foreground
 ***************************************************************************/

static void _ViewerFlashHandler (kec_conn, event, viewer_ptr)
    KecConnection kec_conn ;
    KecEvent      event ;
    KePointer     viewer_ptr ;

{
    ViewerFlash ((Viewer) viewer_ptr) ;
    return ;
}

static void _ViewerFlasher (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    viewer->flash_count-- ;

    if ( viewer->flash_count % 2 )
	XtVaSetValues (viewer->wtop
		       , XtNbackground, Defaults.viewer_flash_color
		       , NULL) ;
    else
	XtVaSetValues (viewer->wtop
		       , XtNbackground, viewer->background
		       , NULL) ;

    if ( viewer->flash_count > 0 )
	viewer->timeout = XtAppAddTimeOut (ctxt->app_con
					   , Defaults.viewer_flash_on
					   , _ViewerFlasher
					   , (XtPointer) viewer) ;
    else
	viewer->timeout = (XtIntervalId) 0 ;

    return ;
}

void ViewerFlash (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    viewer->flash_count = Defaults.viewer_flash_count ;

    if ( ! viewer->timeout )
	viewer->timeout = XtAppAddTimeOut (ctxt->app_con
					   , Defaults.viewer_flash_on
					   , _ViewerFlasher
					   , (XtPointer) viewer) ;

    return ;
}
 
/***************************************************************************
 * ViewerSetValue
 * ViewreGetValue
 *
 * Update viewer according to server's value, or server's value according
 * to viewer.
 ***************************************************************************/

void ViewerSetValue (viewer)
    Viewer viewer ;

{
    Context   	   ctxt   = viewer->inspector->context ;
    void      	   *ptr ;
    KePointer 	   value ;
    int       	   size ;
    KeStatus  	   status ;
    char      	   *emsg ;
    KeErrorFreeFct efree ;
/*
 * Go through encoder:
 */

    if ( ! viewer->fromstring ) {
	SetUserMessage (ctxt, "Cannot convert string value to its type.") ;
	return ;
    }

    ptr = (viewer->fromstring)(viewer) ;

    if (status = KeCodecEncode (viewer->codec,ptr,&size,&value,&emsg,&efree)) {
	SetUserMessage (ctxt, emsg) ;
	if ( efree ) 
	    (*efree)(emsg) ;
	return ;
    }

/*
 * Set value:
 */

    KecSet (ctxt->kec_conn, viewer->stat->id, size, value) ;
    KecConnectionDoSync (ctxt->kec_conn) ;

    return ;
}

void ViewerGetValue (viewer)
    Viewer viewer ;

{
    Context   	   ctxt   = viewer->inspector->context ;
    KePointer 	   pointer = (KePointer) 0 ;
    KecValue  	   value ;
    int       	   size ;
    KeStatus  	   status ;
    char      	   *emsg ;
    KeErrorFreeFct efree ;

/*
 * Get and decode:
 */

    if ( status = KecGetSync (ctxt->kec_conn
			      , viewer->stat->id
			      , &value) )
    {
	DisplayKeditError (ctxt, "Get failed: ", status) ;
	return ;
    }

    if ( value->size ) {
	if ( status = KeCodecDecode (viewer->codec
				     , value->size
				     , (KePointer) value->data
				     , KecConnectionSwapped (ctxt->kec_conn)
				     , &pointer
				     , &emsg, &efree) ) {
	    SetUserMessage (ctxt, emsg) ;
	    if ( efree ) 
		(*efree) (emsg) ;
	    return ;
	}

	if ( ! viewer->tostring ) {
	    SetUserMessage (ctxt, "No way to convert value to string.") ;
	    KeCodecDecoderFree (viewer->codec, pointer) ;
	    return ;
	}
	
	(viewer->tostring) (viewer, pointer) ;
	
	KeCodecDecoderFree (viewer->codec, pointer) ;
    }

    KeFree (value) ;
    
    return ;
}

/***************************************************************************
 * ViewerExec: execute viewer's node
 *
 * RETURNS: void
 ***************************************************************************/

void ViewerExec (viewer)
    Viewer viewer ;

{
    Context  ctxt   = viewer->inspector->context ;
    KeStatus status ;

    if ( status = KecExec (ctxt->kec_conn, viewer->stat->id) )
    {
	DisplayKeditError (ctxt, "KecExec failed:", status) ;
	return ;
    }

    KecConnectionDoSync (ctxt->kec_conn) ;
    
    return ;
}

/***************************************************************************
 * ViewerReadWithValueMonitor: notify and update on reads
 *
 * RETURN: void.
 ***************************************************************************/

static void _ViewerReadWithValueHandler (kec_conn, any_event, viewer_ptr)
    KecConnection kec_conn ;
    KecEvent      any_event ;
    KePointer     viewer_ptr ;

{
    KecReadWithValueEvent event = (KecReadWithValueEvent) any_event ;
    Viewer                viewer = (Viewer) viewer_ptr ;
    Context		  ctxt = viewer->inspector->context ;
    KeStatus		  status ;
    KePointer		  pointer = (KePointer) 0 ;
    char		  *emsg ;
    KeErrorFreeFct        efree ;

    ViewerFlash (viewer) ;

/*
 * Get and decode:
 */

    if ( status = KeCodecDecode (viewer->codec
				 , event->size
				 , (KePointer) event->data
				 , KecConnectionSwapped (ctxt->kec_conn)
				 , &pointer
				 , &emsg, &efree) )
    {
	SetUserMessage (ctxt, emsg) ;
	if ( efree )
	    (*efree) (emsg) ;
	return ;
    }

/*
 * Update viewer:
 */

    if ( ! viewer->tostring )
    {
	SetUserMessage (ctxt, "No way to convert value to string.") ;
	KeCodecDecoderFree (viewer->codec, pointer) ;
	return ;
    }

    (viewer->tostring) (viewer, pointer) ;

    KeCodecDecoderFree (viewer->codec, pointer) ;

    return ;
}

void ViewerReadWithValueMonitor (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    if ( viewer->read_with_value )
    {
	KecNotifyRemove (ctxt->kec_conn, viewer->read_with_value) ;
	viewer->read_with_value = (KecNotifyId) 0 ;
    }
    else
    {
	viewer->read_with_value = KecNotifyAdd (ctxt->kec_conn
						, viewer->stat->id
						, KEDIT_ReadWithValueMask
						, _ViewerReadWithValueHandler
						, (KePointer) viewer) ;
    }

    KecConnectionDoSync (ctxt->kec_conn) ;
    return ;
}


/***************************************************************************
 * ViewerWriteWithValueMonitor: update node on writes
 *
 * RETURNS: void.
 ***************************************************************************/

static void _ViewerWriteWithValueHandler (kec_conn, any_event, viewer_ptr)
    KecConnection kec_conn ;
    KecEvent      any_event ;
    KePointer     viewer_ptr ;

{
    KecWriteWithValueEvent event = (KecWriteWithValueEvent) any_event ;
    Viewer                 viewer = (Viewer) viewer_ptr ;
    Context		   ctxt = viewer->inspector->context ;
    KeStatus		   status ;
    KePointer		   pointer = (KePointer) 0 ;
    char		   *emsg ;
    KeErrorFreeFct         efree ;

/*
 * Flash the viewer:
 */

    ViewerFlash (viewer) ;

/*
 * Decode value, convert to string, and free temp:
 */

    if ( status = KeCodecDecode (viewer->codec
				 , event->size
				 , (KePointer) event->data
				 , KecConnectionSwapped (ctxt->kec_conn)
				 , &pointer
				 , &emsg, &efree) )
    {
	SetUserMessage (ctxt, emsg) ;
	if ( efree )
	    (*efree) (emsg) ;
	return ;
    }

    if ( ! viewer->tostring )
    {
	KeCodecDecoderFree (viewer->codec, pointer) ;
	SetUserMessage (ctxt, "No way to convert value to string.") ;
	return ;
    }

    (viewer->tostring) (viewer, pointer) ;
    KeCodecDecoderFree (viewer->codec, pointer) ;

    return ;
}

void ViewerWriteWithValueMonitor (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    if ( viewer->write_with_value )
    {
	KecNotifyRemove (ctxt->kec_conn, viewer->write_with_value) ;
	viewer->write_with_value = (KecNotifyId) 0 ;
    }
    else
    {
	viewer->write_with_value = KecNotifyAdd (ctxt->kec_conn
						 , viewer->stat->id
						 , KEDIT_WriteWithValueMask
						 , _ViewerWriteWithValueHandler
						 , (KePointer) viewer) ;
    }

    KecConnectionDoSync (ctxt->kec_conn) ;
    return ;
}

/***************************************************************************
 * ViewerReadMonitor: monitor on notify
 *
 * RETURNS: void
 ***************************************************************************/

void ViewerReadMonitor (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    if ( viewer->read )
    {
	KecNotifyRemove (ctxt->kec_conn, viewer->read) ;
	viewer->read = (KecNotifyId) 0 ;
    }
    else
    {
	viewer->read = KecNotifyAdd (ctxt->kec_conn
				     , viewer->stat->id
				     , KEDIT_ReadMask
				     , _ViewerFlashHandler
				     , (KePointer) viewer) ;
    }

    KecConnectionDoSync (ctxt->kec_conn) ;
    return ;
}

/***************************************************************************
 * ViewerWriteMonitor: monitor on write
 *
 * RETURNS: void
 ***************************************************************************/

void ViewerWriteMonitor (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    if ( viewer->write )
    {
	KecNotifyRemove (ctxt->kec_conn, viewer->write) ;
	viewer->write = (KecNotifyId) 0 ;
    }
    else
    {
	viewer->write = KecNotifyAdd (ctxt->kec_conn
				      , viewer->stat->id
				      , KEDIT_WriteMask
				      , _ViewerFlashHandler
				      , (KePointer) viewer) ;
    }

    KecConnectionDoSync (ctxt->kec_conn) ;
    return ;
}

/***************************************************************************
 * ViewerExecMonitor: monitor exec for this node
 *
 * RETURNS: void
 ***************************************************************************/

void ViewerExecMonitor (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    if ( viewer->exec )
    {
	KecNotifyRemove (ctxt->kec_conn, viewer->exec) ;
	viewer->exec = (KecNotifyId) 0 ;
    }
    else
    {
	viewer->exec = KecNotifyAdd (ctxt->kec_conn
				     , viewer->stat->id
				     , KEDIT_ExecMask
				     , _ViewerFlashHandler
				     , (KePointer) viewer) ;
    }

    KecConnectionDoSync (ctxt->kec_conn) ;
    return ;
}

/***************************************************************************
 * ViewerNotifyMonitor: monitor on notify
 *
 * RETURNS: void
 ***************************************************************************/

void ViewerNotifyMonitor (viewer)
    Viewer viewer ;

{
    Context ctxt = viewer->inspector->context ;

    if ( viewer->notify )
    {
	KecNotifyRemove (ctxt->kec_conn, viewer->notify) ;
	viewer->notify = (KecNotifyId) 0 ;
    }
    else
    {
	viewer->notify = KecNotifyAdd (ctxt->kec_conn
				       , viewer->stat->id
				       , KEDIT_NotifyMask
				       , _ViewerFlashHandler
				       , (KePointer) viewer) ;
    }

    KecConnectionDoSync (ctxt->kec_conn) ;
    return ;
}

