/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KeClient.h
 *
 * This file should be included by file using the client side library of
 * kedit.
 ***************************************************************************/

#ifndef _KCLIENT_H_
#define _KCLIENT_H_

#include <KeLib.h>

/*
 * We have to predeclare all those data types, because they refer to each 
 * other.
 */

typedef struct _KecConnection *KecConnection ;
typedef union  _KecEvent      *KecEvent ; 
typedef struct _KecError      *KecError ;
typedef struct _KecReply      *KecReply ;
typedef struct _KecPending    *KecPending ;

/*
 * Pending requests, and their handlers.
 *
 * - Pending blocks are opaque data types (programmer should never even 
 * *see* them), although, they are needed because referenced from the 
 * connection data type.
 * - KEC_PENDWATERMARK defines the number of pending block that Kedit client
 * library will keep allocated, even if not used.
 */

#define KEC_PENDWATERMARK	100

typedef void (*KecHandler) (
#ifdef KEDIT_NeedsFunctionPrototype
      KecReply         reply
    , KePointer        closure
#endif
) ;

/*
 * A value is just a variable size struct
 */

typedef struct _KecValue {
    int 		size ;			/* value size */
    char		data[1] ;		/* data */
} KecValueRec, *KecValue ;

#define KecSizeOfValue(v)	((v)->size)
#define KecDataOfValue(v)	((v)->data)

/*
 * Event data type:
 *
 * As event data type depends on event type, the scheme is quite complex.
 * - First of all *every* event begins with the _KecGenericEvent part,
 * - According to events, they than have specific fields.
 * - Finally all events are encapuslated in a union KecEvent.
 * Although some events just need the same kind of fields (ie: size, data for
 * WithValue events, we still define two different types for them (in case
 * those events change in the future).
 */

typedef struct _KecGenericEvent {
    KecConnection	kec_conn ;	/* Connection of this event */
    unsigned int	type ;		/* KEDIT_Event */
    unsigned int	event_type ;	/* See Kproto.h for events type */
    KID			node ;		/* node that is notified */
    unsigned long	timestamp ;	/* server's timestamp */
    unsigned long	client_tag ;	/* client that triggered the event */
} KecDeleteEventRec, *KecDeleteEvent
, KecExecEventRec, *KecExecEvent
, KecReadEventRec, *KecReadEvent
, KecWriteEventRec, *KecWriteEvent 
, KecAnyEventRec, *KecAnyEvent ;

typedef struct _KecNotifyEvent {
    KecConnection	kec_conn ;	/* Connection of this event */
    unsigned int	type ;
    unsigned int	event_type ;
    KID			node ;
    unsigned long	timestamp ;
    unsigned long	client_tag ;
    unsigned long	mask ;		/* new mask for this node */
} KecNotifyEventRec, *KecNotifyEvent ;

typedef struct _KecCreateEvent {
    KecConnection	kec_conn ;	/* Connection of this event */
    unsigned int	type ;
    unsigned int	event_type ;
    KID			node ;
    unsigned long	timestamp ;
    unsigned long	client_tag ;
    KID			child ;		/* father of created node */
} KecCreateEventRec, *KecCreateEvent ;

typedef struct _KecWithValueEvent {
    KecConnection	kec_conn ;	/* Connection of this event */
    unsigned int	type ;
    unsigned int	event_type ;
    KID			node ;
    unsigned long	timestamp ;
    unsigned long	client_tag ;
    unsigned long	size ;		/* size of value */
    char		data[1] ;	/* value bytes */
} KecReadWithValueEventRec, *KecReadWithValueEvent 
, KecWriteWithValueEventRec, *KecWriteWithValueEvent ;

typedef union _KecEvent {
    KecAnyEventRec		any ;
    KecDeleteEventRec		delete_event ;
    KecCreateEventRec		create_event ;
    KecNotifyEventRec 		notify_event ;
    KecExecEventRec		exec_event ;
    KecReadEventRec		read_event ;
    KecWriteEventRec		write_event ;
    KecReadWithValueEventRec	read_with_value_event ;
    KecWriteWithValueEventRec   write_with_value_event ;
} KecEventRec ;

/*
 * Error data type:
 *
 * There is only one possible type for errors, wich is (should be !) 
 * derived from the ICE error datatype.
 */

typedef struct _KecError {
    KecConnection       kec_conn ;
    unsigned int	type ;		/* KEDIT_Error */
    unsigned int	subtype ;	/* subtype */
    KID			node ;		/* the error node */
    unsigned int	serial ;	/* the serial of request */
    KeStatus		status ;	/* any K-edit status */
    char		*msg ;		/* error message */
} KecErrorRec ;

#define KecTypeOfError(e)	((e)->type)
#define KecSubtypeOfError(e)	((e)->subtype)
#define KecNodeOfError(e)	((e)->node)
#define KecSerialOfError(e)	((e)->serial)
#define KecStatusOfError(e)	((e)->status)
#define KecMsgOfError(e)	((e)->msg)

/*
 * Reply data type
 *
 * There is only one reply type, wich have fields for all the specifities
 * of requests.
 *
 * I'm not really convinced with this one: I think it should be better to
 * have one reply type per request type, and have a union for all replies
 * (as is done for events).
 */

typedef struct _KecReply {
    KecConnection         connection ;	/* connnection this reply comes */
    KeStatus		  status ;	/* reply status */
    union {
	KePointer	  pointer ;	/* right pointer to set */
	KecValue	  value ;	/* if a value was returned */
	KeIdArray	  ids ;		/* if a list of ids was returned */
	KeNodeStat	  stat ;	/* if a node stat was returned */
	KecError  	  error ;	/* error on status != KE_SUCCESS */
    } ru ;
} KecReplyRec ;

#define KecConnectionOfReply(r)	((r)->connection)
#define KecStatusOfReply(r)	((r)->status)

#define KecValueOfReply(r)	((r)->ru.value)
#define KecIdArrayOfReply(r)	((r)->ru.ids)
#define KecStatOfReply(r)	((r)->ru.stat)
#define KecErrorOfReply(r)	((r)->ru.error)

/*
 * Handlers data type
 * 
 * There is a full range of hooks and handlers. 
 * - KecSyncHandler: is called after every 'write' to the connection. It is 
 * used to turn the connection in sync mode (by setting this handler to the
 * function that performs the Sync: KecConnectionDoSync).
 * - KecErrorHandler: is called when the connection receives an error message
 * from the server. The default handler prints an ASCII text for error.
 * - KecEventHandler: is called when an event is received on the connection.
 * You should better use the default event handler, wich provide some goodies.
 * - KecIOErrorHandler: is invoked when the connection detects some IO error
 * problems. It should close the connection (at least).
 */

typedef KeStatus (*KecSyncHandler) (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection to be synced */
#endif
) ;
typedef void (*KecErrorHandler) (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection    kec_conn		/* connection on wich error occured */
    , struct _KecError *error		/* decoded event */
#endif 
) ;

typedef void (*KecEventHandler) (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection on wich event occured */
    , KecEvent      event		/* decoded event */
#endif
) ;

typedef void (*KecIOErrorHandler) (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection of IO Error occured */
    , int           errno		/* UNIX errno of IO error */
#endif
) ;

/*
 * Connection data type:
 *
 * This is the main data structure for connection to server.
 */

typedef struct _KecConnection {

    char		*name ;			/* name of target server */
    IceConn		ice_connection ; 	/* ice connection to server */

    int                 disconnected ;		/* is connection valid ? */
    unsigned int	serial ;		/* current K-Edit serial */
    unsigned int 	last_reply ;		/* last reply received */

    KID                 root ;			/* server's root node */
    unsigned long	tag ;			/* client_tag on this conn */

    KeMallocPurgeHook   purge_hook ;
    KeMallocFailHook    fail_hook ;

    int                 (*lock)(), (*unlock)() ;
    KecSyncHandler	sync ; 
    KecIOErrorHandler   io_error ; 
    KecErrorHandler	error_handler ; 
    KecEventHandler 	event_handler ; 

    KecPending		pending ;		/* list of pending requests */
    KecPending		pending_tail ;		/* last pending request */

    KecErrorRec		last_error ;		/* last error recieved */
    KeHashtable		notifications ;		/* notification management */
    KePointer		user_data ; 		/* any user data */

} KecConnectionRec ;

/*
 * Some macros around the KecConnection struct :
 */

#define KCLIENT_LOCK(c)   ((c)->lock ? (*((c)->lock))(c) : 0)
#define KCLIENT_UNLOCK(c) ((c)->unlock ? (*((c)->unlock))(c) : 0)
#define KCLIENT_SYNC(c)   ((c)->sync ? (*((c)->sync))(c) : 0)

#define KecErrorOfConnection(k)	   	(&((k)->last_error))
#define KecFdOfConnection(k)	   	IceConnectionNumber((k)->ice_connection)
#define KecIceConnOfConnection(k)	((k)->ice_connection)
#define KecNameOfConnection(k)	   	((k)->name)
#define KecRootNodeOfConnection(k)	((k)->root)
#define KecTagOfConnection(k)		((k)->tag)
#define KecGetUserDataOfConnection(k) 	((k)->user_data)
#define KecSetUserDataOfConnection(k,v) ((k)->user_data = ((KePointer) v))
#define KecSerialOfConnection(k)	((k)->serial)
#define KecConnectionSwapped(k)		(IceSwapping ((k)->ice_connection))

#define KecMalloc(k,s) \
    KeMalloc ((k)->purge_hook, (k)->fail_hook, (s))

/*
 * KCOpenC.c
 */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

KecConnection KecOpenConnection (
#ifdef KEDIT_NeedsFunctionPrototype
      char *name 			/* name or coord of server */
#endif
) ;

KeStatus KecCloseConnection (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection connection		/* connection to be closed */
#endif
) ;

/*
 * KCSet.c defines:
 */

KeStatus KecSet (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KID           node		/* node to be set */
    , unsigned int  size		/* size of value */
    , KePointer     value		/* pointer to value */
#endif
) ;

/*
 * KCNotify.c defines:
 */

KeStatus KecNotify (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KID           node		/* node on wich to notify */
    , unsigned long mask		/* notification mask */
#endif
) ;

/*
 * KCExec.c defines:
 */

KeStatus KecExec (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KID           node		/* node to be executed */
#endif
) ;

/*
 * KCGet.c defines:
 */

KeStatus _KecGet (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KID           node		/* node to be get'ed */
    , KecValue      *retval		/* retval if synchronous mode */
    , KecHandler    handler		/* handler if async mode */
    , KePointer     closure		/* and handler's closure */
#endif
) ;

#define KecGet(k,i,h,c)	     _KecGet((k),(i),(KecValue *) 0,(h),(c))
#define KecGetSync(kc,id,r)  _KecGet((kc),(id),(r),((KecHandler)0),((KePointer)0))

/*
 * KCLs.c defines:
 */

KeStatus _KecLs (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KID           node		/* node to be listed */
    , KeIdArray     *retval		/* retval, if sync mode */
    , KecHandler    handler		/* handler if async mode */
    , KePointer     closure		/* and handler's closure */
#endif
) ;

#define KecLs(k,i,h,c)	     _KecLs((k),(i),(KeIdArray *) 0,(h),(c))
#define KecLsSync(kc,id,r)   _KecLs((kc),(id),(r),((KecHandler)0),((KePointer)0))

/*
 * KCStat.c defines:
 */

KeStatus _KecStat (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KID           node		/* node to be stat'ed */
    , KeNodeStat    *retval		/* retval if sync mode */
    , KecHandler    handler		/* handler for async mode */
    , KePointer     closure		/* and its closure */
#endif
) ;    

#define KecStat(k,i,h,c)     _KecStat((k),(i),(KeNodeStat *) 0,(h),(c))
#define KecStatSync(kc,id,r) _KecStat((kc),(id),(r),((KecHandler)0),((KePointer)0))

/*
 * KCIo.c defines:
 */

KeStatus KecGobbleData (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection to be 'synced' */
#endif
) ;

/*
 * KCSync.c defines :
 */

KeStatus KecConnectionDoSync (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection to be sync'ed */
#endif
) ;

KeStatus KecConnectionSync (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection to be turned in sync */
    , int           on_off		/* turn sync mode on/off */
#endif
) ;

/*
 * KCEvent.c defines:
 */

void KecEventFree (
#ifdef KEDIT_NeedsFunctionPrototype
      KecEvent event 
#endif
) ;

KecErrorHandler KecSetErrorHandler (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection   kec_conn		/* connection to be setted */
    , KecErrorHandler handler		/* new error protocol handler */
#endif
) ;

KecEventHandler KecSetEventHandler (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection   kec_conn		/* connection to be setted */
    , KecEventHandler handler		/* new event handler for connection */
#endif
) ;

KecSyncHandler KecSetSyncHandler (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection to be setted */
    , KecSyncHandler handler		/* new sync handler for connection */
#endif
) ;

KecIOErrorHandler KecSetIOErrorHandler (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection to be setted */
    , KecIOErrorHandler handler		/* new IO error handler */
#endif
) ;

#ifdef __cplusplus
} /* End of c++ verbosities */
#endif /* __cplusplus */

/* 
 * KCTNotify.c defines :
 */

#define KEC_NOTIFICATION_HASHTABLE_SIZE	101

typedef unsigned long KecNotifyId ;

typedef void (*KecNotifyHandler) (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KecEvent event			/* received event */
    , KePointer closure			/* closure */
#endif
) ;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

KecNotifyId KecNotifyAdd (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KID           node		/* node to be notified on */
    , unsigned long mask		/* notification mask */
    , KecNotifyHandler handler		/* handler to trigger on events */
    , KePointer        closure		/* closure to the handler */
#endif
) ;

KeStatus KecNotifyRemove (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KecNotifyId   handler		/* notification to remove */
#endif
) ;


void KecToolkitEventHandler (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kec_conn		/* connection */
    , KecEvent      event		/* event to be handled */
#endif
) ;

/*
 * KCTLsl.c defines:
 */

int KecLsl (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection connection 		/* connection */
    , KID           node		/* node to list */
    , KeNodeStat    **stats_ret		/* returned array of stats */
#endif
) ;

void KecLslFree (
#ifdef KEDIT_NeedsFunctionPrototype
      KeNodeStat *stats			/* stats to be free */
#endif
) ;

/*
 * KCTFind.c defines;
 */

KID KecFind (
#ifdef KEDIT_NeedsFunctionPrototype
      KecConnection kconn
    , KID           kid
    , char          **path
    , int           npath
#endif
) ;

#define KecFindRoot(k,p,n) \
    KecFind ((k), KecRootNodeOfConnection(k),(p),(n))

#ifdef __cplusplus
} /* End of c++ verbosities */
#endif /* __cplusplus */


#endif /* _KCLIENT_H_ */
