/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * resource.c
 *
 * This file manage the "*resources*" directory for widgets, by combining
 * the real resources and the fake ones into a single directory.
 * Real and Fake resources are also managed here with the help of 
 * - real-resource.c: to access real resources
 * - fake-resource.c: to access fake resources
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <KeServer.h>
#include <KesCodec.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "KesXtP.h"

static KesHandlerSet KesResourceDirectoryHandlerSet = (KesHandlerSet) 0 ;
static KesHandlerSet KesResourceHandlerSet          = (KesHandlerSet) 0 ;

/*
 * Private prototypes:
 */

static KeStatus KesXtResourceInfoFromRequest (
#ifdef KEDIT_NeedsFunctionPrototype
      KesHeader         request
    , KesXtResourceInfo info_ret
#endif
) ;

static int KesXtResourceSize (
#ifdef KEDIT_NeedsFunctionPrototype
      Widget            widget
    , KesXtResourceInfo info
#endif
) ;

/*
 * Error messages emitted:
 */

static char *ErrResourceInfo = "Bad resource ID." ;
static char *ErrNoHandler    = "Operation not supported by resource." ;
static char *ErrEncoder      = "Encoder failed." ;
static char *ErrDecoder      = "Decoder failed." ;

/***************************************************************************
 * KesResourceDirectoryHandlerSetGet
 ***************************************************************************/

static KeStatus KesResourceDirectoryLsHandler (request, reply)
    KesLsReq request ;
    KesReply reply ;

{
    Widget widget = (Widget) request->closure ;
    int    i, count ;
    KID    *pnodes ;
    KID    base ;

    if ( (count = _KesXtWidgetRealResourceCount (widget))  < 0 )
	return (KEDIT_Failure) ;
    
    if ( (i = _KesXtWidgetFakeResourceCount (widget)) < 0 )
	return (KEDIT_Failure) ;
    
    count += i ;

/*
 * Create (or Get) the associated node-range:
 */

    if ( KesNodeRangeMake (request->server
			   , KesResourceHandlerSetGet()
			   , (KePointer) widget
			   , count
			   , &base) != KEDIT_Success )
    {
	return (KEDIT_Failure) ;
    }

/*
 * Setup the reply :
 */

    pnodes = (KID *) KesMalloc (request->server, sizeof (KID) * (count)) ;

    for ( i = 0 ; i < count ; i++ )
	pnodes[i] = base + i ;

    KesLsReplyFill (reply, count, pnodes, KeFree) ;

    return (KEDIT_Success) ;
}

static KeStatus KesResourceDirectoryStatHandler (request, reply) 
    KesStatReq request ;
    KesReply   reply ;

{
    KesStatReplyFillStatic (reply
			    , "*resources*"
			    , KEDIT_DirectoryMode
			    , (unsigned long) 0
			    , (KID) 0
			    , 0) ;
    return (KEDIT_Success) ;
}

KesHandlerSet KesResourceDirectoryHandlerSetGet ()

{
    if ( ( ! KesResourceDirectoryHandlerSet )
	&& ( KesResourceDirectoryHandlerSet = KesHandlerSetMake () ) )
    {
	KesHandlerSetLsHandler (KesResourceDirectoryHandlerSet
				, KesResourceDirectoryLsHandler) ;
	KesHandlerSetStatHandler (KesResourceDirectoryHandlerSet
				  , KesResourceDirectoryStatHandler) ;
    }

    return (KesResourceDirectoryHandlerSet) ;
}

/***************************************************************************
 * KesResourceHandlerSet: manage resource nodes
 *
 * A few utility functions (wich deals with fake and real resources):
 * - KesXtResourceInfoFromRequest:
 *   Retreive the description of the resource from a request.
 * - KesXtResourceSize:
 *   Retreive the size of encoded widget resource value
 ***************************************************************************/

static KeStatus KesXtResourceInfoFromRequest (request, rinfo)
    KesHeader         request ;
    KesXtResourceInfo rinfo ;

{
    Widget               widget = (Widget) request->closure ;
    KID                  base ;
    int                  real_count, fake_count, position ;
    KeStatus             status ;

/*
 * Get the resource count, in order to retreive the base node identifier
 * for this node.
 */

    if ( (real_count = _KesXtWidgetRealResourceCount (widget)) < 0 )
	return (KEDIT_Failure) ;

    if ( (fake_count = _KesXtWidgetFakeResourceCount (widget)) < 0 )
	return (KEDIT_Failure) ;

    if ( KesNodeRangeMake (request->server
			   , KesResourceHandlerSetGet ()
			   , (KePointer) widget
			   , real_count + fake_count
			   , &base) != KEDIT_Success )
    {
	return (KEDIT_Failure) ;
    }

/*
 * Now, we can deduce the position of node in node range, and from this
 * fill in the resource description
 */

    position = KesNodeRangeIndex (((KesLsReq) request)->node, base) ;

    if ( position < real_count )
	status = _KesXtGetOneRealResource (widget, position, rinfo) ;
    else
	status = _KesXtGetOneFakeResource (widget, position-real_count, rinfo);

    return (status) ;
}

static int KesXtResourceSize (widget, rinfo)
    Widget            widget ;
    KesXtResourceInfo rinfo ;

{
    if ( ! rinfo->get )
	return (-1) ;
    return (0) ;
}

static KeStatus KesResourceStatHandler (request, reply)
    KesStatReq request ;
    KesReply   reply ;

{
    Widget               widget = (Widget) request->closure ;
    KID                  codec_node ;
    KesXtResourceInfoRec info_rec ;
    KeStatus             status ;

/*
 * Get theresource description:
 */

    if (status = KesXtResourceInfoFromRequest((KesHeader) request, &info_rec))
	return (status) ;

/*
 * Ok, perform he job (at least, sighhhh)
 * - Register codec, to get its associated node
 * - Fill in the stat reply
 */

    if ( KesCodecRegister (request->server, info_rec.codec, &codec_node) )
	return (KEDIT_Failure) ;

    KesStatReplyFillStatic (reply
			    , info_rec.name
			    , info_rec.mode
			    , info_rec.mask
			    , codec_node
			    , KesXtResourceSize (widget, &info_rec)) ;

    return (KEDIT_Success) ;
}

static KeStatus KesResourceGetHandler (request, reply)
    KesGetReq request ;
    KesReply  reply ;

{
    Widget               widget = (Widget) request->closure ;
    KesXtResourceInfoRec info_rec ;
    KePointer            value ;
    KePointer		 coded_value = (KePointer) 0 ;
    int                  coded_size ;
    KeStatus             status ;
    char		 *emsg ;
    KeErrorFreeFct       efree ;

/*
 * Get the resource description:
 */

    if (status = KesXtResourceInfoFromRequest ((KesHeader)request,&info_rec)) {
	KesErrorReplyFillStatic (reply, status, ErrResourceInfo) ;
	return (KEDIT_Success) ;
    }

/*
 * Perform the job:
 */

    if ( ! info_rec.get ) {
	KesErrorReplyFillStatic (reply, KEDIT_InvalidOp, ErrNoHandler) ;
	return (KEDIT_Success) ;
    }

    if (status = (info_rec.get) (widget,&info_rec,&value,&emsg,&efree)) {
	KesErrorReplyFill (reply, status, emsg, (KesReplyFreeFct) efree) ;
	return (KEDIT_Success) ;
    }

    if ( (status = KeCodecEncode (info_rec.codec
				  , value
				  , &coded_size
				  , &coded_value
				  , &emsg, &efree)) != KEDIT_Success ) {
	KesErrorReplyFill (reply, status, emsg, (KesReplyFreeFct) efree);
	return (KEDIT_Success) ;
    }

    KesGetReplyFill (reply
		     , coded_size
		     , coded_value
		     , (KesReplyFreeFct)KeCodecEncoderFreeer(info_rec.codec));

    return (KEDIT_Success) ;
}

static KeStatus KesResourceSetHandler (request, reply)
    KesSetReq request ;
    KesReply  reply ;

{
    Widget               widget = (Widget) request->closure ;
    KesXtResourceInfoRec info_rec ;
    KeStatus		 status ;
    KePointer		 decoded_value = (KePointer) 0 ;
    char		 *emsg ;
    KeErrorFreeFct       efree ;

/*
 * Retreive the resource description:
 */
    
    if (status = KesXtResourceInfoFromRequest ((KesHeader)request,&info_rec)) {
	KesErrorReplyFillStatic (reply, status, ErrResourceInfo) ;
	return (KEDIT_Success) ;
    }

/*
 * Perform the job:
 */

    if ( ! info_rec.set ) {
	KesErrorReplyFillStatic (reply, KEDIT_InvalidOp, ErrNoHandler) ;
	return (KEDIT_Success) ;
    }

    if ( status = KeCodecDecode (info_rec.codec
				 , request->size
				 , (KePointer) request->value
				 , KesClientSwapped (request->client)
				 , &decoded_value
				 , &emsg, &efree) ) {
	KesErrorReplyFill (reply, status, emsg, (KesReplyFreeFct) efree) ;
	return (KEDIT_Success) ;
    }

    status = (info_rec.set)(widget,&info_rec,decoded_value,&emsg,&efree) ;
    if ( status ) 
	KesErrorReplyFill (reply, status, emsg, (KesReplyFreeFct) efree) ;

/* BUG:
 *    KeCodecDecoderFree (info_rec.codec, decoded_value) ;
 */

    return (KEDIT_Success) ;
}

static KeStatus KesResourceExecHandler (request, reply)
    KesExecReq request ;
    KesReply   reply ;

{
    Widget               widget = (Widget) request->closure ;
    KesXtResourceInfoRec info_rec ;
    KeStatus		 status ;
    char		 *emsg ;
    KeErrorFreeFct       efree ;

/*
 * Retreive the resource description:
 */
    
    if (status = KesXtResourceInfoFromRequest ((KesHeader)request,&info_rec)) {
	KesErrorReplyFillStatic (reply, status, ErrResourceInfo) ;
	return (KEDIT_Success) ;
    }

/*
 * Perform the job:
 */

    if ( ! info_rec.exec ) {
	KesErrorReplyFillStatic (reply, status, ErrNoHandler) ;
	return (KEDIT_Success) ;
    }

    if ( status = (info_rec.exec)(widget, &info_rec, &emsg, &efree) ) 
	KesErrorReplyFill (reply, status, emsg, (KesReplyFreeFct) efree) ;
    return (KEDIT_Success) ;
}

KesHandlerSet KesResourceHandlerSetGet ()

{
    if ( ( ! KesResourceHandlerSet )
	&& ( KesResourceHandlerSet = KesHandlerSetMake() ) )
    {
	KesHandlerSetStatHandler (KesResourceHandlerSet
				  , KesResourceStatHandler) ;
	KesHandlerSetSetHandler  (KesResourceHandlerSet
				  , KesResourceSetHandler) ;
	KesHandlerSetGetHandler  (KesResourceHandlerSet
				  , KesResourceGetHandler) ;
	KesHandlerSetExecHandler (KesResourceHandlerSet
				  , KesResourceExecHandler) ;
    }
    return (KesResourceHandlerSet) ;
}

