/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KCEvent.c
 *
 * Event handling (at the protocol level, not IO).
 * _KecEventDecode: decode an error packet
 * _KecDefaultErrorHandler: the default error handler
 * _KecDefaultEventHandler: the default event handler
 *
 * KecSetErrorHandler: set a new error handler
 * KecSetEventHandler: set a new event handler
 * KecSetIOErrorHandler: set a new error handler
 * KecSetSyncHandler: set a new sync handler
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <errno.h>

extern int errno ;

#include <KeClient.h>

/***************************************************************************
 * _KecEventDecode : decode the given buffer as an error packet
 * KecEventFree    : free a decoded event
 *
 * Decode the provided event into a KecEvent structure. The structure is
 * allocated here, and should be freed using KecEventFree
 *
 * RETURN: a KecEvent (NULL on failure)
 ***************************************************************************/

KeStatus _KecEventDecode (kec_conn, header, body, event_ret, swap)
    KecConnection kec_conn ;
    keditRequest  *header;	/* beginning of event specific infos */
    char          *body ; 	
    KecEvent      *event_ret ;
    Bool          swap ;

{
    unsigned int  size = 0 ;
    unsigned int  event_type ;
    KID           node ;
    unsigned long timestamp, client_tag ;
    KecEvent      pevent ;
    char         *p;
/*
 * Retreive the common event header:
 */
    p=header->data;
    EXTRACT_CARD8 (p, event_type) ;

    /* now, we are in body */
    EXTRACT_CARD32 (body, swap, node) ;
    EXTRACT_CARD32 (body, swap, timestamp) ;
    EXTRACT_CARD32 (body, swap, client_tag) ;
    
    if ( (event_type == KEDIT_ReadWithValueEvent) 
	|| (event_type == KEDIT_WriteWithValueEvent) )
    {
	EXTRACT_CARD32 (body, swap, size) ;
    }
/*
 * Allocate the event:
 */

    pevent = (KecEvent) KecMalloc (kec_conn, sizeof (KecEventRec) + size) ;

    *event_ret = pevent ;

    pevent->any.kec_conn   = kec_conn ;
    pevent->any.event_type = event_type ;
    pevent->any.node       = node ;
    pevent->any.timestamp  = timestamp ;
    pevent->any.client_tag = client_tag ;
    
    switch ( event_type )
    {
      case KEDIT_NotifyEvent:
	EXTRACT_CARD16 (body, swap, pevent->notify_event.mask) ;
	break ;
      case KEDIT_CreateEvent:
	EXTRACT_CARD32 (body, swap, pevent->create_event.child) ;
	break ;
      case KEDIT_ReadWithValueEvent:
	pevent->read_with_value_event.size = size ;
	bcopy (body, pevent->read_with_value_event.data, size) ;
	break ;
      case KEDIT_WriteWithValueEvent:
	pevent->write_with_value_event.size = size ;
	bcopy (body, pevent->write_with_value_event.data, size) ;
	break ;
    }
    return (KEDIT_Success) ;
}

void KecEventFree (event)
    KecEvent event ;

{
    KeFree (event) ;
    return ;
}
    
/***************************************************************************
 * KecErrorDecode: decode the packet as an error.
 ***************************************************************************/

KeStatus _KecErrorDecode (kec_conn, pbuf, error_ret, swap)
    KecConnection kec_conn ;
    char          *pbuf ;	/* at error specific fields */
    KecError      error_ret ;
    Bool          swap ;

{
    int  msg_len ;


    error_ret->kec_conn = kec_conn ;
    error_ret->type     = KEDIT_Error ;
    error_ret->subtype  = 0 ;
    pbuf += sizeof (CARD8) ;				/* skip offedning */
    pbuf += sizeof (CARD8) ;				/* skip severity */
    pbuf += sizeof (CARD16) ;				/* skip unused */
    EXTRACT_CARD32 (pbuf, swap, error_ret->serial) ;
    EXTRACT_CARD32 (pbuf, swap, error_ret->status) ;
    EXTRACT_CARD32 (pbuf, swap, error_ret->node) ;
    EXTRACT_CARD16 (pbuf, swap, msg_len) ;
    
    if ( msg_len ) 
    {
	error_ret->msg = (char *) KecMalloc (kec_conn, msg_len + 1) ;
	bcopy (pbuf, error_ret->msg, msg_len) ;
	error_ret->msg[msg_len] = '\0' ;
    }
    else
    {
	error_ret->msg = (char *) 0 ;
    }

    return (KEDIT_Success) ;
}

/***************************************************************************
 * _KecDefaultErrorHandler: the default client protocol error handler
 *
 * this function is called by KCRep.c when an event is recieved with the
 * type K_Error. Default is to print the error packet to stderr
 * (no exit)
 *
 * RETURN: void.
 ***************************************************************************/

void _KecDefaultErrorHandler (kec, error) 
    KecConnection kec ;
    KecError      error ;

{
    fprintf (stderr, "kEdit: failed request.\n") ;
    fprintf (stderr, "\tserial   : 	%d\n", error->serial) ;
    fprintf (stderr, "\ttype     :	%d\n", error->type) ;
    fprintf (stderr, "\tsubtype  :	%d\n", error->subtype) ;
    fprintf (stderr, "\tnode     :	%d\n", error->node) ;
    fprintf (stderr, "\tstatus   :      %d\n", error->status) ;
    fprintf (stderr, "\tmsg      :      %s\n", ((int)(error->msg)) ? (char *)
	     error->msg : "<NONE>") ;
    fflush (stderr) ;
    return ;
}

/***************************************************************************
 * _KecDefaultEventHandler: the default event handler
 *
 * If DEBUG, print the event, else return
 *
 * RETURN: void.
 ***************************************************************************/
	     
void _KecDefaultEventHandler (kec, event)
    KecConnection kec ;
    KecEvent      event ;

{
    KV ( fprintf (stderr, "_KecDefaultEventHandler: got following event.\n"););
    KV ( fprintf (stderr, "\tsubtype  :	%d\n", event->subtype) ; ) ;
    KV ( fprintf (stderr, "\tnode     :	%d\n", event->node_id) ; ) ;
    KV ( fprintf (stderr, "\ttimestamp:	%d\n", event->timestamp) ; ) ;
    return ;
}

/***************************************************************************
 * _KecDefaultIOErrorHandler: the default error handler
 *
 * This handler should (at least in "serious" applications) always be 
 * redefined. Its default behavior is to quit, because the user must
 * at least remove tyhe corresponding input sources from its mainloop.
 *
 * RETURN: void.
 ***************************************************************************/

void _KecDefaultIOErrorHandler (kec, err)
    KecConnection kec ;
    int err ;

{
    fprintf (stderr, "_KecDefaultIOErrorHandler: closing connection (errno %d)\n", err) ;
    exit (1) ;
}

/***************************************************************************
 * KecSetErrorHandler: set a new error handler
 * KecSetEventHandler: set a new event handler
 * KecSetIOErrorHandler: set a new IO Error handler.
 * KecSetSyncHandler: set a new sync handler
 *
 * RETURN: the old value.
 ***************************************************************************/

KecErrorHandler KecSetErrorHandler (kec, handler)
    KecConnection   kec ;
    KecErrorHandler handler ;

{
    KecErrorHandler retval = kec->error_handler ;

    kec->error_handler = handler ;
    return (retval) ;
}

KecEventHandler KecSetEventHandler (kec, handler)
    KecConnection   kec ;
    KecEventHandler handler ;

{
    KecEventHandler retval = kec->event_handler ;

    kec->event_handler = handler ;
    return (retval) ;
}

KecIOErrorHandler KecSetIOErrorHandler (kec, handler)
    KecConnection     kec ;
    KecIOErrorHandler handler ;

{
    KecIOErrorHandler retval = kec->io_error ;

    kec->io_error = handler ;
    return (retval) ;
}

KecSyncHandler KecSetSyncHandler (kec, handler)
    KecConnection  kec ;
    KecSyncHandler handler ;

{
    KecSyncHandler retval = kec->sync ;

    kec->sync = handler ;
    return (retval) ;
}
