/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KCPend.c
 *
 * Manage the pending blocks for reply, you should also read the file  
 * KCRep.c and one of KCGet.c, KCLs.c ... to see the whole architecture
 * We keep statically allocated a few pending blocks to avoid free/malloc
 * In fact, it would be much better to allocate pending blocks by blocks.
 * 
 * Defined functions:
 * KecPendingMake(): create a pending block
 * KecPendingFree(): free an allocated pending block
 ***************************************************************************/

#include <stdio.h>

#include <KeClient.h>
#include <KeClientP.h>

static KecPending PendingList = (KecPending) 0 ;
static int PendingListLength  = 0 ;

/***************************************************************************
 * KecPendingMake: create a pending bock, and add it to KecConnection
 *
 * The pending list should keep following assertions:
 * . It is ordered in increasing serial (not checkec)
 * . pending_tail points to the last pending block (not chceked)
 * . Only one pending block per serial (not checked)
 *
 * To avoid dead lock on reads, we enforce a call to GobbleData if too 
 * many pending blocks have been created.
 *
 * RETURN: a KecPending or NULL.
 ***************************************************************************/


KecPending _KecPendingMake (kc, serial, decode, handler, closure)
    KecConnection kc ;
    unsigned int  serial ;
    KecDecoder    decode ;
    KecHandler    handler ;
    KePointer     closure ;

{
    KecPending kp ;

/* 
 * try to get an already allocated pending block 
 */

    if ( ! ( kp = PendingList ) )
    {
	kp = KecMalloc (kc, sizeof (KecPendingRec)) ;
    }
    else
    {
	PendingList = kp->next ;
	PendingListLength-- ;
    }

/*
 * Fill in the pending block, and add it to the list:
 */

    kp->serial  = serial ;
    kp->decode  = decode ;
    kp->handler = handler ;
    kp->closure = closure ;
    kp->next    = (KecPending) 0 ;

    if ( kc->pending )
    {
	kc->pending_tail->next = kp ;
	kc->pending_tail       = kp ;
    }
    else
    {
	kc->pending      = kp ;
	kc->pending_tail = kp ;
    }
    return (kp) ;
}

/***************************************************************************
 * KecPendingFree: free the previously allocated pending block
 *
 * Remove this pending block (see assertions in KecPendingMake)
 *
 * RETURN: void
 ***************************************************************************/

void _KecPendingFree (kc, kp)
    KecConnection kc ;
    KecPending    kp ;

{
    KecPending *pkp, kpl ;

/*
 * Find this pending block in the list :
 */
    pkp = &(kc->pending) ;
    for ( kpl = kc->pending ; kpl ; )
    {
	if ( kp == kpl )
	    break ;
	pkp = &(kpl->next) ;
	kpl = kpl->next ;
    }

/*
 * Than free it, and reset the queue :
 */
    if ( kpl )
    {
	*pkp = kp->next ;
	if ( kc->pending_tail == kp )
	{
	    KecPending last = kc->pending ;

/*
 * Nasty hack !!! Poor coding ...
 */

	    if ( ! last )
	    {
		kc->pending_tail = (KecPending) 0 ;
	    }
	    else
	    {
		while ( last->next )
		    last = last->next ;
		kc->pending_tail = last ;
	    }
	}
    }
    else
    {
	fprintf (stderr, "KecPendingFree: couldn't find kp = x%08x.\n", kp) ;
	fprintf (stderr, "implementation -*- BUG -*- (continuing).\n") ;
    }

/*
 * Depending on KEC_PENDWATERMARK, free or keep it :
 */

    if ( PendingListLength > KEC_PENDWATERMARK )
    {
	KeFree (kp) ;
    }
    else
    {
	PendingListLength++ ;
	kp->next    = PendingList ;
	PendingList = kp ;
    }

    return ;
}

