/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KCTNotify.c
 *
 * Toolkit access to notification. This access provides a mean to set a 
 * particular handler for each node.
 ***************************************************************************/


#include <stdio.h>

#include <KeClient.h>

/*
 * This is an opaque structure for notification management:
 */

typedef struct _NotificationHandle {
    KecConnection	kec ;		/* for wich connection */
    KID			kid ;		/* on wich ID */
    unsigned long	mask ;
    KecNotifyHandler	handler ;	/* the handler */
    KePointer		closure ;	/* the closure */
} NotificationHandleRec, *NotificationHandle ;

/***************************************************************************
 * notification_hash_fct
 * notification_cmp_fct 
 *
 * Interface to the hashtable module.
 ***************************************************************************/

static unsigned long notification_hash_fct (handle)
    NotificationHandle handle ;

{
    return ( (((unsigned long) handle->kec) >> 2) ^ ((unsigned long) handle->kid) ) ;
}

static int notification_cmp_fct (handle1, handle2)
    NotificationHandle handle1, handle2 ;

{
    return ( ! ( (handle1->kec == handle2->kec) && (handle1->kid == handle2->kid)) ) ;
}

/***************************************************************************
 * KecNotifyAdd: add a handler for this node events
 *
 * RETURN: a notification ID.
 ***************************************************************************/

KecNotifyId KecNotifyAdd (kec, kid, mask, handler, closure)
    KecConnection    kec ;
    KID	  	     kid ;
    unsigned long    mask ;
    KecNotifyHandler handler ;
    KePointer	     closure ;

{
    KeHashtable             hashtable ;
    NotificationHandleRec notification_rec ;
    NotificationHandle    notification ;

/*
 * Check to see if this connection already has set notification on this node:
 */
    
    if ( ! (hashtable = kec->notifications) )
    {
	kec->notifications = KeHashtableMake (KEC_NOTIFICATION_HASHTABLE_SIZE
					    , notification_hash_fct
					    , notification_cmp_fct) ;
	hashtable = kec->notifications ;
    }

    notification_rec.kec = kec ;
    notification_rec.kid = kid ;
    if ( notification = (NotificationHandle) KeHashtableFind (hashtable, &notification_rec) )
    {
	unsigned long omask = 0 ;

	do
	{
	    omask |= notification->mask ;
	}
	while ( notification = (NotificationHandle) KeHashtableFindNext (hashtable) ) ;
	
	if ( ( omask | mask != mask) && KecNotify (kec, kid, omask | mask) )
	    return ((KecNotifyId) 0) ;
    }
    else if ( mask )
    {
	if ( KecNotify (kec, kid, mask) != KEDIT_Success )
	    return ((KecNotifyId) 0) ;
    }

/*
 * Now, add this new item in the hashtable :
 */

    notification = KecMalloc (kec, sizeof (NotificationHandleRec)) ;

    notification->kec     = kec ;
    notification->kid     = kid ;
    notification->handler = handler ;
    notification->closure = closure ;
    notification->mask    = mask ;
	
    KeHashtableAdd (hashtable, notification) ;

    return ((KecNotifyId) notification) ;
}

/***************************************************************************
 * KecNotifyRemove: remove the given handler
 *
 * If a null handle is given, then remove all the notifications.
 ***************************************************************************/

static void notify_remove_one (handle)
    NotificationHandle handle ;
{
    KeHashtable             hashtable = handle->kec->notifications ;
    NotificationHandleRec notification_rec ; 
    NotificationHandle    notification ;
    unsigned long         nmask = 0 ;

    KeHashtableRemove (hashtable, handle) ;

    notification_rec.kec = handle->kec ;
    notification_rec.kid = handle->kid ;

    if ( notification =  KeHashtableFind (hashtable, &notification_rec) )
    {
	do
	{
	    nmask |= notification->mask ;
	} 
	while ( notification = KeHashtableFindNext (hashtable) ) ;
    }

    KecNotify (handle->kec, ((NotificationHandle) handle)->kid, nmask) ;

    KeFree (handle) ;

    return ;
}

KeStatus KecNotifyRemove (kec, handle)
    KecConnection kec ;
    KecNotifyId   handle ;

{
    KeHashtable             hashtable ;

    if ( ! (hashtable = kec->notifications) )
	return (KEDIT_Failure) ;

/*
 * . Remove it from hashtable, 
 * . If no more notifications on this id for this connection, emit a FALSE notification
 * . Free the notification block.
 */

    if ( handle )
	notify_remove_one (handle) ;
    else
	KeHashtableMap (hashtable, notify_remove_one, (void *) 0) ;

    return (KEDIT_Success) ;
}

/* non documented (yet) function: acces the closure 
   of a notification handler */
KePointer KecClosureOfNotifyHandler (handle)
    NotificationHandle handle ;
{
    return (handle->closure);
}

/***************************************************************************
 * KecToolkitEventHandler: this should be the default event handler
 *
 * This event handler take care to invoke each of the specific node handler
 * on each event it recieves.
 ***************************************************************************/

static void _CallEventHandler (notification_ptr, event_ptr)
    void *notification_ptr ;
    void *event_ptr ;

{
    static int type2mask[] = { 
	0 ,
	KEDIT_DeleteMask ,
	KEDIT_CreateMask ,
	KEDIT_NotifyMask ,
        KEDIT_ExecMask ,
	KEDIT_ReadMask ,
	KEDIT_ReadWithValueMask ,
	KEDIT_WriteMask ,
	KEDIT_WriteWithValueMask 
    } ;
    NotificationHandle notification = (NotificationHandle) notification_ptr ;
    KecEvent           event        = (KecEvent) event_ptr ;

    if ( notification->mask & type2mask[event->any.event_type] )
    {
	(notification->handler) (notification->kec
				 , event
				 , notification->closure) ;
    }

    return ;
}

void KecToolkitEventHandler (kec, event)
    KecConnection kec ;
    KecEvent      event ;

{
    NotificationHandleRec notification_rec ;

/*
 * Simple check , save memory :
 */

    if ( ! kec->notifications )
	return ;

/*
 * Ok, perform high cost :
 */

    notification_rec.kec = kec ;
    notification_rec.kid = event->any.node ;

    KeHashtableKeyedMap (kec->notifications
		       , &notification_rec
		       , _CallEventHandler
		       , event) ;

    KecEventFree (event) ;

    return ;
}
