/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * codecs.c
 *
 * Codecs are used in clients as in servers. That's why they lie here.
 * A codec takes a language depend type, and convert it to a (size,value)
 * couple transmissble over the net. It also performs the operation the
 * other way round (decode).
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <KeLib.h>
#include <KeCodec.h>

static KeCodec KeCodecCARD32 = (KeCodec) 0 ;
static KeCodec KeCodecSTRING = (KeCodec) 0 ;
static KeCodec KeCodecBYTE   = (KeCodec) 0 ;
static KeCodec KeCodecCARD16 = (KeCodec) 0 ;
static KeCodec KeCodecUNDEF  = (KeCodec) 0 ;

static KeHashtable _KeCodecHashtable = (KeHashtable) 0 ;

static char *ErrBadMalloc   = "Malloc failed." ;
static char *ErrInvalidSize = "Invalid data size for this format." ;
static char *ErrNoEncoder   = "No encoder for this format." ;
static char *ErrNoDecoder   = "No decoder for this format." ;

/***************************************************************************
 * codec_hash_fct
 * codec_cmp_fct
 *
 * Hashtable interface.
 ***************************************************************************/

static unsigned long codec_hash_fct (codec)
    KeCodec codec ;

{
    unsigned long hashval = 0, carry ;
    char          *p ;

    for ( p = codec->name ; *p ; p++ )
    {
	if ( carry = (hashval & 0xFF000000) )
	    hashval = (hashval << 8) | ((carry >> 24) ^ ((unsigned long) *p)) ;
	else
	    hashval = (hashval << 8) | ((unsigned long) *p) ;
    }

    return (hashval) ;
}

static int codec_cmp_fct (codec1, codec2)
    KeCodec codec1, codec2 ;

{
    return ( strcmp (codec1->name, codec2->name) ) ;
}


/***************************************************************************
 * KeCodecDefine: define a new basic codec
 * KeCodecGet   : try to get the named codec
 *
 * The codec description is either retreived from the stati hashtable, or
 * created.
 *
 * RETURNS: a codec pointer
 ***************************************************************************/

KeCodec KeCodecGet (name)
    char *name ;

{
    KeCodecRec codec_rec ;

/*
 * Initalize if needed (creating a codec enforce a call to CodecDefine,
 * wich triggers the hastable creation (:-)
 */

    if ( ! _KeCodecHashtable )
	(void) KeCodecBYTEGet () ;

/*
 * Try to find it :
 */

    codec_rec.name = name ;

    return ((KeCodec) KeHashtableFind (_KeCodecHashtable, &codec_rec)) ;
}

KeCodec KeCodecDefine (name, sizer, encoder, enc_free, decoder, dec_free)
    char                *name ;
    KeCodecSizer	sizer ;
    KeCodecEncoder	encoder ;
    KeCodecFreeer	enc_free ;
    KeCodecDecoder	decoder ;
    KeCodecFreeer	dec_free ;

{
    KeCodecRec codec_rec ;
    KeCodec    new_codec ;

/*
 * Initialize the static codec hashtable if not done, and setup default codecs
 */

    if ( ! _KeCodecHashtable )
    {
	if ( _KeCodecHashtable = KeHashtableMake (KE_CODEC_HASHTABLE_SIZE
					       , codec_hash_fct
					       , codec_cmp_fct) )
	{
	    (void) KeCodecBYTEGet () ;
	    (void) KeCodecCARD16Get () ;
	    (void) KeCodecCARD32Get () ;
	    (void) KeCodecSTRINGGet () ;
	    (void) KeCodecUNDEFGet () ;
	}
	else
	{
	    return ((KeCodec) 0) ;
	}
    }

/*
 * Try to find it first :
 */
    
    codec_rec.name = name ;

    if ( new_codec = (KeCodec) KeHashtableFind (_KeCodecHashtable, &codec_rec) )
    {
	return (new_codec) ;
    }


/*
 * Add it :
 */

    if ( new_codec = (KeCodec) malloc (sizeof (KeCodecRec)) )
    {
	new_codec->name          = (char *) strdup (name) ;
	new_codec->sizer         = sizer ;
	new_codec->encoder       = encoder ;
	new_codec->encoder_free  = enc_free ;
	new_codec->decoder       = decoder ;
	new_codec->decoder_free  = dec_free ;
	
	KeHashtableAdd (_KeCodecHashtable, new_codec) ;
    }

    return (new_codec) ;
}

/***************************************************************************
 * KeCodecDecode
 * KeCodecEncode
 * KeCodecSize 
 * KeCodecFree
 *
 * Short hands for invoking a function of a given codec.
 ***************************************************************************/

KeStatus KeCodecDecode (codec, size, value, swapped, to, emsg, efree)
    KeCodec   	   codec ;
    int       	   size ;
    KePointer 	   value ;
    int       	   swapped ;
    KePointer 	   *to ;
    char           **emsg ;
    KeErrorFreeFct *efree ;

{
    if ( codec->decoder )
	return ((codec->decoder)(size, value, swapped, to, emsg, efree)) ;
    *emsg  = ErrNoDecoder ;
    *efree = (KeErrorFreeFct) 0 ;
    return (KEDIT_InvalidOp) ;
}

KeStatus KeCodecEncode (codec, pointer, size_ret, value_ret, emsg, efree)
    KeCodec  	   codec ;
    KePointer 	   pointer ;
    int       	   *size_ret ;
    KePointer 	   *value_ret ;
    char      	   **emsg ;
    KeErrorFreeFct *efree ;
{
    if ( codec->encoder )
	return ((codec->encoder) (pointer, size_ret, value_ret, emsg, efree)) ;
    *emsg  = ErrNoEncoder ;
    *efree = (KeErrorFreeFct) 0 ;
    return (KEDIT_Failure) ;
}

KeStatus KeCodecSize (codec, pointer, size_ret)
    KeCodec  codec ;
    KePointer pointer ;
    int       *size_ret ;

{
    if ( codec->sizer )
	return ((codec->sizer) (pointer, size_ret)) ;
    return (KEDIT_Failure) ;
}

void KeCodecEncoderFree (codec, pointer)
    KeCodec  codec ;
    KePointer pointer ;

{
    if ( pointer && ( codec->encoder_free ) )
	(codec->encoder_free) (pointer) ;
    return ;
}

void KeCodecDecoderFree (codec, pointer)
    KeCodec  codec ;
    KePointer pointer ;

{
    if ( pointer && ( codec->decoder_free ) )
	(codec->decoder_free) (pointer) ;
    return ;
}

/***************************************************************************
 * KeCodecBYTEGet: default BYTE codec handling
 *
 * RETURNS: codec object
 ***************************************************************************/

static KeStatus KeCodecBYTESize (pointer, size_ret)
    KePointer pointer ;
    int       *size_ret ;

{
    *size_ret = sizeof (BYTE) ;
    return (KEDIT_Success) ;
}

static KeStatus KeCodecBYTEEncode (pointer, size_ret, value_ret, mret, fret)
    KePointer 	    pointer ;
    int       	    *size_ret ;
    KePointer 	    *value_ret ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;

{
    *size_ret  = sizeof (BYTE) ;
    *value_ret = pointer ;
    return (KEDIT_Success) ;
}

static KeStatus KeCodecBYTEDecode (size, value, swapped, to, mret, fret)
    int       	    size ;
    KePointer 	    value ;
    int       	    swapped ;
    KePointer 	    *to ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;

{
    static BYTE scratch ;

    if ( ! *to )
	*to = (KePointer) &scratch ;

    if ( size == sizeof (BYTE) ) {
	*((BYTE *) *to) = *((BYTE *) value) ;
	return (KEDIT_Success) ;
    }
    *mret = ErrInvalidSize ;
    *fret = (KeErrorFreeFct) 0 ;
    return (KEDIT_InvalidData) ;
}

KeCodec KeCodecBYTEGet ()

{
    if ( ! KeCodecBYTE )
    {
	KeCodecBYTE = KeCodecDefine ("BYTE"
				     , KeCodecBYTESize
				     , KeCodecBYTEEncode
				     , (KeCodecFreeer) 0
				     , KeCodecBYTEDecode
				     , (KeCodecFreeer) 0) ;
    }
    return (KeCodecBYTE) ;
}

/***************************************************************************
 * KeCodecCARD16Get: default CARD16 codec handling
 *
 * RETURNS: codec object.
 ***************************************************************************/

static KeStatus KeCodecCARD16Size (pointer, size_ret)
    KePointer pointer ;
    int       *size_ret ;

{
    *size_ret = sizeof (CARD16) ;
    return (KEDIT_Success) ;
}

static KeStatus KeCodecCARD16Encode (pointer, size_ret, value_ret, mret, fret)
    KePointer 	    pointer ;
    int       	    *size_ret ;
    KePointer 	    *value_ret ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;
{
    *size_ret  = sizeof (CARD16) ;
    *value_ret = pointer ;
    return (KEDIT_Success) ;
}

static KeStatus KeCodecCARD16Decode (size, value, swapped, to, mret, fret)
    int       	    size ;
    KePointer 	    value ;
    int       	    swapped ;
    KePointer 	    *to ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;
{
    static CARD16 scratch ;

    if ( ! *to )
	*to = (KePointer) &scratch ;
    
    if ( size == sizeof (CARD16) ) {
	CARD16 card16 = *((CARD16 *) value) ;

	*((CARD16 *) *to) = (swapped ? lswaps (card16) : card16) ;
	return (KEDIT_Success) ;
    }
    *mret = ErrInvalidSize ;
    *fret = (KeErrorFreeFct) 0 ;
    return (KEDIT_Failure) ;
}

KeCodec KeCodecCARD16Get ()

{
    if ( ! KeCodecCARD16 )
    {
	KeCodecCARD16 = KeCodecDefine ("CARD16"
				       , KeCodecCARD16Size
				       , KeCodecCARD16Encode
				       , (KeCodecFreeer) 0
				       , KeCodecCARD16Decode
				       , (KeCodecFreeer) 0) ;
    }
    return (KeCodecCARD16) ;
}


/***************************************************************************
 * KeCodecCARD32Get: default CARD32 codec handling
 *
 * RETURNS: codec object.
 ***************************************************************************/

static KeStatus KeCodecCARD32Size (pointer, size_ret)
    KePointer pointer ;
    int       *size_ret ;

{
    *size_ret = sizeof (CARD32) ;
    return (KEDIT_Success) ;
}

static KeStatus KeCodecCARD32Encode (pointer, size_ret, value_ret, mret, fret)
    KePointer 	    pointer ;
    int       	    *size_ret ;
    KePointer 	    *value_ret ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;
{
    *size_ret  = sizeof (CARD32) ;
    *value_ret = pointer ;
    return (KEDIT_Success) ;
}

static KeStatus KeCodecCARD32Decode (size, value, swapped, to, mret, fret)
    int       	    size ;
    KePointer 	    value ;
    int       	    swapped ;
    KePointer 	    *to ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;

{
    static CARD32 scratch ;

    if ( ! *to )
	*to = (KePointer) &scratch ;

    if ( size == sizeof (CARD32) ) {
	CARD32 card32 = *((CARD32 *) value) ;

	*((CARD32 *) *to) = (swapped ? lswapl (card32) : card32) ;
	return (KEDIT_Success) ;
    }
    *mret = ErrInvalidSize ;
    *fret = (KeErrorFreeFct) 0 ;
    return (KEDIT_Failure) ;
}

KeCodec KeCodecCARD32Get ()

{
    if ( ! KeCodecCARD32 )
    {
	KeCodecCARD32 = KeCodecDefine ("CARD32"
				       , KeCodecCARD32Size
				       , KeCodecCARD32Encode
				       , (KeCodecFreeer) 0
				       , KeCodecCARD32Decode
				       , (KeCodecFreeer) 0) ;
    }
    return (KeCodecCARD32) ;
}


/***************************************************************************
 * KeCodecSTRINGGet: default STRING codec handling
 *
 * RETURNS: codec object
 ***************************************************************************/

static KeStatus KeCodecSTRINGSize (pointer, size_ret)
    KePointer pointer ;
    int       *size_ret ;

{
    if ( pointer ) 
	*size_ret = strlen ((char *) pointer) ;
    else
	*size_ret = 0 ;
    return (KEDIT_Success) ;
}

static KeStatus KeCodecSTRINGEncode (pointer, size_ret, value_ret, mret, fret)
    KePointer 	    pointer ;
    int       	    *size_ret ;
    KePointer 	    *value_ret ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;
{

    if ( pointer ) {
	*size_ret  = strlen ((char *) pointer) ;
	*value_ret = (KePointer) pointer ;
    } else {
	*size_ret  = 0 ;
	*value_ret = (KePointer) 0 ;
    }
    return (KEDIT_Success) ;
}

static KeStatus KeCodecSTRINGDecode (size, value, swapped, to, mret, fret)
    int       	    size ;
    KePointer 	    value ;
    int       	    swapped ;
    KePointer 	    *to ;
    char      	    **mret ;
    KeErrorFreeFct  *fret ;
{
    if ( ( ! *to ) && ( ! (*to = (KePointer) malloc (size + 1)) ) ) {
	*mret = ErrBadMalloc ;
	*fret = (KeErrorFreeFct) 0 ;
	return (KEDIT_BadMalloc) ;
    }
    if ( size ) 
	bcopy ((char *) value, (char *) *to, size) ;
    *(((char *) *to) + size) = '\0' ;
    return (KEDIT_Success) ;
}

KeCodec KeCodecSTRINGGet ()

{
    if ( ! KeCodecSTRING )
    {
	KeCodecSTRING = KeCodecDefine ("STRING"
				       , KeCodecSTRINGSize
				       , KeCodecSTRINGEncode
				       , (KeCodecFreeer) 0
				       , KeCodecSTRINGDecode
				       , (KeCodecFreeer) KeFree) ;
    }
    return (KeCodecSTRING) ;
}
   
/***************************************************************************
 * KeCodeUNDEFGet: undef codec, for data that you don't export
 *
 * This type just means: I just warn you such data exists, but I don't 
 * want (can) you to access it.
 ***************************************************************************/

KeCodec KeCodecUNDEFGet ()

{
    if ( ! KeCodecUNDEF )
    {
	KeCodecUNDEF = KeCodecDefine ("UNDEF"
				      , (KeCodecSizer) 0 
				      , (KeCodecEncoder) 0
				      , (KeCodecFreeer) 0
				      , (KeCodecDecoder) 0
				      , (KeCodecFreeer) 0) ;
    }
    return (KeCodecUNDEF) ;
}

