/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KSIce.c
 *
 * Ice settings
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <KeServer.h>
#include <KesServerP.h>

typedef KesRequest (*KeditDecoder) (KesServer s, char *pbuf, Bool swap) ;
typedef void       (*KeditEncoder) (KesClient client, KesReply reply) ;

static KesRequest _KesBadOpcode (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesOpenConnectionDecoder (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesLsDecoder (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesStatDecoder (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesGetDecoder (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesSetDecoder (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesExecDecoder (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesNotifyDecoder (KesServer s, char *pbuf, Bool swap) ;
static KesRequest _KesNoopDecoder (KesServer s, char *pbuf, Bool swap) ;

static void _KesBadReplyEncoder (KesClient client, KesReply reply) ;
static void _KesErrorEncoder (KesClient client, KesReply reply) ;
static void _KesOpenConnectionReplyEncoder (KesClient client, KesReply reply) ;
static void _KesLsReplyEncoder (KesClient client, KesReply reply) ;
static void _KesStatReplyEncoder (KesClient client, KesReply reply) ;
static void _KesGetReplyEncoder (KesClient client, KesReply reply) ;
static void _KesEventEncoder (KesClient client, KesReply reply) ;
static void _KesNoopReplyEncoder (KesClient client, KesReply reply) ;

static KeditDecoder _KeditDecoderArray[] =
    { _KesBadOpcode
    , _KesBadOpcode
    , _KesOpenConnectionDecoder
    , _KesLsDecoder
    , _KesStatDecoder
    , _KesGetDecoder
    , _KesSetDecoder
    , _KesExecDecoder
    , _KesNotifyDecoder
    , _KesBadOpcode
    , _KesNoopDecoder } ;

static KeditEncoder _KeditEncoderArray[] = 
    { _KesErrorEncoder
    , _KesBadReplyEncoder			/* KEDIT_Reply */
    , _KesOpenConnectionReplyEncoder
    , _KesLsReplyEncoder
    , _KesStatReplyEncoder
    , _KesGetReplyEncoder
    , _KesBadReplyEncoder			/* KEDIT_Set */
    , _KesBadReplyEncoder			/* KEDIT_Exec */
    , _KesBadReplyEncoder			/* KEDIT_Notify */
    , _KesEventEncoder 
    , _KesNoopReplyEncoder } ;

/***************************************************************************
 * Encoding replies, one function per reply
 ***************************************************************************/

static void _KesBadReplyEncoder (client, reply)
    KesClient client ;
    KesReply  reply ;

{
    fprintf (stderr, "libKs[_KesBadReplyEncoder]: invalid reply (fatal)\n") ;
    exit (1) ;
}

static void _KesErrorEncoder (client, oreply)
    KesClient client ;
    KesReply  oreply ;

{
    KesError   error = (KesError) oreply ;
    keditError *werror ;
    char       *pdata ;
    int        msg_len ;
    int        extra = sizeof (keditError) - sizeof (keditRequest) ;
    
    msg_len = (error->msg) ? strlen (error->msg) : 0 ;
    extra  += WORD64COUNT (sizeof (CARD16) + msg_len) ;

    IceGetHeaderExtra (client->ice_connection
		       , _IceKeditPaProtocolNumber
		       , KEDIT_Error
		       , sizeof (keditRequest)
		       , extra 
		       , keditError
		       , werror
		       , pdata) ;

    STORE_CARD8 (pdata, KEDIT_Reply) ;        /* offending opcode */
    STORE_CARD8 (pdata, IceCanContinue) ;	/* severity */
    pdata += sizeof (CARD16) ;			/* skip unused */
    STORE_CARD32 (pdata, error->serial) ;
    STORE_CARD32 (pdata, error->status) ;
    STORE_CARD32 (pdata, error->node) ;

    STORE_CARD16 (pdata, msg_len) ;
    if ( msg_len )
	bcopy (error->msg, pdata, msg_len) ;
	
    return ;
}

static void _KesOpenConnectionReplyEncoder (client, oreply)
    KesClient client ;
    KesReply  oreply ;

{
    KesOpenConnectionReply   reply = (KesOpenConnectionReply) oreply ;
    keditOpenConnectionReply *wreply ;
    char 		     *pdata ;

    IceGetHeaderExtra (client->ice_connection
		       , _IceKeditPaProtocolNumber
		       , KEDIT_Reply
		       , sizeof (keditRequest)
		       , (sizeof (keditOpenConnectionReply) - sizeof (keditRequest)) >>3
		       , keditOpenConnectionReply
		       , wreply
		       , pdata) ;

    STORE_CARD32 (pdata, reply->client_tag) ;
    STORE_CARD32 (pdata, reply->root) ;
    
    return ;
}

static void _KesLsReplyEncoder (client, oreply)
    KesClient  client ;
    KesReply   oreply ;

{
    KesLsReply   reply = (KesLsReply) oreply ;
    keditLsReply *wreply ;
    char         *pdata ;
    int          bytes = (reply->num_children * sizeof (KID)) ;

    IceGetHeaderExtra (client->ice_connection
		       , _IceKeditPaProtocolNumber
		       , KEDIT_Reply
		       , sizeof (keditRequest)
		       , (sizeof (keditLsReply) - sizeof (keditRequest)) >> 3
		       , keditLsReply
		       , wreply
		       , pdata) ;
    
    if ( ! pdata ) {
	fprintf (stderr, "libKs[_KesLsReplyEncoder]: unable to get pdata\n") ;
	exit (1) ;
    }

    STORE_CARD32 (pdata, reply->serial) ;
    STORE_CARD32 (pdata, reply->num_children) ;

    if ( bytes ) {
	/*
	 * It is safe to write in request *before* IceWriteData:
	 */
	wreply->length += WORD64COUNT (bytes) ;
	IceWriteData (client->ice_connection, bytes, (char *) reply->children);
	IceWritePad (client->ice_connection, PAD64 (bytes)) ;
    }

    return ;
}

static void _KesStatReplyEncoder (client, oreply)
    KesClient client ;
    KesReply  oreply ;

{
    KesStatReply   reply = (KesStatReply) oreply ;
    keditStatReply *wreply ;
    char           *pdata ;
    int		   extra   = sizeof (keditStatReply) - sizeof (keditRequest) ;
    int		   namelen = strlen (reply->stat.name) ;

    extra += WORD64COUNT (namelen + sizeof (CARD16)) ;

    IceGetHeaderExtra (client->ice_connection
		       , _IceKeditPaProtocolNumber
		       , KEDIT_Reply
		       , sizeof (keditRequest)
		       , extra
		       , keditStatReply
		       , wreply
		       , pdata) ;

    STORE_CARD32 (pdata, reply->serial) ;
    STORE_CARD16 (pdata, reply->stat.mode) ;
    STORE_CARD16 (pdata, reply->stat.mask) ;
    STORE_CARD32 (pdata, reply->stat.id) ;
    STORE_CARD32 (pdata, reply->stat.size) ;
    STORE_CARD32 (pdata, reply->stat.info) ;
    pdata += sizeof (CARD32) ;				/* skip unused32 */
    STORE_CARD16 (pdata, namelen) ;
    bcopy (reply->stat.name, pdata, namelen) ;
	
    return ;
}

static void _KesGetReplyEncoder (client, oreply)
    KesClient client ;
    KesReply  oreply ;

{
    KesGetReply   reply = (KesGetReply) oreply ;
    keditGetReply *wreply ;
    char          *pdata ;
    
    IceGetHeaderExtra (client->ice_connection
		       , _IceKeditPaProtocolNumber
		       , KEDIT_Reply
		       , sizeof (keditRequest)
		       , (sizeof (keditGetReply) - sizeof (keditRequest)) >> 3
		       , keditGetReply
		       , wreply
		       , pdata) ;

    STORE_CARD32 (pdata, reply->serial) ;
    STORE_CARD32 (pdata, reply->size) ;

    if ( reply->size ) {
	wreply->length += WORD64COUNT (reply->size) ;
	IceWriteData (client->ice_connection, reply->size, reply->value) ;
	IceWritePad (client->ice_connection, PAD64 (reply->size)) ;
    }

    return ;
}

static void _KesEventEncoder (client, oreply)
    KesClient client ;
    KesReply  oreply ;

{
    KesEvent   event = (KesEvent) oreply ;
    keditEvent *wevent ;
    char       *pdata ;
    int	       extra = sizeof (keditEvent) - sizeof (keditRequest) ;

    switch ( event->any.event_type )
    {
      case KEDIT_ReadWithValueEvent:
	extra += WORD64COUNT (event->read_with_value_event.size) ;
	break ;
      case KEDIT_WriteWithValueEvent:
	extra += WORD64COUNT (event->write_with_value_event.size) ;
	break ;
    }

    IceGetHeaderExtra (client->ice_connection
		       , _IceKeditPaProtocolNumber
		       , KEDIT_Event
		       , sizeof (keditRequest)
		       , extra 
		       , keditEvent
		       , wevent
		       , pdata) ;

    wevent->event_type = event->any.event_type ;

    STORE_CARD32 (pdata, event->any.node) ;
    STORE_CARD32 (pdata, event->any.timestamp) ;
    STORE_CARD32 (pdata, event->any.client_tag) ;

    switch ( event->any.event_type )
    {
      case KEDIT_NotifyEvent:
	STORE_CARD16 (pdata, event->notify_event.mask) ;
	break ;
      case KEDIT_CreateEvent:
	STORE_CARD32 (pdata, event->create_event.child) ;
	break ;
      case KEDIT_ReadWithValueEvent:
	STORE_CARD32 (pdata, event->read_with_value_event.size) ;
	if ( event->read_with_value_event.size )
	    bcopy (event->read_with_value_event.value
		   , pdata
		   , event->read_with_value_event.size) ;
	break ;
      case KEDIT_WriteWithValueEvent:
	STORE_CARD32 (pdata, event->write_with_value_event.size) ;
	if ( event->write_with_value_event.size )
	    bcopy (event->write_with_value_event.value
		   , pdata
		   , event->write_with_value_event.size) ;
	break ;
    }

    return ;
}

static void _KesNoopReplyEncoder (client, oreply)
    KesClient client ;
    KesReply  oreply ;

{
    KesNoopReply   reply = (KesNoopReply) oreply ;
    keditNoopReply *wreply ;
    char           *pdata ;

    IceGetHeaderExtra (client->ice_connection
		       , _IceKeditPaProtocolNumber
		       , KEDIT_Reply
		       , sizeof (keditRequest)
		       , (sizeof (keditNoopReply) - sizeof (keditRequest)) >> 3
		       , keditNoopReply
		       , wreply
		       , pdata) ;
    
    STORE_CARD32 (pdata, reply->serial) ;
    
    return ;
}

void _KesWriteReplyToClient (reply)
    KesReply  reply ;

{
    (_KeditEncoderArray[KesTypeOfReply(reply)])(KesClientOfReply(reply), reply) ;
    return ;
}

/***************************************************************************
 * Decoding requests, one function per network request:
  ***************************************************************************/

static KesRequest _KesBadOpcode (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    fprintf (stderr, "libKs[KesBadOpcode]: invalid request type (fatal).\n") ;
    exit (1) ;
}
    
static KesRequest _KesOpenConnectionDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KesOpenConnectionReq request ;

    request = KesMalloc (server, sizeof (KesOpenConnectionReqRec)) ;
    return ((KesRequest) request) ;
}

static KesRequest _KesLsDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KesLsReq request ;

    request = KesMalloc (server, sizeof (KesLsReqRec)) ;

    EXTRACT_CARD32 (pbuf, swap, request->node) ;

    return ((KesRequest) request) ;
}

static KesRequest _KesStatDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KesStatReq request ;

    request = KesMalloc (server, sizeof (KesStatReqRec)) ;
    
    EXTRACT_CARD32 (pbuf, swap, request->node) ;

    return ((KesRequest) request) ;
}

static KesRequest _KesGetDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KesGetReq request ;

    request = KesMalloc (server, sizeof (KesGetReqRec)) ;

    EXTRACT_CARD32 (pbuf, swap, request->node) ;

    return ((KesRequest) request) ;
}

static KesRequest _KesSetDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KID           node ;
    unsigned int  size ;
    KesSetReq     request ;
    
    EXTRACT_CARD32 (pbuf, swap, node) ;
    EXTRACT_CARD32 (pbuf, swap, size) ;
    
    request = KesMalloc (server, sizeof (KesSetReqRec) + size) ;
    request->node = node ;
    if ( request->size = size ) 
	bcopy (pbuf, request->value, size) ;
   
    return ((KesRequest) request) ;
}

static KesRequest _KesExecDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KesExecReq request ;

    request = KesMalloc (server, sizeof (KesExecReqRec)) ;
    
    EXTRACT_CARD32 (pbuf, swap, request->node) ;
    
    return ((KesRequest) request) ;
}

static KesRequest _KesNotifyDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KesNotifyReq request ;

    request = KesMalloc (server, sizeof (KesNotifyReqRec)) ;

    EXTRACT_CARD32 (pbuf, swap, request->node) ;
    EXTRACT_CARD16 (pbuf, swap, request->mask) ;
    
    return ((KesRequest) request) ;
}

static KesRequest _KesNoopDecoder (server, pbuf, swap)
    KesServer server ;
    char      *pbuf ;
    Bool      swap ;

{
    KesNoopReq request ;

    request = KesMalloc (server, sizeof (KesNoopReqRec)) ;
    return ((KesRequest) request) ;
}

/***************************************************************************
 * Free any request
 ***************************************************************************/

void KesRequestFree (request)
    KesRequest request ;

{
    switch ( KesTypeOfRequest (request) )
    {
      case KEDIT_OpenConnection:
      case KEDIT_Ls:
      case KEDIT_Stat:
      case KEDIT_Set:
      case KEDIT_Get:
      case KEDIT_Exec:
      case KEDIT_Notify:
      case KEDIT_Noop:
        KeFree (request) ;
        break ;
      default:
        fprintf (stderr, "KesRequestFree: unknown request.\n") ;
        break ;
    }
    return ;
}

/***************************************************************************
 * _IceKeditPAProcessMessage: process incomming K-edit messages
 *
 * RETURN: void (according to Ice specs)
 ***************************************************************************/

void _IceKeditPAProcessMessage_1_0 (ice_connection,ice_ptr,opcode,length,swap)
    IceConn       ice_connection ;
    IcePointer	  ice_ptr ;
    int	          opcode ;
    unsigned long length ;
    Bool	  swap ;

{
    KesClient     client = _IceConnToClient (ice_connection) ;
    KesRequest    request ;
    keditRequest  *packet ;
    char          *pdata ;

    IceReadCompleteMessage (ice_connection
			    , sizeof (keditRequest)
			    , keditRequest
			    , packet
			    , pdata) ;

/*
 * Decode the packet according to its type, *than* the common header
 */

    request = (_KeditDecoderArray[opcode]) (client->server, pdata, swap) ;

    KesTypeOfRequest (request)       = opcode ;
    KesSubtypeOfRequest (request)    = 0 ;
    KesServerOfRequest (request)     = KesServerOfClient (client) ;
    KesClientOfRequest (request)     = client ;
    KesSerialOfRequest (request)     = (client->serial)++ ;
    KesClosureOfRequest(request)     = (KePointer) 0 ;
    KesHandlerSetOfRequest (request) = (KesHandlerSet) 0 ;

    IceDisposeCompleteMessage (ice_connection, pdata) ;

/*
 * Now dispatch the request:
 */

    KesRequestDispatch (request) ;
    KesRequestFree (request) ;

    return ;
}
