/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KSTDir.c
 *
 ***************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <KeServer.h>

static KesHandlerSet KesDirectoryHandlerSet = (KesHandlerSet) 0 ;

/***************************************************************************
 * KesDirectoryHandlerSet: directory callback set.
 *
 * This callback set handles only the Ls and Stat requests through:
 * KesDirectoryLsHandler
 * KesDirectoryStatHandler
 * KesDirectoryNotifyHandler
 *
 * Also provided, an accesor for the KesDirectoryHandlerSet, which eventually
 * initialize it.
 ***************************************************************************/

static KeStatus KesDirectoryLsHandler (request, reply)
    KesLsReq request ;
    KesReply reply ;

{
    KesDirectory directory ;

    directory = (KesDirectory) KesClosureOfRequest (request) ;

    if ( ! directory->filter ) {
	KesLsReplyFillStatic (reply, directory->used, directory->subdirs) ;
    } else {
	KID *ids = (KID *) KesMalloc (KesServerOfRequest (request)
				      , sizeof (KID) * directory->used) ;
	KID *pid = ids ;
	int i ;
	
	for (i = 0 ; i < directory->used ; i++ ) {
	    if ( directory->filter (directory,request,directory->subdirs[i]) )
		*pid++ = directory->subdirs[i] ;
	}
	KesLsReplyFill (reply, pid - ids, ids, (KesReplyFreeFct) KeFree) ;
    }

    return (KEDIT_Success) ;
}

static KeStatus KesDirectoryStatHandler (request, reply)
    KesStatReq request ;
    KesReply   reply ;

{
    KesDirectory directory ;

    directory = (KesDirectory) KesClosureOfRequest (request) ;
    
    KesStatReplyFillStatic (reply
			    , directory->name
			    , KEDIT_DirectoryMode
			    , (KEDIT_CreateMask | KEDIT_DeleteMask | KEDIT_NotifyMask)
			    , (KID) 0
			    , 0) ;

    return (KEDIT_Success) ;
}

static KeStatus KesDirectoryNotifyHandler (request, generic_reply)
    KesNotifyReq request ;
    KesReply     generic_reply ;

{
    KesNotifyNotify (request->server
		     , request->client->client_tag
		     , request->node
		     , request->mask) ;

    KesNodeNotifySet (request->client, request->node, request->mask) ;
    return (KEDIT_Success) ;
}

KesHandlerSet KesDirectoryHandlerSetGet ()

{
/*
 * Initialize (if not done) :
 */

    if ( ! KesDirectoryHandlerSet )
    {
	KesDirectoryHandlerSet = KesHandlerSetMake() ;
	
	KesHandlerSetLsHandler     (KesDirectoryHandlerSet, KesDirectoryLsHandler) ;
	KesHandlerSetStatHandler   (KesDirectoryHandlerSet, KesDirectoryStatHandler) ;
	KesHandlerSetNotifyHandler (KesDirectoryHandlerSet, KesDirectoryNotifyHandler) ;
    }

    return (KesDirectoryHandlerSet) ;
}

/***************************************************************************
 * KesSubdirDelete: remove a subdir from given directory
 * KesSubdirMake  : add a subdirectory to the given directory
 *
 * This functions add a new subdirectory to the given directory
 *
 * RETURNS: the augmnented directory or NULL
 ***************************************************************************/

KesDirectory KesDirectoryRemoveChild (directory, subdir_id)
    KesDirectory directory ;
    KID          subdir_id ;

{
    KID *subdirs = directory->subdirs ;
    int used     = directory->used ;

/*
 * Find the subdir_id in the list :
 */
    while ( used )
    {
	if ( *subdirs == subdir_id )
	    break ;
	used-- ;
	subdirs++ ;
    }
/*
 * Remove it, compacting the array :
 */

    if ( used )
    {
	(directory->used)-- ;
	if ( used - 1 > 0 )
	    memcpy (subdirs, subdirs + 1, (used - 1) * sizeof (KID)) ;
    }

    KesDeleteNotify (directory->server, (unsigned long) 0, subdir_id) ;
    
    return (directory) ;
}

/*
 * Quantity to resize subdir array, should be commented in a general config.h file
 */

#define SUBDIR_ARRAY_CHUNK	16

KesDirectory KesDirectoryAddChild (directory, subdir_id)
    KesDirectory directory ;
    KID          subdir_id ;

{
/*
 * If needed, resize the container :
 */

    if ( (! directory->subdirs) 
	|| (directory->used + 1 >= directory->size) )
    {
	int size = directory->size + SUBDIR_ARRAY_CHUNK ;

	if ( directory->subdirs )
	{
	    directory->subdirs = KesRealloc (directory->server
					     , directory->subdirs
					     , size * sizeof (KID)) ;
	}
	else
	{
	    directory->subdirs = KesMalloc (directory->server
					    , size * sizeof (KID)) ;
	}
	
	if ( ! directory->subdirs )
	{
	    KesDirectoryDelete (directory) ;
	    return ((KesDirectory) 0) ;
	}
	else
	{
	    directory->size = size ;
	}
    }

/*
 * Add the new subdir :
 */

    directory->subdirs[(directory->used)++] = subdir_id ;

    KesCreateNotify (directory->server
		     , (unsigned long) 0
		     , directory->node
		     , subdir_id) ;

    return (directory) ;
}

/***************************************************************************
 * KesDirectoryMake  : make a new domain.
 * KesDirectroyDelete: delete (and remove) a directroy
 *
 * This function creates a "directory" which ios able to contain any 
 * domain. The return value is a domain descriptor.
 ***************************************************************************/

void KesDirectoryDelete (directory)
    KesDirectory directory ;

{
/* 
 * Free associated resources, and the directroy itself :
 */

    KesNodeFree (directory->server, directory->node) ;
    free (directory->name) ;
    if ( directory->subdirs )
	KeFree (directory->subdirs) ;
    KeFree (directory) ;
	    
    return ;
}

KesDirectoryFilter KesDirectoryFilterSet (directory, filter)
    KesDirectory       directory ;
    KesDirectoryFilter filter ;

{
    KesDirectoryFilter old = directory->filter ;

    directory->filter = filter ;
    return (old) ;
}
	
KesDirectory KesDirectoryMake (server, name)
    KesServer server ;
    char      *name ;

{
    KesDirectory directory = KesMalloc (server, sizeof (KesDirectoryRec)) ;

/*
 * Create the directory :
 */

    if ( directory )
    {
	directory->closure = (KePointer) 0 ;
	directory->server  = server ;
	directory->filter  = (KesDirectoryFilter) 0 ;
	directory->name    = (char *) strdup (name) ;
	directory->size    = 0 ;
	directory->used    = 0 ;
	directory->subdirs = (KID *) 0 ;
	KesNodeMake (server
		     , KesDirectoryHandlerSetGet()
		     , (KePointer) directory
		     , &(directory->node)) ;
    }

    return (directory) ;
}

