/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KSTDisp.c
 *
 * Dispatch the request according to out id table. Defined functions
 * KesDispatchRequest
 * _KesLsReplyDefaultFreeFct
 * _KesGetReplyDefaultFreFct
 ***************************************************************************/

#include <stdio.h>
#include <string.h>

#include <KeServer.h>
#include <KesServerP.h>

/***************************************************************************
 * apply_free_function: apply free function on reply
 *
 * Datat provided by server programmers may need to be freed once send. This 
 * function applies a 'free-hook' to the reply, in order to free server's
 * programmers reply toi K-edit server library.
 *
 * RETURNS : void
 ***************************************************************************/

static void apply_free_function (reply) 
    KesReply reply ;

{
    if ( reply->free_fct )
    {
	switch ( KesTypeOfReply (reply) )
	{
	  case KEDIT_Error:
	    {
		KesError error = (KesError) reply ;

		if ( error->msg )
		    (reply->free_fct) (error->msg) ;
	    }
	    break;
	  case KEDIT_Ls:
	    {
		KesLsReply ls = (KesLsReply) reply ;

		if ( ls->children )
		    (reply->free_fct) (ls->children) ;
	    }
	    break;
	  case KEDIT_Stat:
	    {
		KesStatReply stat = (KesStatReply) reply ;

		if ( stat->stat.name )
		    (reply->free_fct) (stat->stat.name) ;
	    }
	    break ;
	  case KEDIT_Get:
	    {
		KesGetReply set = (KesGetReply) reply ;

		if ( set->value )
		    (reply->free_fct) (set->value) ;
	    }
	    break ;
	}
    }
    return ;
}

/***************************************************************************
 * do_error: perform action associated to errors.
 *
 * As we want the user to be able to overide the default error mechanism
 * this function first check for a specific hook error before sending
 * back the default error packet.
 * 
 * RETURNS: void.
 ***************************************************************************/

static void do_error (request, reply, oid, status)
    KesRequest request ;
    KesReply   reply ;
    KID	       oid ;
    KeStatus   status ;

{
    KesServer   server = KesServerOfRequest (request) ;
    KesError    error  = (KesError) reply ;

/*
 * Fill in a default error in reply :
 */

    ((KesReply) error)->free_fct = (KesReplyFreeFct) 0 ;

    error->type    = KEDIT_Error ;
    error->subtype = 0 ;
    error->status  = status ;
    error->node    = oid ;
    error->serial  = KesSerialOfRequest (request) ;

    switch ( status )
    {
      case KEDIT_Failure: 
	error->msg = "General failure from handler." ;
	break ;
      case KEDIT_InvalidId:
	error->msg = "Invalid id in request." ;
	break ;
      case KEDIT_InvalidOp:
	error->msg = "This id doesn't support this operation." ;
	break ;
      default:
	error->msg = "Unknown error." ;
	break ;
    }

/*
 * If an error hook as been provided, then invoke it :
 */

    if ( server->error_hook )
    {
	KesReplyRec hook_reply ;

	KesTypeOfReply (&hook_reply)    = KesTypeOfReply (reply) ;
	KesSubtypeOfReply (&hook_reply) = KesSubtypeOfRequest (reply) ;
	KesSerialOfReply (&hook_reply)  = KesSerialOfRequest (reply) ;
	KesServerOfReply (&hook_reply)  = KesServerOfRequest (reply) ;
	KesClientOfReply (&hook_reply)  = KesClientOfRequest (reply) ;
	KesFreeFctOfReply (&hook_reply) = (KesReplyFreeFct) 0 ;

	if ( ! (server->error_hook) (request, &hook_reply, error) )
	{
	    /* Patch reply with hook_reply: (ugly-p side-effect) -> NIL  */
	    bcopy (&hook_reply, reply, sizeof (KesReplyRec)) ;
	}
    }

    return ;
}

/***************************************************************************
 * KesPerformRequest: perform the given request
 *
 * This functions perform the request, triggering the right handler, and
 * fill in the provided empty reply.
 * If the concerned handler isn't able to reply properly, the provided 
 * reply is still filled with a default error message.
 *
 * RETURN: a KeStatus
 ***************************************************************************/

KeStatus KesPerformRequest (request, reply)
    KesRequest request ;
    KesReply   reply ;

{
    KID 	  oid ;
    KesServer 	  ks     = KesServerOfRequest (request) ;
    KesError      error  = (KesError) reply ;
    KeStatus 	  status = KEDIT_Failure ;
    KesHandlerSet handler_set ;
    KePointer	  closure ;

/*
 * Set the reply fields we now right now :
 */

    KesTypeOfReply (reply)    = KesTypeOfRequest (request) ;
    KesSubtypeOfReply (reply) = KesSubtypeOfRequest (request) ;
    KesSerialOfReply (reply)  = KesSerialOfRequest (request) ;
    KesServerOfReply (reply)  = KesServerOfRequest (request) ;
    KesClientOfReply (reply)  = KesClientOfRequest (request) ;
    KesFreeFctOfReply (reply) = (KesReplyFreeFct) 0 ;

#define _InternalCheckOid(oid,handler) \
    if ( ! (oid ) ) \
	  { closure = ks->root_closure ; handler_set = ks->root_handler_set ; } \
    else if ( KesNodeToInfo (ks, oid, &closure, &handler_set) ) \
        { do_error (request, reply, oid, KEDIT_InvalidId) ; \
	  return (KEDIT_InvalidId) ; } \
    if ( ( ! handler_set) || ( ! handler_set->handler ) ) \
        { do_error (request, reply, oid, KEDIT_InvalidOp) ; \
	  return (KEDIT_InvalidOp) ; } 

/*
 * Dispatch to the right handler according to request type :
 */

    switch ( KesTypeOfRequest (request) )
    {
      case KEDIT_OpenConnection:
	{
	    KesOpenConnectionReply oc_reply = (KesOpenConnectionReply) reply ;
	    KesClient client = KesClientOfRequest (request) ;

	    oc_reply->client_tag = client->client_tag ;
	    oc_reply->root       = (KID) 0 ;

	    status = KEDIT_Success ;
	}
	break ;
      case KEDIT_Ls:
	{
	    KesLsReq   ls_req = (KesLsReq)   request ;
	    KesLsReply ls_rep = (KesLsReply) reply ;

	    oid = ls_req->node ;
	    _InternalCheckOid (oid, ls_handler) ;
	    ls_req->closure      = closure ;
	    ls_req->handler_set  = handler_set ;
	    reply->free_fct      = (KesReplyFreeFct) KeFree ;
	    ls_rep->children     = (KID *) 0 ;
	    ls_rep->num_children = 0 ;
	    if ( status = (handler_set->ls_handler)(ls_req, reply) )
		do_error (request, reply, oid, status) ;
	}
	break ;
      case KEDIT_Stat:
	{
	    KesStatReq   stat_req = (KesStatReq) request ;
	    KesStatReply stat_rep = (KesStatReply) reply ;

	    oid = stat_req->node ;
	    _InternalCheckOid (oid, stat_handler) ;
	    stat_req->closure     = closure ;
	    stat_req->handler_set = handler_set ;
	    stat_rep->stat.id = oid ;
	    if ( status = (handler_set->stat_handler)(stat_req, reply) )
		do_error (request, reply, oid, status) ;
	}
	break ;
      case KEDIT_Get:
	{
	    KesGetReq get_req   = (KesGetReq) request ;
	    KesGetReply get_rep = (KesGetReply) reply ;

	    oid = get_req->node ;
	    _InternalCheckOid (oid, get_handler) ;
	    get_req->closure     = closure ;
	    get_req->handler_set = handler_set ;
	    reply->free_fct  = (KesReplyFreeFct) KeFree ;
	    get_rep->size    = 0 ;
	    get_rep->value   = (KePointer) 0 ;
	    if ( status = (handler_set->get_handler)(get_req, reply) )
		do_error (request, reply, oid, status) ;
	}
	break ;
      case KEDIT_Set:
	{
	    KesSetReq set_req = (KesSetReq) request ;

	    oid = set_req->node ;
	    _InternalCheckOid (oid, set_handler) ;
	    set_req->closure     = closure ;
	    set_req->handler_set = handler_set ;
	    KesTypeOfReply (reply)          = KEDIT_Error ;
	    KesSubtypeOfReply (reply)       = 0 ;
	    error->status = KEDIT_Success ;
	    error->msg    = (char *) 0 ;
	    error->node   = oid ;
	    if ( status = (handler_set->set_handler)(set_req, reply) )
		do_error (request, reply, oid, status) ;
	}
	break ;
      case KEDIT_Exec:
	{
	    KesExecReq exec_req = (KesExecReq) request ;

	    oid = exec_req->node ;
	    _InternalCheckOid (oid, exec_handler) ;
	    exec_req->closure     = closure ;
	    exec_req->handler_set = handler_set ;
	    KesTypeOfReply (reply)            = KEDIT_Error ;
	    KesSubtypeOfReply (reply)         = 0 ;
	    error->status  = KEDIT_Success ;
	    error->msg     = ((char *) 0) ;
	    error->node    = oid ;
	    if ( status = (handler_set->exec_handler)(exec_req, reply) )
		do_error (request, reply, oid, status) ;
	}
	break ;
      case KEDIT_Notify:
	{
	    KesNotifyReq notify_req = (KesNotifyReq) request ;

	    oid = notify_req->node ;
	    _InternalCheckOid (oid, notify_handler) ;
	    notify_req->closure     = closure ;
	    notify_req->handler_set = handler_set ;
	    KesTypeOfReply (reply)                = KEDIT_Error ;
	    KesSubtypeOfReply (reply)             = 0 ;
	    error->status = KEDIT_Success ;
	    error->msg    = ((char *) 0) ;
	    error->node   = oid ;
	    if ( status = (handler_set->notify_handler)(notify_req, reply) )
		do_error (request, reply, oid, status) ;
	}
	break ;
      case KEDIT_Noop:
	/* Always reply with an emplty reply, wich has already been initliazed */
	status = KEDIT_Success ;
	break ;
      default:
	fprintf (stderr, "KesDispatchRequest: unknown type %d\n(Fatal error).\n"
		 , KesTypeOfRequest (request)) ;
	exit (1) ;
    }

    return (status) ;

#undef _InternalCheckOid

}

/***************************************************************************
 * KesDispatchRequest: dispatch the request according to callbacks
 *
 * This function is quiet tricky, here is the main (per request) algo:
 * 1. Check oid validity
 * 2. If this is root oid, go to root node management functions
 * 3. Invoke the handler with an appropriately filled reply
 * 4. If request needs reply, be sure that at least one reply is sent
 * 5. If request failed, send an error
 * 6. If something has been sent, flush all output
 *
 * RETURN: void
 ***************************************************************************/

void KesRequestDispatch (request)
    KesRequest request ;

{
    KesReplyRec reply_rec ;
    KeStatus    status ;

/*
 * Dispatch to the right handler according to request type :
 */

    status = KesPerformRequest (request, &reply_rec) ;
    switch (KesTypeOfRequest (request))
    {
      case KEDIT_OpenConnection:
      case KEDIT_Ls:
      case KEDIT_Get:
      case KEDIT_Stat:
      case KEDIT_Noop:

/*
	fprintf (stderr, "KesRequestDispatch: reply %d type %d\n"
		 , KesSerialOfReply (&reply_rec)
		 , KesTypeOfReply (&reply_rec)) ;
 */

	if ( KesTypeOfReply (&reply_rec) == KEDIT_Error )
	    KesRequestPrint (stderr, request) ;

	KesWriteToClient (&reply_rec) ;
	apply_free_function (&reply_rec) ;
	break ;
      case KEDIT_Exec:
      case KEDIT_Set:
      case KEDIT_Notify:
	if ( (( ! status ) && reply_rec.ru.error.status) || status )
	{
	    KesWriteToClient (&reply_rec) ;
	    apply_free_function (&reply_rec) ;
	}
	break ;
    }

    IceFlush (KesClientOfRequest(request)->ice_connection) ;

    return ;
}
