/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * KSTNotify.c
 *
 * Toolkit function for the notify functionnality of kedit. This is a 
 * quick and dirty implementation...
 * Defined functions:
 * KesNodeNotifyRemoveById: remove all notified client of
 *     this node.
 * KesNodeNotifyClientKillHandler: the default handler to call or set 
 *     when using this functions on a client.
 * KesNotify: send notification packet for this node.
 *
 * When using these functions, you should at least call KesNodeNotifyRemove
 * in the client kill hook (default kill hook is set to call this function.
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include <KeServer.h>
#include <KesServerP.h>

/*
 * Those struct are used as opaque type :
 */

typedef struct _NotificationHandle {
    KesClient	  client ;		/* client to notify */
    KID		  id ;			/* id to notify */
    unsigned long mask ;		/* with this mask */
} NotificationHandleRec, *NotificationHandle ;
    
/***************************************************************************
 * notification_hash_fct
 * notification_cmp_fct
 *
 * KeHashtable interface
 ***************************************************************************/

static unsigned long notification_hash_fct (handle)
    NotificationHandle handle ;

{
    return ((unsigned long) handle->id) ;
}

static int notification_cmp_fct (handle1, handle2)
    NotificationHandle handle1, handle2 ;

{
    return ( ! (handle1->id == handle2->id) ) ;
}

/***************************************************************************
 * _KesFindInterestedClients: list of clients interested in some events
 *
 * This function returns a staticaaly allocated array of clients interested
 * in the given event mask, and the associated count.
 *
 * RETURNS: # of clients in array
 ***************************************************************************/

int _KesFindInterestedClients (server, node, mask, pclients) 
    KesServer     server ;
    KID           node ;
    unsigned long mask ;
    KesClient     **pclients ;

{
    static KesClient 	  *clients_ret = (KesClient *) 0 ;
    static int       	  clients_size = 0 ;
    int                   count_ret    = 0 ;
    KeHashtable        	  hashtable ;
    NotificationHandleRec handle_rec ;
    NotificationHandle    handle ;

/*
 * Get rid of fast case :
 */

    *pclients = (KesClient *) 0 ;
    if ( ! ( hashtable = server->notifications) )
	return (0) ;

/*
 * Sweep the hashtable to find clients:
 */

    handle_rec.id = node ;

    if ( handle = (NotificationHandle) KeHashtableFind (hashtable, &handle_rec) )
    {
	do
	{
	    if ( handle->mask & mask )
	    {
		/*
		 * Does clients_ret needs some resize ?
		 */
		if ( count_ret + 1 >= clients_size )
		{
		    if ( clients_size )
		    {
			int newsize ;

			clients_size += 16 ;
			newsize       = sizeof (KesClient) * clients_size ;
			clients_ret = (KesClient *) realloc (clients_ret, newsize) ;
		    }
		    else
		    {
			int newsize ;

			clients_size = 16 ;
			newsize      = sizeof (KesClient) * clients_size ;
			clients_ret  = (KesClient *) malloc (sizeof (KesClient) * newsize);
		    }

		    if ( ! clients_ret )
		    {
			fprintf (stderr, "libKs[_KesFindInterestedClients]: bad malloc\n");
			exit (1) ;
		    }

		}
		/*
		 * Add it to clients_ret :
		 */
		clients_ret[count_ret++] = handle->client ;
	    }
	}
	while ( handle = (NotificationHandle) KeHashtableFindNext (hashtable) ) ;
    }

    *pclients = clients_ret ;

    return ( count_ret ) ;
}

/***************************************************************************
 * KesNodeNotifyClientKillHandler: default kill handler for client
 *
 * RETURN : void
 ***************************************************************************/

static void client_notification_remove (handle, client)
    NotificationHandle handle ;
    KesClient          client ;

{
    if ( handle->client == client )
	KeHashtableRemove (client->server->notifications, handle) ;
    return ;
}

void KesNodeNotifyClientKillHandler (client)
    KesClient client ;

{
    KeHashtable hashtable = client->server->notifications ;

    if ( hashtable )
	KeHashtableMap (hashtable, client_notification_remove, client) ;

    return ;
}

/***************************************************************************
 * KesNodeNotifySet: add the given client to this node
 *
 * Add this client to the list of client to be notified on node changes.
 * This is done using the generic hashtable package.
 *
 * RETURN: a KeStatus
 ***************************************************************************/

KeStatus KesNodeNotifySet (client, id, mask)
    KesClient     client ;
    KID           id ;
    unsigned long mask ;

{
    KeHashtable             hashtable ;
    KesServer             server = client->server ;
    NotificationHandleRec handle_rec ;
    NotificationHandle    handle ;
    KePointer	 	  closure ;
    KesHandlerSet	  handler_set ;

/*
 * Check for this server notification hashtable :
 */

    if ( ! (hashtable = server->notifications) )
    {
	hashtable = KeHashtableMake (KES_NOTIFICATIONS_HASHTABLE_SIZE
				   , notification_hash_fct
				   , notification_cmp_fct) ;
	server->notifications = hashtable ;
    }

/* 
 * Check that this id really exists :
 */

    if ( KesNodeToInfo (client->server, id, &closure, &handler_set) )
	return (KEDIT_InvalidId) ;

/*
 * If this has already been set, return it, else change mask :
 */

    handle_rec.id   = id ;

    if ( handle = (NotificationHandle) KeHashtableFind (hashtable, &handle_rec) )
    {
	do
	{
	    if ( handle->client == client )
	    {
		if ( ! mask )
		{
		    KeHashtableRemove (hashtable, handle) ;
		    KeFree (handle) ;
		}
		else
		{
		    handle->mask = mask ;
		}
		return (KEDIT_Success) ;
	    }
	}
	while ( handle = KeHashtableFindNext (hashtable) ) ;
    }

    if ( mask )
    {
	handle = KesMalloc (server, sizeof (NotificationHandleRec)) ;
	
	handle->id     = id ;
	handle->mask   = mask ;
	handle->client = client ;
	KeHashtableAdd (hashtable, handle) ;
	
    }
    
    return (KEDIT_Success) ;

}

/***************************************************************************
 * KesNodeNotifyRemoveById: remove all notifications on given id.
 *
 * RETURN: a KeStatus
 ***************************************************************************/

KeStatus KesNodeNotifyRemoveById (server, id)
    KesServer server ;
    KID       id ;

{
    NotificationHandleRec handle_rec ;
    NotificationHandle    handle ;
    KeHashtable             hashtable = server->notifications ;

    if ( hashtable )
    {
	handle_rec.id = id ;
	
	if ( handle = (NotificationHandle) KeHashtableFind (hashtable, &handle_rec) )
	{
	    do
	    {
		KeHashtableRemove (hashtable, handle) ;
		KeFree (handle) ;
	    }
	    while ( handle = KeHashtableFindNext (hashtable) ) ;
	}
    }

    return (KEDIT_Success) ;
}

/***************************************************************************
 * KesDeleteNotify: send a DeleteEvent to every interested client
 * KesCreateNotify: send a CreateEvent to every interested client
 *
 * RETURNS: # of sent events
 ***************************************************************************/

int KesDeleteNotify (server, client_tag, node)
    KesServer     server ;
    unsigned long client_tag ;
    KID	          node ;

{
    KesDeleteEventRec event_rec ;
    KesClient         *clients ;
    int               count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server,node,KEDIT_DeleteMask,&clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_DeleteEvent ;
	event_rec.node       = node ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}

int KesCreateNotify (server, client_tag, father, node)
    KesServer     server ;
    unsigned long client_tag ;
    KID		  father ;
    KID           node ;

{
    KesCreateEventRec event_rec ;
    KesClient         *clients ;
    int               count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server
					     , father
					     , KEDIT_CreateMask
					     , &clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_CreateEvent ;
	event_rec.node       = father ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	event_rec.child      = node ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}

/***************************************************************************
 * KesNotifyNofify: send a NotifyEvent to every interested clients
 *
 * RETURNS: # of sent events
 ***************************************************************************/

int KesNotifyNotify (server, client_tag, node, mask)
    KesServer     server ;
    unsigned long client_tag ;
    KID           node ;
    unsigned long mask ;

{
    KesNotifyEventRec event_rec ;
    KesClient         *clients ;
    int               count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server,node,KEDIT_NotifyMask,&clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_NotifyEvent ;
	event_rec.node       = node ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	event_rec.mask       = mask ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}

/***************************************************************************
 * KesExecNofify: send a ExecEvent to every interested clients
 *
 * RETURNS: # of sent events
 ***************************************************************************/

int KesExecNotify (server, client_tag, node)
    KesServer     server ;
    unsigned long client_tag ;
    KID           node ;

{
    KesExecEventRec event_rec ;
    KesClient       *clients ;
    int             count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server
					     , node
					     , KEDIT_ExecMask
					     , &clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_ExecEvent ;
	event_rec.node       = node ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}

/***************************************************************************
 * KesReadNotify: send a read event to interested clients.
 * KesWriteEvent: send a write event to interested clients.
 *
 * RETURNS: # of sent events
 ***************************************************************************/

int KesReadNotify (server, client_tag, node)
    KesServer     server ;
    unsigned long client_tag ;
    KID		  node ;

{
    KesReadEventRec event_rec ;
    KesClient       *clients ;
    int             count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server,node,KEDIT_ReadMask,&clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_ReadEvent ;
	event_rec.node       = node ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}


int KesWriteNotify (server, client_tag, node)
    KesServer     server ;
    unsigned long client_tag ;
    KID           node ;

{
    KesWriteEventRec event_rec ;
    KesClient        *clients ;
    int              count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server,node,KEDIT_WriteMask,&clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_WriteEvent ;
	event_rec.node       = node ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}

/***************************************************************************
 * KesReadWithValueNotify:  send a ReadWithValue event to interested clients
 * KesWriteWithValueNotify: send a WriteWithValue event
 *
 * RETURNS: # of sent events
 ***************************************************************************/

int KesReadWithValueNotify (server, client_tag, node, size, value)
    KesServer     server ;
    unsigned long client_tag ;
    KID           node ;
    unsigned int  size ;
    KePointer     value ;

{
    KesWriteWithValueEventRec event_rec ;
    KesClient                 *clients ;
    int                       count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server
					     , node
					     , KEDIT_ReadWithValueMask
					     , &clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_ReadWithValueEvent ;
	event_rec.node       = node ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	event_rec.size       = size ;
	event_rec.value      = value ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}

int KesWriteWithValueNotify (server, client_tag, node, size, value)
    KesServer     server ;
    unsigned long client_tag ;
    KID           node ;
    unsigned int  size ;
    KePointer     value ;

{
    KesWriteWithValueEventRec event_rec ;
    KesClient                 *clients ;
    int                       count, sent = 0 ;

    if ( (count = _KesFindInterestedClients (server
					     , node
					     , KEDIT_WriteWithValueMask
					     , &clients)) > 0 )
    {
	event_rec.type       = KEDIT_Event ;
	event_rec.server     = server ;
	event_rec.event_type = KEDIT_WriteWithValueEvent ;
	event_rec.node       = node ;
	event_rec.timestamp  = (server->timestamp)++ ;
	event_rec.client_tag = client_tag ;
	event_rec.size       = size ;
	event_rec.value      = value ;
	for ( sent = 0 ; sent < count ; sent++, clients++ )
	{
	    event_rec.client = *clients ;
	    KesWriteToClient ((KesReply) &event_rec) ;
	    IceFlush ((*clients)->ice_connection) ;
	}
    }

    return (sent) ;
}


