/***************************************************************************
 * tclServer.c
 * A k-Edit server interface to tcl
 ***************************************************************************/

#include <setjmp.h>

#include <tcl.h>
#include <tk.h>

#include <KeServer.h>
#include <KesCodec.h>
#include <KeCodec.h>

typedef struct _TclServer {
    Tcl_Interp   *interp ;
    KesServer    server ;
    int          listen_count ;
    IceListenObj *listens ;
    KeHashtable    name_space ;
    KID          string_codec ;
} TclServerRec, *TclServer ;
 
typedef struct _TclNode {
    TclServer    tcl_server ;
    KID		 id ;
    unsigned int mask ;
    char         *name ;
    char	 *varname ;
    int          wtrace_count ;
    int          rtrace_count ;
    char         *command ;
    KesDirectory directory ;
} TclNodeRec, *TclNode ;

static KesHandlerSet Tcl_handlerSet ;
static KeCodec       Tcl_stringCodec ;

/* This is broken, I have much better hashtable package now */
#define DEF_NAME_SPACE_SIZE	128

static int nodeConfigure () ;
    
/*
 * Utilities:
 */

static unsigned long node_hash_fct (tcl_node)
    TclNode tcl_node ;

{
    unsigned long hashval = 0, carry ;
    char          *p = tcl_node->name ;

    for ( p = tcl_node->name ; *p ; p++ ) {
	if ( carry = (hashval & 0xFF000000) )
	    hashval = (hashval << 8) | ((carry >> 24) ^ ((unsigned long) *p)) ;
	else
	    hashval = (hashval << 8) | ((unsigned long) *p) ;
    }
    return (hashval) ;
}

static int node_cmp_fct (tcl_node1, tcl_node2)
    TclNode tcl_node1, tcl_node2 ;

{
    return ( strcmp (tcl_node1->name, tcl_node2->name) ) ;
}

static TclNode nameToNode (tcl_server, name)
    TclServer tcl_server ;
    char      *name ;
{
    TclNodeRec node_rec ;

    if ( ! tcl_server->name_space )
        return ((TclNode) 0) ;
    node_rec.name = name ;
    return ((TclNode) KeHashtableFind (tcl_server->name_space, &node_rec)) ;
}

static void nodeRegister (tcl_server, tcl_node)
    TclServer tcl_server ;
    TclNode   tcl_node ;
{
    if ( ! tcl_server->name_space ) {
        tcl_server->name_space = KeHashtableMake (DEF_NAME_SPACE_SIZE
                                                , node_hash_fct
                                                , node_cmp_fct) ;
    }
    KeHashtableAdd (tcl_server->name_space, tcl_node) ;
    return ;
}

/*
 * Trace functions:
 */

static char *nodeWriteTrace (tcl_node_ptr, interp, name1, name2, flags)
    ClientData tcl_node_ptr ;
    Tcl_Interp *interp ;
    char       *name1, *name2 ;
    int        flags ;
{
    TclNode tcl_node = (TclNode) tcl_node_ptr ;
    char    *value ;
    
    value = Tcl_GetVar (tcl_node->tcl_server->interp
                        , tcl_node->varname
                        , TCL_GLOBAL_ONLY) ;
    if ( tcl_node->mask & KEDIT_WriteWithValueMask ) {
        unsigned int   size ;
        KePointer      coded_value ;
        char	       *emsg ;
        KeErrorFreeFct efree ;
        
        if ( ! KeCodecEncode (Tcl_stringCodec
                              , value
                              , &size
                              , &coded_value
                              , &emsg, &efree) ) {
            KesWriteWithValueNotify (tcl_node->tcl_server->server
                                     , (unsigned long) 0
                                     , tcl_node->id
                                     , size
                                     , coded_value) ;
            KeCodecEncoderFree (Tcl_stringCodec, coded_value) ;
        } else if ( efree ) {
            (efree) (emsg) ;
        }
    }
    if ( tcl_node->mask & KEDIT_WriteMask ) {
        KesWriteNotify (tcl_node->tcl_server->server
                        , (unsigned long) 0
                        , tcl_node->id) ;
    }
    return (NULL) ;;
}

static char *nodeReadTrace (tcl_node_ptr, interp, name1, name2, flags)
    ClientData tcl_node_ptr ;
    Tcl_Interp *interp ;
    char       *name1, *name2 ;
    int        flags ;
{
    TclNode tcl_node = (TclNode) tcl_node_ptr ;
    char    *value ;
    
    value = Tcl_GetVar (tcl_node->tcl_server->interp
                        , tcl_node->varname
                        , TCL_GLOBAL_ONLY) ;
    
    if ( tcl_node->mask & KEDIT_ReadWithValueMask ) {
        unsigned int   size ;
        KePointer      coded_value ;
        char	       *emsg ;
        KeErrorFreeFct efree ;
        
        if ( ! KeCodecEncode (Tcl_stringCodec
                              , value
                              , &size
                              , &coded_value
                              , &emsg, &efree) ) {
            KesReadWithValueNotify (tcl_node->tcl_server->server
                                    , (unsigned long) 0
                                    , tcl_node->id
                                    , size
                                    , coded_value) ;
            KeCodecEncoderFree (Tcl_stringCodec, coded_value) ;
        } else if ( efree ) {
            (efree) (emsg) ;
        }
    }
    if ( tcl_node->mask & KEDIT_ReadMask ) {
        KesReadNotify (tcl_node->tcl_server->server
                       , (unsigned long) 0
                       , tcl_node->id) ;
    }
    return (NULL) ;
}

/*
 * Node comands:
 */

static int Tcl_DirectoryCmd (tcl_node_ptr, interp, argc, argv)
    ClientData tcl_node_ptr ;
    Tcl_Interp *interp ;
    int        argc ;
    char       *argv[] ;
{
    TclNode tcl_node = (TclNode) tcl_node_ptr ;
    return (TCL_OK) ;
}

static int Tcl_NodeCmd (tcl_node_ptr, interp, argc, argv)
    ClientData tcl_node_ptr ;
    Tcl_Interp *interp ;
    int        argc ;
    char       *argv[] ;
{
    TclNode tcl_node = (TclNode) tcl_node_ptr ;
    
    return (nodeConfigure(interp
                          , tcl_node->tcl_server
                          , tcl_node
                          , argc - 1, argv + 1));
}

/*
* Server stuff:
 */

static jmp_buf err_jmpbuf ;

static void iceErrorHandler (ice_conn)
    IceConn ice_conn ;
{
    longjmp (err_jmpbuf, 1) ;
    /* NOT REACHED */
}

static void clientHandler (ice_conn_ptr, mask)
    ClientData ice_conn_ptr ;
    int        mask ;
{
    IceConn ice_conn = (IceConn) ice_conn_ptr ;
    Bool    reply_ready ;

    if ( setjmp (err_jmpbuf) ) {
        Tk_DeleteFileHandler (IceConnectionNumber (ice_conn)) ;
    } else {
        IceProcessMessages (ice_conn, NULL, &reply_ready) ;
    }
    return ;
}

static void connectionHandler (listen_obj_ptr, mask)
    ClientData listen_obj_ptr ;
    int        mask ;
{
    IceListenObj    listen_obj = (IceListenObj) listen_obj_ptr ;
    IceAcceptStatus status ;
    IceConn         ice_conn = IceAcceptConnection (listen_obj, &status) ;
    Bool            reply_ready ;

    if ( ice_conn ) {
        do {
            IceProcessMessages (ice_conn, NULL, &reply_ready) ;
        } while ( IceConnectionStatus (ice_conn) == IceConnectPending ) ;
        if ( IceConnectionStatus (ice_conn) != IceConnectAccepted ) 
            return ;
        Tk_CreateFileHandler (IceConnectionNumber (ice_conn)
                              , TK_READABLE
                              , clientHandler
                              , (ClientData) ice_conn) ;
    }
    return ;
}

static int createDirectory (interp, tcl_server, path, nopt, opts)
    Tcl_Interp *interp ;
    TclServer  tcl_server ;
    char       *path ;
    int        nopt ;
    char       *opts[] ;
{
    TclNode      tcl_node ;
    char         *p, *name ;
    KesDirectory father ;
    int          err ;
    
    if ( nameToNode (tcl_server, path) ) {
        Tcl_AppendResult (interp, "Already existing node.", NULL) ;
        return (TCL_ERROR) ;
    }
    name = strdup (path) ;
    p = strrchr (name, '.') ;
    if ( ! p ) {
        Tcl_AppendResult (interp, "bad path name \"", path, "\"", NULL) ;
        free (name) ;
        return (TCL_ERROR) ;
    }
    *p = '\0' ;
    /* Get father directory */
    if ( ! strcmp (KesNameOfServer (tcl_server->server), name) ) { /* root */
        father = KesRootDirectoryOfServer (tcl_server->server) ;
    } else {							   /* any */
        TclNode node = nameToNode (tcl_server, name) ;
        if ( ( ! node ) || ( ! node->directory ) ) {
            free (name) ;
            Tcl_AppendResult (interp
                              , "Invalid father node \"", name, "\"", NULL);
            return (TCL_ERROR) ;
        }
        father = node->directory ;
    }
    /* Create the directory */
    tcl_node = (TclNode) malloc (sizeof (TclNodeRec)) ;
    memset (tcl_node, 0, sizeof (TclNodeRec)) ;
    *p = '.' ;
    tcl_node->tcl_server = tcl_server ;
    tcl_node->name       = name ;
    tcl_node->directory  = KesDirectoryMake (tcl_server->server, p+1) ;
    KesDirectoryAddChild (father, KesNodeOfDirectory (tcl_node->directory)) ;
    /* Create the command and resgiter the node */
    Tcl_CreateCommand (interp
                       , tcl_node->name
                       , Tcl_DirectoryCmd
                       , (ClientData) tcl_node
                       , (Tcl_CmdDeleteProc *) 0) ;
    nodeRegister (tcl_server, tcl_node) ;
    return (TCL_OK) ;
}

static int nodeConfigure (interp, tcl_server, tcl_node, nopt, opts)
    Tcl_Interp *interp ;
    TclServer  tcl_server ;
    TclNode    tcl_node ;
    int        nopt ;
    char       **opts ;
{
    while ( nopt-- ) {
        if ( nopt && ! strcmp (*opts, "-variable") ) {
            if ( tcl_node->varname )
                free (tcl_node->varname) ;
            tcl_node->varname = (char *) strdup (*++opts) ;
            nopt-- ;
        } else if ( nopt && ! strcmp (*opts, "-command") ) {
            if ( tcl_node->command )
                free (tcl_node->command) ;
            tcl_node->command = (char *) strdup (*++opts) ;
            nopt-- ;
        } else if ( tcl_node->command && ! strcmp (*opts, "+exec") ) {
            tcl_node->mask |= KEDIT_ExecMask ;
        } else if ( ! strcmp (*opts, "-exec") ) {
            tcl_node->mask &= (~KEDIT_ExecMask) ; 
        } else if ( tcl_node->varname && ! strcmp (*opts, "+read") ) {
            tcl_node->mask |= KEDIT_ReadMask ;
            if ( ! tcl_node->rtrace_count ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_READS
                              , nodeReadTrace
                              , (ClientData) tcl_node) ;
            }
            tcl_node->rtrace_count++ ;
        } else if ( ! strcmp (*opts, "-read") ) {
            tcl_node->mask &= (~KEDIT_ReadMask) ;
            if ( ! --(tcl_node->rtrace_count) ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_WRITES
                              , nodeReadTrace
                              , (ClientData) tcl_node) ;
            }
        } else if (tcl_node->varname && ! strcmp (*opts, "+read-with-value")) {
            tcl_node->mask |= KEDIT_ReadWithValueMask ;
            if ( ! tcl_node->rtrace_count ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_READS
                              , nodeReadTrace
                              , (ClientData) tcl_node) ;
            }
            tcl_node->rtrace_count++ ;
        } else if ( ! strcmp (*opts, "-read-with-value") ) {
            tcl_node->mask &= (~KEDIT_ReadWithValueMask) ;
            if ( ! --(tcl_node->rtrace_count) ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_WRITES
                              , nodeReadTrace
                              , (ClientData) tcl_node) ;
            }
        } else if ( tcl_node->varname && ! strcmp (*opts, "+write") ) {
            tcl_node->mask |= KEDIT_WriteMask ;
            if ( ! tcl_node->wtrace_count ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_WRITES
                              , nodeWriteTrace
                              , (ClientData) tcl_node) ;
            }
            tcl_node->wtrace_count++ ;
        } else if ( ! strcmp (*opts, "-write") ) {
            tcl_node->mask &= (~KEDIT_WriteMask) ;
            if ( ! --(tcl_node->wtrace_count) ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_WRITES
                              , nodeWriteTrace
                              , (ClientData) tcl_node) ;
            }
        } else if ( tcl_node->varname && ! strcmp(*opts,"+write-with-value")) {
            tcl_node->mask |= KEDIT_WriteWithValueMask ;
            if ( ! tcl_node->wtrace_count ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_WRITES
                              , nodeWriteTrace
                              , (ClientData) tcl_node) ;
            }
            tcl_node->wtrace_count++ ;
        } else if ( ! strcmp (*opts, "-write-with-value") ) {
            tcl_node->mask &= (~KEDIT_WriteWithValueMask) ;
            if ( ! --(tcl_node->wtrace_count) ) {
                Tcl_TraceVar (tcl_node->tcl_server->interp
                              , tcl_node->varname
                              , TCL_TRACE_WRITES
                              , nodeWriteTrace
                              , (ClientData) tcl_node) ;
            }
        } else {
            Tcl_AppendResult (interp, "invalid argument", NULL) ;
            return (TCL_ERROR) ;
        }
    }
    return (TCL_OK) ;
}

static int createNode (interp, tcl_server, path, nopt, opts)
    Tcl_Interp *interp ;
    TclServer  tcl_server ;
    char       *path ;
    int        nopt ;
    char       *opts[] ;
{
    TclNode      tcl_node ;
    char         *p, *name=path ;
    KesDirectory father ;
    int          err ;
    
    if ( nameToNode (tcl_server, path) ) {
        Tcl_AppendResult (interp, "Already existing node.", NULL) ;
        return (TCL_ERROR) ;
    }
    p = strrchr (name, '.') ;
    if ( ! p ) {
        Tcl_AppendResult (interp, "bad path name \"", path, "\"", NULL) ;
        return (TCL_ERROR) ;
    }
    *p = '\0' ;
    /* Get father directory */
    if ( ! strcmp (KesNameOfServer (tcl_server->server), name) ) { /* root */
        father = KesRootDirectoryOfServer (tcl_server->server) ;
    } else {							   /* any */
        TclNode node = nameToNode (tcl_server, name) ;
        if ( ( ! node ) || ( ! node->directory ) ) {
            Tcl_AppendResult (interp
                              , "Invalid father node \"", name, "\"", NULL);
            return (TCL_ERROR) ;
        }
        father = node->directory ;
    }
    /* Create the node */
    tcl_node = (TclNode) malloc (sizeof (TclNodeRec)) ;
    memset (tcl_node, 0, sizeof (TclNodeRec)) ;
    *p = '.' ;
    tcl_node->tcl_server = tcl_server ;
    tcl_node->name       = strdup(p+1) ;
    if ( KesNodeMake (tcl_server->server
                      , Tcl_handlerSet
                      , (KePointer) tcl_node
                      , &(tcl_node->id)) != KEDIT_Success ) {
        Tcl_AppendResult (interp, "KesNodeMake failed.", NULL) ;
        return (TCL_ERROR) ;
    }
    /* Configure node */
    if ( err = nodeConfigure (interp, tcl_server, tcl_node, nopt, opts) ) {
        /* BUG (should free tcl_node) */
        return (err) ;
    }
    KesDirectoryAddChild (father, tcl_node->id) ;
    /* Create the command and register node */
    Tcl_CreateCommand (interp
                       , name
                       , Tcl_NodeCmd
                       , (ClientData) tcl_node
                       , (Tcl_CmdDeleteProc *) 0) ;
    nodeRegister (tcl_server, tcl_node) ;
    return (TCL_OK) ;
}

/*
 * Syntax:
 * <server> <path>: create a new node, named path.
 * <server> -directory <path>: create a new directory node, named path
 * <server> -netids: return net id.
 */

int Tcl_ServerCmd (tcl_server_ptr, interp, argc, argv)
    ClientData tcl_server_ptr ;
    Tcl_Interp *interp ;
    int        argc ;
    char       *argv[];
{
    TclServer tcl_server = (TclServer) tcl_server_ptr ;

    if ( argc < 2 ) {
        Tcl_AppendResult (interp
                          , "wrong # args: should be kedit:server option ?arg?"
                          , NULL);
        return (TCL_ERROR) ;
    }
    if ( (argc == 3) && ! strcmp (argv[1], "-directory") ) {
        return (createDirectory (interp, tcl_server, argv[2], argc-3, argv+3));
    } else if ( (argc == 2) && ! strcmp (argv[1], "-netids") ) {
        int i ;
        for (i = 0 ; i < tcl_server->listen_count ; i++ ) {
            Tcl_AppendResult
                (interp
                 , IceGetListenConnectionString (tcl_server->listens[i])
		 , " "
		 , NULL) ;
        }
        return (TCL_OK) ;
    } else {
        return (createNode (interp, tcl_server, argv[1], argc - 2, argv + 2)) ;
    } 
    /* NOT REACHED */
}
    
/* Authentication defaults to none :-) */
static Bool hostBasedAuthDisabled (hostname)
    char *hostname ;
{
    fprintf (stderr, "host %s with no auth accepted.\n", hostname) ;
    return (True) ;
}

int Tcl_KesServerInitializeCmd (clientData, interp, argc, argv)
    ClientData clientData ;
    Tcl_Interp *interp ;
    int        argc ;
    char       *argv[] ;
{
    TclServer    tcl_server ;
    KesServer    server ;
    int          listen_count, i ;
    IceListenObj *listens ;

    if ( argc != 2 ) {
        Tcl_AppendResult (interp
                          , "wrong # args: should be kedit:server name"
                          , NULL);
        return (TCL_ERROR) ;
    }
/*
 * Set up the server, and connection file handler:
 */
    if ( ! ( server = KesServerInitialize (argv[1]) ) ) {
        Tcl_AppendResult (interp, "KesServerInitialize failed.", NULL) ;
        return (TCL_ERROR) ;
    }
    if ( ! IceListenForConnections (&listen_count, &listens, 0, (char *) 0) ) {
        Tcl_AppendResult (interp, "ListenForConnections failed.", NULL) ;
        return (TCL_ERROR) ;
    }
    tcl_server = (TclServer) malloc (sizeof (TclServerRec)) ;
    memset (tcl_server, 0, sizeof (TclServerRec)) ;
    tcl_server->interp       = interp ;
    tcl_server->server       = server ;
    tcl_server->listen_count = listen_count ;
    tcl_server->listens      = listens ;
    for (i = 0 ; i < listen_count ; i++ ) {
        IceSetHostBasedAuthProc (listens[i], hostBasedAuthDisabled) ;
        Tk_CreateFileHandler (IceGetListenConnectionNumber (listens[i])
                              , TK_READABLE
                              , connectionHandler
                              , (ClientData) listens[i]) ;
    }
/*
 * Create the new command
 */
    KesCodecRegister (tcl_server->server
                      , Tcl_stringCodec
                      , &(tcl_server->string_codec)) ;
    Tcl_CreateCommand (interp
                       , (char *) strdup (argv[1])
                       , Tcl_ServerCmd
                       , (ClientData) tcl_server
                       , (Tcl_CmdDeleteProc *) 0) ;
    return (TCL_OK) ;
}

/*
 * Handler set for Tcl nodes:
 */

static KeStatus Tcl_statHandler (req, rep)
    KesStatReq req ;
    KesReply   rep ;
{
    TclNode tcl_node = (TclNode) KesClosureOfRequest (req) ;
    unsigned long mode = 0 ;

    if ( tcl_node->directory )
        mode |= KEDIT_DirectoryMode ;
    if ( tcl_node->varname )
        mode |= (KEDIT_ReadMode | KEDIT_WriteMode) ;
    if ( tcl_node->command )
        mode |= KEDIT_ExecMode ;
    
    KesStatReplyFillStatic (rep
                            , tcl_node->name
                            , mode
                            , tcl_node->mask
                            , tcl_node->tcl_server->string_codec 
                            , 0) ;
    return (KEDIT_Success) ;
}

static KeStatus Tcl_setHandler (req, rep)
    KesSetReq req ;
    KesReply rep ;
{
    TclNode        tcl_node = (TclNode) KesClosureOfRequest (req) ;
    KePointer      value = (KePointer) 0 ;
    KeStatus       status ;
    char           *emsg ;
    KeErrorFreeFct efree ;
    
    if ( ! tcl_node->varname ) {
        KesErrorReplyFillStatic (rep
                                 , KEDIT_InvalidOp
                                 , "Node doesn't support set.") ;
        return (KEDIT_Success) ;
    }
    if ( status = KeCodecDecode (Tcl_stringCodec
                                 , req->size
                                 , (KePointer) req->value
                                 , KesClientSwapped (req->client)
                                 , &value
                                 , &emsg, &efree) ) {
        KesErrorReplyFill (rep, status, emsg, (KesReplyFreeFct) efree) ;
        return (KEDIT_Success) ;
    }
    Tcl_SetVar (tcl_node->tcl_server->interp
                , tcl_node->varname
                , value
                , TCL_GLOBAL_ONLY) ;
    return (KEDIT_Success) ;
}

static KeStatus Tcl_getHandler (req, rep)
    KesGetReq req ;
    KesReply  rep ;
{
    TclNode        tcl_node = (TclNode) KesClosureOfRequest (req) ;
    char           *value ;
    KeStatus       status ;
    char           *emsg ;
    KeErrorFreeFct efree ;
    int            size ;
    KePointer      coded_value ;
    
    if ( ! tcl_node->varname ) {
        KesErrorReplyFillStatic (rep
                                 , KEDIT_InvalidOp
                                 , "Node doesn't support set.") ;
        return (KEDIT_Success) ;
    }
    value = Tcl_GetVar (tcl_node->tcl_server->interp
                        , tcl_node->varname
                        , TCL_GLOBAL_ONLY) ;
    if ( status = KeCodecEncode (Tcl_stringCodec
                                 , value
                                 , &size
                                 , &coded_value
                                 , &emsg, &efree) ) {
        KesErrorReplyFill (rep, status, emsg, (KesReplyFreeFct) efree) ;
        return (KEDIT_Success) ;
    }
    KesGetReplyFill (rep
                     , size
                     , coded_value
                     , (KesReplyFreeFct)KeCodecEncoderFreeer(Tcl_stringCodec));
    return (KEDIT_Success) ;
}

static KeStatus Tcl_notifyHandler (req, rep)
    KesNotifyReq req ;
    KesReply     rep ;
{
    TclNode tcl_node = (TclNode) KesClosureOfRequest (req) ;

    if ( tcl_node->mask & req->mask ) {
        KesNodeNotifySet (req->client, req->node, req->mask) ;
    } else {
        KesErrorReplyFillStatic (rep
                                 , KEDIT_InvalidOp
                                 , "Invalid notification mask.") ;
    }
    return (KEDIT_Success) ;
}

static KeStatus Tcl_execHandler (req, rep)
    KesExecReq req ;
    KesReply   rep ;
{
    TclNode tcl_node = (TclNode) KesClosureOfRequest (req) ;

    if  ( tcl_node->command ) {
        Tcl_Eval (tcl_node->tcl_server->interp, tcl_node->command) ;
        if ( tcl_node->mask & KEDIT_ExecMask )
            KesExecNotify (tcl_node->tcl_server->server
                           , req->client->client_tag
                           , req->node) ;
    } else {
        KesErrorReplyFillStatic (rep
                                 , KEDIT_InvalidOp
                                 , "No command associated with node.") ;
    }
    return (KEDIT_Success) ;
}

void Tcl_ServerInit (interp)
    Tcl_Interp *interp ;
{
    Tcl_CreateCommand (interp
                       , "kes:server"
                       , Tcl_KesServerInitializeCmd
                       , (ClientData) 0
                       , (Tcl_CmdDeleteProc *) 0) ;
    Tcl_stringCodec = KeCodecGet ("STRING") ;
    Tcl_handlerSet = KesHandlerSetMake () ;
    KesHandlerSetStatHandler (Tcl_handlerSet, Tcl_statHandler) ;
    KesHandlerSetSetHandler (Tcl_handlerSet, Tcl_setHandler) ;
    KesHandlerSetGetHandler (Tcl_handlerSet, Tcl_getHandler) ;
    KesHandlerSetNotifyHandler (Tcl_handlerSet, Tcl_notifyHandler) ;
    KesHandlerSetExecHandler (Tcl_handlerSet, Tcl_execHandler) ;
    IceSetIOErrorHandler (iceErrorHandler) ;
    return ;
}
