/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * iceio.c
 *
 * IO functions for ICE emulation library.
 ***************************************************************************/

#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>

extern int errno ;

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICElibint.h>

#if defined(EAGAIN) && defined(EWOULDBLOCK)
#define ETEST(err) ( (err == EAGAIN) || (err == EWOULDBLOCK) )
#else
#if defined(EAGAIN)
#define ETEST(err) (err == EAGAIN)
#else
#define ETEST(err) (err == EWOULDBLOCK)
#endif
#endif

/*
 * This file is in some part really tricky. Lets get some debug traces:
 */

/* #define ICEIO_DEBUG */

#ifdef ICEIO_DEBUG
#  define ICEIO_TRACE(x) x
#else
#  define ICEIO_TRACE(x)
#endif

/*
 * Buffer list management:
 */

typedef struct _IceBuf {
    struct _IceBuf *next ;		/* chain them in free list */
    int		   size ;		/* size of this buffer */
    char	   buffer[1] ;		/* ! variable length struct */
} *IceBuf, IceBufRec ;

static IceBuf IceFreeBuffers     = (IceBuf) 0 ;
static int    IceFreeBufferCount = 0 ; 

/***************************************************************************
 * _IceWaitForWritable: wait for this connection to be writable
 * _IceWaitForReadable: wait for connection to be readable
 *
 * RETURNS: when > 0 when writable, < 0 on error
 ***************************************************************************/

int _IceWaitForWritable (ice_conn) 
    IceConn ice_conn ;

{
    fd_set wfds ;
    int    found ;

    FD_ZERO ( &wfds ) ;
    FD_SET  (ice_conn->fd, &wfds) ;
    return ( select (ice_conn->fd + 1, NULL, &wfds, NULL, NULL) ) ;
}

int _IceWaitForReadable (ice_conn) 
    IceConn ice_conn ;

{
    fd_set rfds ;

    FD_ZERO ( &rfds ) ;
    FD_SET  (ice_conn->fd, &rfds) ;
    return ( select (ice_conn->fd + 1, &rfds, NULL, NULL, NULL) ) ;
}

/***************************************************************************
 * _IceSecureWrite: write safely the given data to socket
 *
 * As sockets are turned into NODELAY, this is to be done with care. 
 *
 * RETURNS: # of bytes written, or < 0 
 ***************************************************************************/

int _IceSecureWrite (ice_conn, buffer, size)
    IceConn ice_conn ;
    char    *buffer ;
    int     size ;

{
    int	    not_written = size ;
    int     result ;

    while ( 1 )
    {
	if ( (result = write (ice_conn->fd, buffer, not_written)) <= 0 )
	{
	    if ( (result < 0) && ( ! ETEST (errno) ) )
		return (-1) ;
	}
	else
	{
            buffer += result ;
	    not_written -= result ;
	}

	if ( ! not_written )
	    break ;

	if ( _IceWaitForWritable (ice_conn) <= 0 )
	    return (-1) ;
    }

    return (size) ;
}

/***************************************************************************
 * _IceDumpInputBuffer
 * _IceDumpOutputBuffer
 *
 * These function print a summary of the request list in resp. buffers. For
 * debug purposes only !
 *
 * RETURN: void
 ***************************************************************************/

static void _IceDumpOutputBuffer (ice_conn)
    IceConn ice_conn ;

{
    char	  *reqbase ;
    int           obufcnt ;
    unsigned long length ;

    reqbase = ice_conn->obuf ;
    obufcnt = ice_conn->obufptr - ice_conn->obuf ;

    fprintf (stderr, "_IceDumpOutputBuffer: %d bytes\n", obufcnt) ;

    while ( (obufcnt > 0) && (obufcnt > sizeof (iceHeader)) )
    {
	int opcode = ((iceHeader *) reqbase)->minorOpcode ;

	ICE_REQUEST_LENGTH (ice_conn, reqbase, length) ;

	if ( opcode != 4 )
	{
	    fprintf (stderr, "\t]%4d[opcode %d length %d\n"
		     , obufcnt
		     , opcode
		     , ((iceHeader *) reqbase)->length) ;
	}
	else
	{
	    fprintf (stderr, "\t]%4d[opcode %d length %d on node %d\n"
		     , obufcnt
		     , opcode
		     , ((iceHeader *) reqbase)->length
		     , *(((CARD32 *) reqbase) + 2)) ;
	}

	reqbase += (sizeof (iceHeader) + length) ;
	obufcnt -= (sizeof (iceHeader) + length) ;
    }

    return ;
}

static void _IceDumpInputBuffer (ice_conn)
    IceConn ice_conn ;

{
    char          *reqbase ;
    int           ibufcnt ;
    unsigned long length ;

    reqbase = ice_conn->ibuf ;
    ibufcnt = ice_conn->ibufcnt ;

    fprintf (stderr, "_IceDumpInputBuffer: %d bytes\n", ibufcnt) ;

    while ( (ibufcnt > 0) && ( ibufcnt > sizeof (iceHeader)) )
    {
	int opcode = ((iceHeader *) reqbase)->minorOpcode ;

	ICE_REQUEST_LENGTH (ice_conn, reqbase, length) ;
	
	if ( opcode != 4 )
	{
	    fprintf (stderr, "\t[%4d]opcode %d length %d\n"
		     , ibufcnt
		     , ((iceHeader *) reqbase)->minorOpcode 
		     , ((iceHeader *) reqbase)->length) ;
	}
	else
	{
	    fprintf (stderr, "\t[%4d]opcode %d length %d on node %d\n"
		     , ibufcnt
		     , opcode
		     , ((iceHeader *) reqbase)->length
		     , *(((CARD32 *) reqbase) + 2)) ;
	}

	reqbase += (sizeof (iceHeader) + length)  ;
	ibufcnt -= (sizeof (iceHeader) + length) ;
    }

    return ;
}
	

/***************************************************************************
 * _IceSecureRead: read safely the given # of bytes
 *
 * Needed because of the O_NDELAY stuff
 *
 * RETURNS: # of readed bytes
 ***************************************************************************/

int _IceSecureRead (ice_conn, buffer, size)
    IceConn ice_conn ;
    char    *buffer ;
    int     size ;

{
    int     not_read = size ;
    int     result ;

    while ( 1 )
    {
	if ( (result = read (ice_conn->fd, buffer, not_read)) <= 0 )
	{
	    if ( (result < 0) && ( ! ETEST (errno) ) )
		return (-1) ;
	}
	else
	{
            buffer += result ;
	    not_read -= result ;
	}

	if ( ! not_read )
	    break ;

	if ( _IceWaitForReadable (ice_conn) <= 0 )
	    _IceIOError (ice_conn) ;
    }

    return (size) ;
}

/***************************************************************************
 * IceFlush         : flush the output buffer freing output buffer
 * _IceInternalFlush: flush the output, keeping current buffer
 *
 * IceFlush is the secure one: it will write everything in the output buffer
 * before returning (this is what the user want).
 *
 * - I would have like to have a _IceInternalFlush() function, which purpose
 * would be to flush just what the net accept at this time. Unfortunately,
 * I cannot see how to resync on padding without having to select/write
 * after the first flush' write.
 *
 * RETURNS: void (but might invoke some IO error handler)
 ***************************************************************************/

void _IceInternalFlush (ice_conn)
    IceConn ice_conn ;

{
/*
 * No test against NULL obuf, because the following does it implicitly :
 */  

    if ( ice_conn->obufptr > ice_conn->obuf )
    {
	if ( _IceSecureWrite (ice_conn
			      , ice_conn->obuf
			      , ice_conn->obufptr - ice_conn->obuf) < 0 ) 
	{
	    _IceIOError (ice_conn) ;
	}
	
	ICEIO_TRACE ( fprintf (stderr, "_IceInternalFlush: wrote %d bytes x%08x\n"
			       , ice_conn->obufptr - ice_conn->obuf
			       , ice_conn->obuf) ; ) ;

	ice_conn->obufptr = ice_conn->obuf ;
    }
}

void IceFlush (ice_conn)
    IceConn ice_conn ;

{
/*
 * This one is the external: it frees the output buffer :
 */

    if ( ice_conn->obufobj )
    {
	_IceInternalFlush (ice_conn) ;
	_IceReleaseBuffer (ice_conn->obufobj) ;

	ice_conn->obufobj = (void *) 0 ;
	ice_conn->obuf    = (char *) 0 ;
	ice_conn->obufptr = (char *) 0 ;
	ice_conn->obuflen = 0 ;
    }

    return ;
}

/***************************************************************************
 * _IceAllocateBuffer : allocate a fresh buffer
 * _IceGetOutputBuffer: allocate an output buffer for this connection
 * _IceGetInputBuffer : allocate an input buffer for this connection
 *
 * The given ice_conn is updated. If no buffer could be allocated, this
 * function sets pointers to NULL. The caller must check against that.
 *
 * RETURNS: pointer to buffer
 ***************************************************************************/

static IceBuf _IceAllocateBuffer (ice_conn, size)
    IceConn ice_conn ;
    int     size ;

{
    IceBuf buf, *pbuf ;

/*
 * Set the default size. Can we find this in free-list ?
 */

    if ( size < ICE_BUF_DEFSIZE )
	size = ICE_BUF_DEFSIZE ;

    if ( size > ICE_BUF_WATERMARK ) 
	goto allocate_it ;

/*
 * Ok, look in free-list :
 */

    if ( IceFreeBufferCount )
    {
	buf  = IceFreeBuffers ;
	pbuf = &IceFreeBuffers ;
	while ( buf )
	{
	    if ( buf->size >= size )
	    {
		IceFreeBufferCount-- ;
		*pbuf = buf->next ;
		return (buf) ;
	    }
	    pbuf = &(buf->next) ;
	    buf  = buf->next ;
	}
    }

/*
 * Enforce this buffer allocation:
 */

  allocate_it:

    if ( ! ( buf = (IceBuf) malloc (sizeof (IceBufRec) + size)) ) 
	return (buf) ;
    
    buf->size = size ;
    buf->next = (IceBuf) 0 ;
    
    ICEIO_TRACE ( fprintf (stderr, "_IceAllocateBuffer: allocating new buffer x%08x\n"
			   , buf) ;) ;

    return (buf) ;
}
    
static IceBuf _IceResizeBuffer (buf, size)
    IceBuf buf ;
    int    size ;

{
    int    new_size = sizeof (IceBufRec) + (size & ( ~ 0x3FF)) + 1024 ;
    IceBuf new_buf ;


    if ( new_buf = (IceBuf) realloc (buf, new_size) )
    {
	ICEIO_TRACE ( fprintf (stderr, "_IceResizeBuffer: resizing x%08x to %d -> x%08x\n"
			       , buf
			       , new_size
			       , new_buf) ; ) ;
	new_buf->size = new_size ;
    }

    return (new_buf) ;
}

void _IceReleaseBuffer (buf)
    IceBuf buf ;

{
    if ( ( IceFreeBufferCount < ICE_BUF_KEEPTHEM )
	&& ( buf->size < ICE_BUF_WATERMARK ) )
    {
	buf->next      = IceFreeBuffers ;
	IceFreeBuffers = buf ;
	IceFreeBufferCount++ ;
	return ;
    }

    free (buf) ;

    return ;
}

void _IceGetOutputBuffer (ice_conn, size)
    IceConn ice_conn ;
    int	    size ;

{
    IceBuf buf = ice_conn->obufobj ;

    if ( ! buf )
    {
	if ( ! ( buf = _IceAllocateBuffer (ice_conn, size) ) )
	    _IceIOError (ice_conn) ;
	ice_conn->obufobj = buf ;
	ice_conn->obuf    = ice_conn->obufptr = buf->buffer ;
	ice_conn->obuflen = buf->size ;
    }
    else if ( ice_conn->obuflen > size ) 
    {
	_IceInternalFlush (ice_conn) ;
    }
    else 
    {
	_IceInternalFlush (ice_conn) ;

	if ( ! ( buf = _IceResizeBuffer (buf, size)) )
	    _IceIOError (ice_conn) ;

	ice_conn->obufobj = buf ;
	ice_conn->obuf    = buf->buffer ;
	ice_conn->obufptr = ice_conn->obuf ;
	ice_conn->obuflen = buf->size ;

	return ;
    }

    return ;
}

/***************************************************************************
 * _IceGetRequest: retreive request from input buffer
 *
 * This function gets the next request from the given connection. If the
 * returned request is the last one, and no requests are pending to be
 * fully read, it returns as pfreep True. The caller can than invoke
 * _IceReleaseBuffer, once the returned request has been processed.
 *
 * RETURNS: a request until no more.
 ***************************************************************************/

iceHeader *_IceGetMessage (ice_conn, pfreep)
    IceConn ice_conn ;
    int     *pfreep ;

{
    Bool           done = False ;
    unsigned long  length ;
    char	   *reqbase ;

/*
 * Check current request against needed bytes:
 */

    ICEIO_TRACE (fprintf (stderr, "IceGetMessage: ibufcnt %d x%08x < x%08x < x%08x\n"
			  , ice_conn->ibufcnt
			  , ice_conn->ibuf
			  , ice_conn->reqbase
			  , ice_conn->ibuf + ice_conn->ibuflen) ; ) ;

    *pfreep = 0 ;

    ice_conn->processing = (char *) 0 ;
    if ( ( ! ice_conn->reqbase ) || ( ice_conn->needed ) ) 
    {
	return ((iceHeader *) 0) ;
    }

/*
 * Check that a header is ready (if not return NULL):
 */

    if ( ice_conn->ibufcnt < sizeof (iceHeader) ) 
    {

	ICEIO_TRACE ( fprintf (stderr, "IceGetMessage: no header.\n") ; ) ;

	ice_conn->needed = sizeof (iceHeader) - ice_conn->ibufcnt ;
	if ( ice_conn->reqbase != ice_conn->ibuf )
	{
	    bcopy (ice_conn->reqbase, ice_conn->ibuf, ice_conn->ibufcnt) ;
	    ice_conn->reqbase = ice_conn->ibuf ;
	    ice_conn->ibufptr = ice_conn->ibuf + ice_conn->ibufcnt ;
	}
	
	return ((iceHeader *) 0) ;
    }

/*
 * Header is ready, get length of request, see if it is fully read:
 */

    ICE_REQUEST_LENGTH (ice_conn, ice_conn->reqbase, length) ;

    if ( length + sizeof (iceHeader) > ice_conn->ibufcnt ) 
    {
	/*
	 * Setup for this request:
	 * - Set needed to number of needed bytes to complete this request
	 * - Shift buffer to left
	 * - Resize buffer if request length doesn't hold in current
	 */

	ice_conn->needed = length + sizeof (iceHeader) - ice_conn->ibufcnt ;

	if ( ice_conn->reqbase != ice_conn->ibuf )
	{
	    bcopy (ice_conn->reqbase, ice_conn->ibuf, ice_conn->ibufcnt) ;
	    ice_conn->reqbase = ice_conn->ibuf ;
	    ice_conn->ibufptr = ice_conn->ibuf + ice_conn->ibufcnt ;
	}

	if ( ice_conn->ibuflen < length + sizeof (iceHeader) )
	{
	    IceBuf newbuf ;
	    int    ptrloc = ice_conn->ibufptr - ice_conn->ibuf ;
	    int    bufneeded = length + sizeof (iceHeader) ;

	    if ( newbuf = _IceResizeBuffer (ice_conn->ibufobj, bufneeded) )
	    {
		ice_conn->ibufobj = newbuf ;
		ice_conn->ibuf    = newbuf->buffer ;
		ice_conn->ibufptr = ice_conn->ibuf + ptrloc ;
		ice_conn->reqbase = ice_conn->ibuf ;
		ice_conn->ibuflen = newbuf->size ;
	    }
	}

	return ((iceHeader *) 0) ;
    }

/*
 * At least one full request ready (in reqbase):
 */

    ice_conn->processing = ice_conn->reqbase ;
    ice_conn->ibufcnt   -= (length + sizeof (iceHeader)) ;
    ice_conn->reqbase   += (length + sizeof (iceHeader)) ;

    ICEIO_TRACE (fprintf (stderr, "IceGetMessage: full request (%d bytes)\n", length) ;) ;

    if ( ! ice_conn->ibufcnt )
    {
	/*
	 * Buffer has been fully emptied. We can now mark it as candidate
	 * for a free on next call
	 */

	*pfreep = 1 ;
	ice_conn->reqbase = (char *) 0 ;
	
	ICEIO_TRACE (fprintf (stderr, "IceGetMessage: input buffer can be freed\n") ;) ;

    }
	
    return ((iceHeader *) ice_conn->processing) ;
}

/***************************************************************************
 * IceReadConnection: read bytes for this connection
 *
 * In my sense, this should really be in the spec. It reads available bytes
 * from the connection into the ice_conn buffer, and signal if a full 
 * request is ready to be processed by IceProcessMessage. 
 * Anyawy, let's glue to the ICE lib specs.
 *
 * RETURNS: > 0 if requests are ready (might invoke _IceIOError)
 ***************************************************************************/

int IceReadConnection (ice_conn)
    IceConn ice_conn ;

{
    int           result ;

/*
 * Buffer management:
 * - Make sure current IceConn has an input buffer.
 */

    if ( ! ice_conn->ibuf )
    {
	IceBuf buf = _IceAllocateBuffer (ice_conn, ICE_BUF_DEFSIZE) ;

	if ( ! buf )
	    _IceIOError (ice_conn) ;	/* NO RET, bad stuff */

	ice_conn->ibufobj = buf ;
	ice_conn->ibuf    = ice_conn->ibufptr = buf->buffer ;
	ice_conn->ibuflen = buf->size ;
	ice_conn->ibufcnt = 0 ;
    }

/*
 * Read available data, and get rid of errors :
 */

    result = read (ice_conn->fd
		   , ice_conn->ibufptr
		   , ice_conn->ibuflen - (ice_conn->ibufptr-ice_conn->ibuf) ) ;

    ICEIO_TRACE ( fprintf (stderr, "IceReadConnection: (%d, x%08x, %d) = %d\n"
			   , ice_conn->fd
			   , ice_conn->ibufptr
			   , ice_conn->ibuflen - (ice_conn->ibufptr - ice_conn->ibuf) 
			   , result) ; ) ;

    if ( result <= 0 ) {
	if ( (result < 0) && ETEST (errno) ) {
	    /* Select() was joking, get the joke */
	    return (0) ;
	}
	else {
	    
	    ICEIO_TRACE (fprintf (stderr, "IceReadConnection: errno = %d\n", errno) ;) ;

	    _IceIOError (ice_conn) ;
	}
    }

/*
 * Did we get (at least) a full request:
 * No current request is being read, if result is less than a header, return
 * right now, needing at least the bytes to complete header. Otherwise, setup
 * current request, and needed to its length.
 */

    ice_conn->ibufcnt += result ;

    if ( ! ice_conn->reqbase ) {
	/* We are at the starting point of a request, set up */

	ice_conn->reqbase  = (char *) ice_conn->ibufptr ;
	ice_conn->ibufptr += result ;

	if ( result < sizeof (iceHeader) ) {
	    /* Haven't read a full request, return (0) */
	    ice_conn->needed   = sizeof (iceHeader) - result ;
	    return (0) ;
	}
	else {
	    unsigned long length ;
	    int           bufneeded ;

	    result            -= sizeof (iceHeader) ;
	    ICE_REQUEST_LENGTH (ice_conn, ice_conn->reqbase, length) ;
	    ice_conn->needed    = length ;

	    /*
	     * Check buffer size against request length:
	     */

	    bufneeded = length + sizeof (iceHeader) + (ice_conn->reqbase - ice_conn->ibuf) ;

	    if ( ice_conn->ibuflen < bufneeded )
	    {
		IceBuf newbuf ;
		int    ptrloc = ice_conn->ibufptr - ice_conn->ibuf ;
		int    reqloc = ice_conn->reqbase - ice_conn->ibuf ;

		if ( newbuf = _IceResizeBuffer (ice_conn->ibufobj, bufneeded) )
		{
		    ice_conn->ibufobj = newbuf ;
		    ice_conn->ibuf    = newbuf->buffer ;
		    ice_conn->ibufptr = ice_conn->ibuf + ptrloc ;
		    ice_conn->reqbase = ice_conn->ibuf + reqloc ;
		    ice_conn->ibuflen = newbuf->size ;

		    {
			unsigned long nlength ;

			ICE_REQUEST_LENGTH (ice_conn, ice_conn->reqbase, nlength) ;
			ICEIO_TRACE ( fprintf (stderr, "IceReadConnection: buffer to %d wait %d\n"
					       , length
					       , nlength) ;) ;
		    }
		}
		else
		{
		    _IceIOError (ice_conn) ;
		}
	    }
	}
    }
    else 
    {
	ice_conn->ibufptr += result ;
    }

/*
 * We have at least a current request, perhaps even a complete one ?
 */

    if ( result >= ice_conn->needed ) {
	/* Ok, got it */
	ice_conn->needed = 0 ;
	return (1) ;
    }
    else {
	/* Not yet ! */
	ice_conn->needed -= result ;
	return (0) ;
    }

    /* NOT REACHED */

}
